/* HickCmd 1.1.0 -- Command line interface to hick.
** Copyright (C) 97, 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <fcntl.h>
#include <unistd.h>

#include "entry.h"		/* from hick lib 1.3 */
#include "common.h"
#include "getpacket.h"

#include "hickcmd.h"

#define UPDATE_DEF 0
#define UPDATE_ADD 1
#define UPDATE_REM 2
#define UPDATE_DEL 3

static char *convert_to_file(char *buf, const char *s);


int
_save_installed(const char *filename, const char *packpath,
		const char *homepath, const char *hicktype)
{
  int err;
  FILE *outFile;
  userEntry *ue;
  HickPackets pack;
  char bufName[PNAME_LEN*2+1], bufVer[PVER_LEN*2+1];

  if(filename) {
    if(!(outFile = fopen(filename, "w"))) {
      printf("error: can't open '%s' for writting.\n", filename);
      return 1;
    }
  } else {
    outFile = stdout;
  }

  if((err = openPackets(packpath, homepath, hicktype, NULL, &pack))) {
    return err;
  }

  for(ue = pack.hp_installed.ul_head; ue; ue = ue->ue_next) {
    fprintf(outFile, "%s:%s:%s\n",
	    (ue->ue_flags == UE_DEFAULT ? "def" : "add"),
	    convert_to_file(bufName, ue->ue_name),
	    convert_to_file(bufVer, ue->ue_version));
  }

  closePackets(&pack);

  if(filename) {
    fclose(outFile);
  }

  return 0;
}


int
_load_install(const char *filename, HickPackets *pack)
{
  FILE *inFile;
  size_t n, type;
  char *p, buf[64], bufName[PNAME_LEN+1], bufVer[PVER_LEN+1];

  if(filename) {
    if(!(inFile = fopen(filename, "r"))) {
      printf("error: can't open '%s' for writting.\n", filename);
      return 1;
    }
  } else {
    inFile = stdin;
  }

  while(fgets(buf, 64, inFile)) {
    for(p = buf, n = 0; *p && (n < PNAME_LEN) && (*p != ':'); p++){
      bufName[n++] = *p;
    }

    bufName[n] = (char)0;

    if(!_token_case_cmp(bufName, "def")) {
      type = UPDATE_DEF;
    } else {
      if(!_token_case_cmp(bufName, "add")) {
	type = UPDATE_ADD;
      } else {
	if(!_token_case_cmp(bufName, "rem")) {
	  type = UPDATE_REM;
	} else {
	  if(!_token_case_cmp(bufName, "del")) {
	    type = UPDATE_DEL;
	  } else {
	    printf("warning: unknown flag '%s'.\n", bufName);
	    continue;
	  }
	}
      }
    }

    if(!*p) {
      continue;
    }

    p = _get_str_word(++p, ':', ':', bufName, PNAME_LEN);

    if(!*p) {
      continue;
    }

    _get_str_word(++p, ':', '\n', bufVer, PVER_LEN);

    switch(type) {
      case UPDATE_DEF: {
	addPacket(pack, bufName, bufVer, 1, 1);
	break;
      }
      case UPDATE_ADD: {
	addPacket(pack, bufName, bufVer, 1, 0);
	break;
      }
      case UPDATE_REM: {
	removePacket(pack, bufName, bufVer, 1, 0);
	break;
      }
      case UPDATE_DEL: {
	removePacket(pack, bufName, bufVer, 1, 1);
	break;
      }
    }
  }

  if(filename) {
    fclose(inFile);
  }

  return 0;
}


static char *
convert_to_file(char *buf, const char *s)
{
  size_t i;

  for(i = 0; *s; ++i, ++s) {
    switch(*s) {
      case '\n': {
	buf[i] = '\\';
	buf[++i] = 'n';
	break;
      }
      case ':':
      case '\\': {
	buf[i] = '\\';
      }
      default: {
	buf[i] = *s;
      }
    }
  }

  buf[i] = (char)0;
  return buf;
}


int
viewLogfile(const char *homepath, const char *hicktype)
{
  int fd;
  size_t len;
  char fname[FNAME_LEN+1];

  if(hicktype) {
    hsh_create_str(fname, FNAME_LEN, homepath, "/.hick/.", hicktype,
		   "/runlog", 0);

    if((fd = open(fname, O_RDONLY, 0)) != -1) {
      while((len = read(fd, fname, FNAME_LEN))) {
	write(1, fname, len);
      }

      close(fd);
      putc('\n', stdout);
      return 1;
    }

    printf("Can't open logfile '%s'.\n", fname);
  } else {
    printf("No hicktype specified.\n");
  }

  return 0;
}
