/* HickCmd 1.1.0 -- Command line interface to hick.
** Copyright (C) 97, 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#include <hickmk.h>

#include "common.h"		/* from hick lib 1.3 */

#include "hickcmd.h"

#define RESCUE_NONE    0
#define RESCUE_IGNORE  1
#define RESCUE_PROCEED 2

#define LIST_TYPES         't'
#define LIST_GROUPS        'g'
#define LIST_PACKETS       'l'
#define LIST_INSTALLED     'd'
#define LIST_NOT_INSTALLED 'n'

#define SHOW_ALL           0
#define SHOW_INSTALLED     1
#define SHOW_NOT_INSTALLED 2

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#undef BUFSIZ
#define BUFSIZ 1024

extern char *optarg;

static int update_packets(const char *packpath, const char *homepath,
			  const char *hicktype, int script, int silence,
			  HickPackets *instInfo, const char *user,
			  const char *name, int settype);
static int update_hickmk(const char *homepath, const char *hicktype,
			 int silence);
static int check_rescue(int rescue, const char *packpath, const char *homepath,
			const char *hicktype, int script, int silence,
			const char *user, const char *name, int settype);
static void print_result(HickResult *result, const char *hicktype, int script,
			 const char *user, const char *name, int settype);
static int print_about(const char *name, const char *version,
		       const char *packpath, const char *homepath,
		       const char *hicktype);
static int list_packets(const char *packpath, const char *homepath,
			const char *hicktype, const char *group, int f);
static int list_groups(const char *packpath, const char *hicktype);
static int split_name(char *name, char *version, char *args);
static void helicopter(HickEvent *event, void *arg);
static void write_error(int err, const char *s);
static void usage(const char *name);


int
hick_command_line(int argc, char **argv)
{
  int viewLog;
  ArrayInfo types;
  unsigned int check;
  struct passwd *pwd;
  HickPackets instInfo;
  const char *packpath;
  char name[PNAME_LEN+1], version[PVER_LEN+1];
  char *hicktype, *group, *homepath, *filename, *user;
  char aboutName[PNAME_LEN+1], aboutVersion[PVER_LEN+1];
  int n, silence, script, update, rescue, rmHicktype, setHicktype, list, file;

  if(argc == 1) {
    usage(argv[0]);
    return 0;
  }

  memset(&instInfo, 0, sizeof(instInfo));

  aboutName[0] = (char)0;
  viewLog      = FALSE;
  setHicktype  = FALSE;
  rmHicktype   = FALSE;
  list         = FALSE;
  file         = FALSE;
  check        = FALSE;
  rescue       = RESCUE_NONE;
  update       = FALSE;
  script       = FALSE;
  silence      = FALSE;
  user         = NULL;
  group        = NULL;
  filename     = NULL;
  hicktype     = getenv("HICKTYPE");
  packpath     = PACKET_PATH;

  if(!(pwd = getpwuid(getuid()))) {
    printf("error: failed to obtain your passwd entries.\n");
    return 1;
  }

  homepath = pwd->pw_dir;

  while((n = getopt(argc, argv, "hvucsPpRVtlgdnCNSLa:H:G:F:U:i:r:")) != EOF) {
    switch(n) {
      case 'h': {		/* Help text (usage) */
	usage(argv[0]);
	return 0;
      }
      case 'v': {		/* version information */
	printf("%s %s\n", PROGRAM_NAME, PROGRAM_VERSION);
	return 0;
      }
      case 'a': {		/* About message for packet or version */
	if(!split_name(aboutName, aboutVersion, optarg)) {
	  return 1;
	}
	break;
      }
      case 'H': {		/* Set hick type */
	hicktype = optarg;
	setHicktype = TRUE;
	break;
      }
      case 'G': {		/* Set group */
	group = optarg;
	break;
      }
      case 'F': {		/* Install packets from file */
	filename = optarg;
	break;
      }
      case 'U': {		/* ROOT: Install as user */
	if(getuid() != 0) {
	  printf("You have to be root to use the -U option.\n");
	  return 1;
	}

	if(!(pwd = getpwnam(optarg))) {
	  printf("error: can't find the user '%s' in the passwd entries.\n",
		 optarg);
	  return 1;
	}

	user     = optarg;
	homepath = pwd->pw_dir;

	if(setgid(pwd->pw_gid)) {
	  printf("error: failed to set GID %d.\n", (int)pwd->pw_gid);
	  return 1;
	}

	if(setuid(pwd->pw_uid)) {
	  printf("error: failed to set UID %d.\n", (int)pwd->pw_uid);
	  return 1;
	}
	break;
      }
      case 'i': {		/* install packet */
	if(!split_name(name, version, optarg)) {
	  return 1;
	}

	addPacket(&instInfo, name, version, 1, (version[0] == (char)0));
	break;
      }
      case 'r': {		/* remove packet */
	if(!split_name(name, version, optarg)) {
	  return 1;
	}

	removePacket(&instInfo, name, version, 1, (version[0] == (char)0));
	break;
      }
      case 'u': {		/* Update installed packets. */
	update = TRUE;
	break;
      }
      case 'c': {		/* Generate call script */
	script = TRUE;
	break;
      }
      case 's': {		/* Silence */
	silence = TRUE;
	break;
      }
      case 'P': {		/* Ignore rescure, if any */
	rescue = RESCUE_IGNORE;
	update = TRUE;
	break;
      }
      case 'p': {		/* Do rescue if needed */
	rescue = RESCUE_PROCEED;
	update = TRUE;
	break;
      }
      case 'R': {		/* Remove hick type */
	rmHicktype = TRUE;
	break;
      }
      case 'V': {		/* View runlog file. */
	viewLog = TRUE;
	break;
      }
      case 't':			/* List availible hick types */
      case 'l':			/* List all packets */
      case 'g':			/* List availible groups */
      case 'd':			/* List all installed packets */
      case 'n': {		/* List packets which isn't installed */
	list = n;
	break;
      }
      case 'E': {		/* Examine packet with versions, or spec. */
	break;
      }
      case 'C': {		/* Check for changes in installed packets */
	check = CHECK_CHANGED;
	break;
      }
      case 'N': {		/* Check for new packets and versions */
	check = CHECK_NEW;
	break;
      }
      case 'S':			/* Save installed packets (stdout). */
      case 'L': {		/* Load packet settings (stdin).*/
	file = n;
	break;
      }
      default: {		/* ERROR! */
	usage(argv[0]);
	return 0;
      }
    }
  }

  if(setHicktype || (list == LIST_TYPES)) {
    if((n = getHickTypes(packpath, &types))) {
      write_error(n, packpath);
      return 1;
    }

    if(list == LIST_TYPES) {
      if(types.siz) {
	printf("Availible hick types:\n");

	for(n = 0; n < types.siz; n++) {
	  printf("  %s\n", types.arr[n]);
	}
      } else {
	printf("No availible hick types.\n");
      }

      return 0;
    }

    if(!(n = hsh_array_find(hicktype, &types))) {
      printf("error: there's no hick type with the name '%s'.\n", hicktype);
      return 1;
    }

    hicktype = types.arr[n - 1];
  }

  switch(list) {
    case LIST_PACKETS: {
      return !list_packets(packpath, homepath, hicktype, group, SHOW_ALL);
    }
    case LIST_GROUPS: {
      return !list_groups(packpath, hicktype);
    }
    case LIST_INSTALLED: {
      return !list_packets(packpath, homepath, hicktype, group,SHOW_INSTALLED);
    }
    case LIST_NOT_INSTALLED: {
      return !list_packets(packpath, homepath, hicktype, group,
			   SHOW_NOT_INSTALLED);
    }
  }

  if(viewLog) {			/* show the logfile */
    return !viewLogfile(homepath, hicktype);
  }

  if(aboutName[0]) {		/* Print about message for packet */
    return !print_about(aboutName, aboutVersion, packpath, homepath, hicktype);
  }

  if(check) {
    if((n = _check_packets(packpath, homepath, hicktype, check, 0))) {
      write_error(n, packpath);
      return 1;
    }

    return 0;
  }

  switch(file) {
    case 'S': {			/* save */
      if((n = _save_installed(filename, packpath, homepath, hicktype))) {
	write_error(n, packpath);
	return 1;
      }
      return 0;
    }
    case 'L': {			/* load */
      if(_load_install(filename, &instInfo)) {
	return 1;
      }

      update = TRUE;
    }
  }

  if(rmHicktype) {
    if(!hicktype) {
      printf("error: no hick type is specified.\n");
      return 1;
    }

    if(!silence) {
      printf("Removing hick type %s..  ", hicktype);
    }

    HickRemoveType(homepath, hicktype, helicopter, NULL);

    if(!silence) {
      puts("\bdone.");
    }
  }

  if(update || instInfo.hp_install.ul_head || instInfo.hp_remove.ul_head) {
    if(!hicktype) {
      printf("error: no hick type is specified.\n");
      return 1;
    }

    if((n = check_rescue(rescue, packpath, homepath, hicktype, script, silence,
			 user, argv[0], setHicktype))) {
      if(n == 2) { 
	update_hickmk(homepath, hicktype, silence);
      }
      return 0;
    }

    if((n = update_packets(packpath, homepath, hicktype, script, silence,
			   &instInfo, user, argv[0], setHicktype))) {
      update_hickmk(homepath, hicktype, silence);
    }

    return !n;
  }

  return 0;
}


static int
update_packets(const char *packpath, const char *homepath,
	       const char *hicktype, int script, int silence,
	       HickPackets *instInfo, const char *user,
	       const char *name, int settype)
{
  int err;
  HickPackets pack;
  HickResult result;
  unsigned int mask;

  if((err = openPackets(packpath, homepath, hicktype, NULL, &pack))) {
    write_error(err, packpath);
    return 0;
  }

  pack.hp_remove.ul_head = instInfo->hp_remove.ul_head;
  pack.hp_remove.ul_tail = instInfo->hp_remove.ul_tail;
  pack.hp_install.ul_head = instInfo->hp_install.ul_head;
  pack.hp_install.ul_tail = instInfo->hp_install.ul_tail;

  instInfo->hp_remove.ul_head = instInfo->hp_remove.ul_tail = NULL;
  instInfo->hp_install.ul_head = instInfo->hp_install.ul_tail = NULL;

  if(silence) {
    mask = HICKEVENT_NONE;
  } else {
    mask = (HICKEVENT_NOTUSED | HICKEVENT_INSTALL | HICKEVENT_UPDATE);

    printf("Updating packet information..  ");
    fflush(stdout);
  }

  if((err = updatePackets(&pack, &result, script, mask, helicopter, NULL))) {
    write_error(err, packpath);
    return 0;
  }

  if(!silence) {
    puts("\bdone.");

    print_result(&result, hicktype, script, user, name, settype);
  }

  closePackets(&pack);
  return 1;
}


static int
update_hickmk(const char *homepath, const char *hicktype, int silence)
{
  if(!silence) {
    printf("Generating GNU Info menu.. ");
    fflush(stdout);
  }
  
  hickmk_info_remove(homepath, hicktype);
  hickmk_info_create(INFO_PATH, homepath, hicktype, HICKMK_DEFAULT_WIDTH);

  if(!silence) {
    printf("done. \nGenerating HTML Index.. ");
    fflush(stdout);
  }

  hickmk_html_remove(homepath, hicktype);
  hickmk_html_create(NULL, 0, HTML_PATH, homepath, hicktype);

  if(!silence) {
    printf("done.\nGenerating shell scripts.. ");
    fflush(stdout);
  }

  hickmk_scripts_remove(homepath, hicktype);
  hickmk_scripts_create(NULL, 0, SCRIPT_PATH, homepath, hicktype);

  if(!silence) {
    printf("done.\nGenerating menus.. ");
    fflush(stdout);
  }

  hickmk_menus_remove(homepath, hicktype);
  hickmk_menus_create(NULL, 0, MENUS_PATH, homepath, hicktype, "Applications");

  if(!silence) {
    printf("done.\n");
  }
  return 0;
}


static int
check_rescue(int rescue, const char *packpath, const char *homepath,
	     const char *hicktype, int script, int silence, const char *user,
	     const char *name, int settype)
{
  int err;
  HickPackets pack;
  HickResult result;
  unsigned int mask;
  char buf[BUFSIZ+1];

  if((rescue == RESCUE_IGNORE) || !HickCheckRescue(homepath, hicktype)) {
    return 0;
  }

  if(rescue == RESCUE_NONE) {
    buf[0] = (char)0;

    if(script) {
      strcat(buf, " -c");
    }

    if(user) {
      strcat(buf, " -U");
      strncat(buf, user, BUFSIZ);
    }

    if(settype) {
      strncat(buf, " -H", BUFSIZ);
      strncat(buf, hicktype, BUFSIZ);
    }

    printf("Have found some indications that last time %s was runned, it somehow\nfailed to complete the installation successful. Packet information can\nhave been installed or not removed correctly, which can cause problems\nwhen installing new packets. This can be corrected by reinstalling the\npackets you had before this happend.  You can ignore this and continue\nif you choose.\n\nTo rescue install all your packets run:\n%s -p%s\n\nIf you wish to ignore this run:\n%s -P%s\n", PROGRAM_NAME, name, buf, name, buf);
    return 1;
  }

  if(silence) {
    mask = HICKEVENT_NONE;
  } else {
    mask = (HICKEVENT_NOTUSED | HICKEVENT_INSTALL | HICKEVENT_UPDATE);
  }

  if((err = HickRescueOpen(&pack, packpath, homepath, hicktype))) {
    write_error(err, packpath);
    return 1;
  }

  if(!silence) {
    printf("Prepareing for rescue..  ");
    fflush(stdout);
  }

  HickRescuePrepare(&pack, helicopter, NULL);

  if(!silence) {
    puts("\bdone.");

    printf("Running rescue..  ");
    fflush(stdout);
  }

  if((err = HickRescueUpdate(&pack, &result, script, mask, helicopter, NULL))){
    write_error(err, packpath);
    return 1;
  }

  if(!silence) {
    puts("\bdone.");

    print_result(&result, hicktype, script, user, name, settype);
  }

  return 2;
}


static void
print_result(HickResult *result, const char *hicktype, int script,
	     const char *user, const char *name, int settype)
{
  HickForced *fo;
  unsigned int i;
  HickRequire *req;
  HickFailure *fail;
  HickConflict *con;
  char buf[BUFSIZ+1];
  HickRequirePacket *rpac, *ror;

  static const char *helps[] = {
    "Failed to install %s, packet not found.\n",
    "Failed to install %s %s, version not found.\n",
    "Failed to install %s, default version not found.\n",
    "Failed to install %s %s, another version was force installed.\n",

    "Failed to update %s, packet is missing.\n",
    "Failed to update %s %s, version is missing.\n",
    "Failed to update %s, default version is missing.\n",
    "Failed to update %s %s, another version was force installed.\n",

    "Failed to remove %s, packet is force installed.\n",
    "Failed to remove %s, no such packet.\n",
    "Failed to remove %s %s, no such version.\n",

    "Packet %s is remove, it have expired.\n",
    "In %s have version %s been removed, it have expired.\n"
  };

  for(fail = result->hr_failure; fail; fail = fail->fail_next) {
    printf(helps[fail->fail_reason], fail->fail_name, fail->fail_version);
  }

  for(fo = result->hr_forced; fo; fo = fo->fo_next) {
    printf("The packet %s %s have been forced installed.\n", fo->fo_name,
	   fo->fo_version);
  }

  for(con = result->hr_conflicts; con; con = con->con_next) {
    printf("The packet %s %s was in conflict with %s %s, and not installed.\n",
	   con->con_name, con->con_version, con->con_wName, con->con_wVersion);
  }

  for(req = result->hr_required; req; req = req->req_next) {
    printf("%s %s requires", req->req_name, req->req_version);

    for(rpac = req->req_list; rpac; rpac = rpac->rpac_next) {
      printf(" %s", rpac->rpac_name);

      for(ror = rpac; ror; ror = ror->rpac_or) {
	switch(ror->rpac_versions.siz) {
	  case 0: {
	    printf(" any version");
	    break;
	  }
	  case 1: {
	    printf(" %s", ror->rpac_versions.arr[0]);
	    break;
	  }
	  default: {
	    printf(" any of the versions {");
	    for(i = 0; i < ror->rpac_versions.siz; i++) {
	      printf(" %s", ror->rpac_versions.arr[i]);
	    }
	    printf("} ");
	  }
	}

	if(ror->rpac_or) {
	  printf(" or");
	}
      }
    }

    putchar('\n');
  }

  if(result->hr_required) {
    printf("\nThere is packets which is needed to be installed to get some program to work\ncorrectly. Here is an example on how you could write.\n");

    buf[0] = (char)0;

    if(script) {
      strcat(buf, " -c");
    }

    if(user) {
      strcat(buf, " -U");
      strncat(buf, user, BUFSIZ);
    }

    if(settype) {
      strncat(buf, " -H", BUFSIZ);
      strncat(buf, hicktype, BUFSIZ);
    }

    printf("\n%s%s", name, buf);

    /* optimizeRequire() <--- that one would be neat */

    for(req = result->hr_required; req; req = req->req_next) {
      for(rpac = req->req_list; rpac; rpac = rpac->rpac_next) {
	printf(" -i%s", rpac->rpac_name);

	if(rpac->rpac_versions.siz) {
	  printf(",%s", rpac->rpac_versions.arr[0]);
	}
      }
    }

    puts("\n");
  }
}


static void
print_about_text(const char *version, HickEntry *entry, time_t curTime, int f)
{
  HickVer *ver;
  char right[2], left[2];

  strcpy(right, (f ? ")" : ""));
  strcpy(left,  (f ? "(" : ""));

  for(ver = entry->he_versions.vl_head; ver; ver = ver->ver_next) {
    if(!strcasecmp(ver->ver_number, version)) {
      break;
    }
  }

  if(!ver) {
    printf("Can't find the version '%s' in '%s'.\n", version, entry->he_name);
    return;
  }

  if(ver->ver_expire && (ver->ver_expire > curTime)) {
    printf("The version '%s' in '%s' have expired.\n", ver->ver_number,
	   entry->he_name);
    return;
  }

  if(ver->ver_about) {
    if(ver->ver_override) {
      printf("%s %s%s%s\n%s\n", entry->he_name, left, ver->ver_number, right,
	     ver->ver_about);
    } else {
      printf("%s %s%s%s\n%s%s\n", entry->he_name, left, ver->ver_number, right,
	     entry->he_about, ver->ver_about);
    }
  } else {
    printf("%s %s%s%s\n%s\n", entry->he_name, left, ver->ver_number, right,
	   entry->he_about);
  }
}


static int
print_about(const char *name, const char *version, const char *packpath,
	    const char *homepath, const char *hicktype)
{
  int err;
  time_t curTime;
  HickEntry entry;
  HickPackets pack;

  time(&curTime);

  if((err = openPackets(packpath, homepath, hicktype, NULL, &pack))) {
    write_error(err, packpath);
    return 0;
  }

  initHickEntry(&entry);

  while(nextPacket(&pack, &entry, 1)) {
    if(strcasecmp(entry.he_name, name)) {
      continue;
    }

    if(entry.he_expire && (entry.he_expire > curTime)) {
      printf("The packet '%s' have expired.\n", entry.he_name);
      break;
    }

    if(!version[0]) {
      if(entry.he_using.siz) {	/* print the users used version */
	print_about_text(entry.he_using.arr[0], &entry, curTime, 1);
      } else {
	printf("%s\n%s\n", entry.he_name, entry.he_about);
      }
    } else {
      print_about_text(version, &entry, curTime, 0);
    }

    closePackets(&pack);
    return 1;
  }

  printf("There is no packet with the name '%s'.\n", name);

  closePackets(&pack);
  return 0;
}


static int
list_packets(const char *packpath, const char *homepath, const char *hicktype,
	     const char *group, int f)
{
  int err;
  HickVer *ver;
  time_t curTime;
  HickEntry entry;
  HickPackets pack;

  time(&curTime);

  if((err = openPackets(packpath, homepath, hicktype, group, &pack))) {
    write_error(err, packpath);
    return 0;
  }

  initHickEntry(&entry);

  while(nextPacket(&pack, &entry, 0)) {
    if((entry.he_expire && (entry.he_expire > curTime)) ||
       ((f == SHOW_INSTALLED) && !entry.he_using.siz) ||
       ((f == SHOW_NOT_INSTALLED) && entry.he_using.siz)) {
      continue;
    }

    printf("%-15s ", entry.he_name);

    for(ver = entry.he_versions.vl_head; ver; ver = ver->ver_next) {
      if(f == SHOW_INSTALLED) {
	if(!hsh_array_find(ver->ver_number, &entry.he_using)) {
	  continue;
	}

	if(ver->ver_expire && (ver->ver_expire > curTime)) {
	  printf("[%s] ", ver->ver_number);
	}

	printf("%s ", ver->ver_number);
	continue;
      }

      if(ver->ver_expire && (ver->ver_expire > curTime)) {
	continue;
      }

      printf("%s ", ver->ver_number);
    }
    
    putchar('\n');
  }

  closePackets(&pack);
  return 1;
}


static int
list_groups(const char *packpath, const char *hicktype)
{
  int err;
  HickEntry entry;
  HickPackets pack;
  ArrayInfo array;
  unsigned int n;

  if((err = openPackets(packpath, NULL, hicktype, NULL, &pack))) {
    write_error(err, packpath);
    return 0;
  }

  initHickEntry(&entry);
  hsh_array_init(&array);

  while(nextPacketHeader(&pack, &entry)) {
    for(n = 0; n < entry.he_groups.siz; n++) {
      hsh_array_add(entry.he_groups.arr[n], 1, &array);
    }
  }

  if(array.siz) {
    printf("Availible groups:\n");

    for(n = 0; n < array.siz; n++) {
      printf("  %s\n", array.arr[n]);
    }
  } else {
    printf("No availible groups.\n");
  }

  closePackets(&pack);
  return 1;
}


static int
split_name(char *name, char *version, char *args)
{
  unsigned int i, len;

  len = strlen(args);

  for(i = 0; args[i] && (args[i] != ','); i++);

  if(i > PNAME_LEN) {
    printf("error: packet name to long.\n");
    return 0;
  }

  if((args[i] == ',') && ((len - i - 1) > PVER_LEN)) {
    printf("error: version length is to long.\n");
    return 0;
  }

  strncpy(name, args, i);
  name[i] = (char)0;

  if(args[i] == ',') {
    strncpy(version, args + i + 1, PVER_LEN);
    version[PVER_LEN] = (char)0;
  } else {
    version[0] = (char)0;
  }

  return 1;
}


static void
helicopter(HickEvent *event, void *arg)
{
  static int heli = 0;

  putchar('\b');
  putchar((heli == 0 ? '/' : (heli == 1 ? '-' : (heli == 2 ? '\\' : '|'))));
  fflush(stdout);

  if(++heli == 4) {
    heli = 0;
  }
}


static void
write_error(int err, const char *s)
{
  printf("error: ");

  switch(err) {
    case HICKERR_HOMEDIR: {
      printf("can't enter the home directory.\n");
      break;
    }
    case HICKERR_HICKDIR: {
      printf("can't enter the hick directory.\n");
      break;
    }
    case HICKERR_TYPEDIR: {
      printf("can't enter hick type directory.\n");
      break;
    }
    case HICKERR_TYPECONF: {
      printf("can't enter hick type configure directory.\n");
      break;
    }
    case HICKERR_INSTPACK: {
      printf("can't enter installed packets directory.\n");
      break;
    }
    case HICKERR_SCRIPTDIR: {
      printf("can't enter inline shell script directory.");
      break;
    }
    case HICKERR_PACKDIR: {
      printf("can't open packet directory '%s' for reading.\n", s);
      break;
    }
    case HICKERR_LOGFILE: {
      printf("can't open the runtime log for writting.\n");
      break;
    }
    case HICKERR_ENVFILE: {
      printf("can't open the environment file for writting.\n");
      break;
    }
    case HICKERR_INFOFILE: {
      printf("can't open the info file for writting.\n");
      break;
    }
    case HICKERR_MENUFILE: {
      printf("can't open the menu file for writting.\n");
      break;
    }
    case HICKERR_XRESFILE: {
      printf("can't open the xresources file for writting.\n");
      break;
    }
    case HICKERR_INSTFILE: {
      printf("can't open the installed file for writting\n");
      break;
    }
    case HICKERR_CALLFILE: {
      printf("can't open the call script file for writting / appending.\n");
      break;
    }
    case HICKERR_RENAME: {
      printf("can't rename the .installed file to installed.\n");
      break;
    }
    case HICKERR_OUTOFMEMORY: {
      printf("out of memory!\n");
      break;
    }
  }
}


static struct _options {
  const char *opt;
  const char *desc;
} options[] = {
  /* information */

  "h", "This help text",
  "v", "Version information",
  "a <packet> [,<version>]", "View about message for packet",

  /* settings */

  "H <hicktype>", "Use this hick type instead of the HICKTYPE",
  "G <group>", "Use this group for packet listings",
  "F <filename>", "Use this filename (insted of stdin or stdout)",
  "U <user>", "ROOT ONLY: Install as this user",

  /* update packet information */

  "i <packet> [,<version>]", "Install packet. No version, using default",
  "r <packet> [,<version>]","Remove packet. No version, remove whole packet",
  "u", "Update installed packets",
  "c", "Generate call script", 
  "s", "Be silent, no output",
  "P", "Proceed without doing rescue (if needed)",
  "p", "Proceed with the rescue (if needed)",
  "R", "Remove all installed packets",
  "V", "View runlog file",

  /* view packet information */

  "t", "List availible hick types",
  "l", "List all packets",
  "g", "List availible groups",
  "d", "List all installed packets",
  "n", "List packets which isn't installed",

  /* Check information */

  "C", "Check for changes in installed packets",
  "N", "Check for new packets and versions",

  /* save or load from file */

  "S", "Save installed packets (no filename set, stdout)",
  "L", "Load packets to install (no filename set, stdin)",
  NULL, NULL
};

static void
usage(const char *name)
{
  int i;

  printf("Usage: %s <option> [ <option> ... ]\n\n", name);

  for(i = 0; options[i].opt; i++) {
    printf("  -%-24s  %s.\n", options[i].opt, options[i].desc);
  }

  putchar('\n');
}
