/* HickPrj 1.0.1 -- Login project chooser for hick.
** Copyright (C) 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtkmain.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtktext.h>
#include <gtk/gtkclist.h>
#include <gtk/gtkhbbox.h>
#include <gtk/gtktable.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkcheckbutton.h>
#include <gtk/gtkprogressbar.h>
#include <gtk/gtkscrolledwindow.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "hickpc.h"
#include "hicksh.h"
#include "chooser.h"
#include "fileenable.h"

#define HICKPRJ_DISABLE_FILE ".hickprj_disable"

typedef struct {
  GtkWidget *window;
  GtkWidget *plist;		/* clist over projects */
  GtkWidget *text;		/* Description over project */
  GtkWidget *tbar;		/* timeout bar */

  GtkWidget *bOK;		/* OK button */
  GtkWidget *bDef;		/* Default select button */

  Projects *prjs;
  time_t maxTime;
  time_t curTime;
  guint timerID;
} PChooser;


static PChooser pchoose;

static void display_info(GtkWidget *text, const char *infofile);


static void 
cb_quit(GtkWidget *w, gpointer data)
{
  gtk_main_quit();
}


static gint
cb_quit_signal(gpointer date)
{
  printf("%s\n", pchoose.prjs->current->name);
  return 0;
}


static gint
cb_timeout(gpointer data)
{
  gfloat r;

  if(pchoose.curTime == pchoose.maxTime) {
    gtk_main_quit();
    return 0;
  }

  pchoose.curTime++;

  r = (gfloat)pchoose.curTime / (gfloat)pchoose.maxTime;
  gtk_progress_bar_update(GTK_PROGRESS_BAR(pchoose.tbar), r);
  return 1;
}


static void 
cb_disable(GtkWidget *w, gpointer data)
{
  char fname[FNAME_LEN+1];

  hsh_create_str(fname, FNAME_LEN, (const char *)data, "/.hick/", 
		 HICKPRJ_DISABLE_FILE, 0);
  
  if(GTK_TOGGLE_BUTTON(w)->active) {
    file_enable(fname);
  } else {
    file_disable(fname);
  }
}


static gint
cb_stop_timer(GtkWidget *w, gpointer data)
{
  if(pchoose.maxTime) {
    gtk_timeout_remove(pchoose.timerID);
  }
}


static void
cb_select_row(GtkWidget *w, gint row, gint column, GdkEventButton *event,
	      gpointer data)
{
  pchoose.prjs->current = parr_get(&pchoose.prjs->projects, Project, row);

  if(event && (event->type == GDK_2BUTTON_PRESS)) {
    gtk_main_quit();
    return;
  }

  display_info(pchoose.text, pchoose.prjs->current->infofile);
}


static void
cb_set_default(GtkWidget *w, gpointer data)
{
  pchoose.prjs->def = pchoose.prjs->current;
  project_save_default(pchoose.prjs);
}


GtkWidget *
chooser_new(Projects *prjs, time_t timeout, const char *header, 
	    const char *homepath)
{
  int n, row;
  gchar *titles[1];
  char fname[FNAME_LEN+1];
  GtkWidget *w, *table, *vbox, *hbox;

  gtk_quit_add(0, cb_quit_signal, NULL);

  pchoose.prjs = prjs;

  if(timeout) {
    pchoose.maxTime = timeout * 100;
    pchoose.curTime = 0;
    pchoose.timerID = gtk_timeout_add(10, cb_timeout, 0);
  }

  pchoose.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_usize(pchoose.window, 500, 230);
  gtk_window_set_policy(GTK_WINDOW(pchoose.window), TRUE, TRUE, FALSE);

  hsh_create_str(fname, FNAME_LEN, PROGRAM, " ", VERSION_STRING, 0);
  gtk_window_set_title(GTK_WINDOW(pchoose.window), fname);
  gtk_window_set_position(GTK_WINDOW(pchoose.window), GTK_WIN_POS_CENTER);

  gtk_signal_connect(GTK_OBJECT(pchoose.window), "destroy",
                     GTK_SIGNAL_FUNC(gtk_widget_destroyed), &pchoose.window);
  gtk_signal_connect(GTK_OBJECT(pchoose.window), "delete_event",
                     GTK_SIGNAL_FUNC(cb_quit), &pchoose);

  vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_container_add(GTK_CONTAINER(pchoose.window), vbox);

  w = gtk_label_new(header);
  gtk_box_pack_start(GTK_BOX(vbox), w, FALSE, TRUE, 0);

  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

  w = gtk_scrolled_window_new(NULL, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), w, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(w),
                                 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  pchoose.plist = gtk_clist_new_with_titles(1, 0);
  gtk_container_add(GTK_CONTAINER(w), pchoose.plist);
  gtk_clist_set_selection_mode(GTK_CLIST(pchoose.plist), GTK_SELECTION_BROWSE);
  gtk_clist_column_titles_passive(GTK_CLIST(pchoose.plist));
  gtk_clist_set_column_auto_resize(GTK_CLIST(pchoose.plist), 0, TRUE);

  table = gtk_table_new(1, 2, FALSE);
  gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);

  pchoose.text = gtk_text_new(NULL, NULL);
  gtk_table_attach(GTK_TABLE(table), pchoose.text, 0, 1, 0, 1,
                   (GtkAttachOptions)(GTK_FILL | GTK_EXPAND | GTK_SHRINK),
                   (GtkAttachOptions)(GTK_FILL | GTK_EXPAND | GTK_SHRINK),
                   0, 0);

  gtk_text_set_word_wrap(GTK_TEXT(pchoose.text), TRUE);

  w = gtk_vscrollbar_new(GTK_TEXT(pchoose.text)->vadj);
  gtk_table_attach(GTK_TABLE(table), w, 1, 2, 0, 1,
                   GTK_FILL, 
                   (GtkAttachOptions)(GTK_EXPAND | GTK_FILL | GTK_SHRINK), 
                   0, 0);

  if(timeout) {
    pchoose.tbar = gtk_progress_bar_new();
    gtk_widget_set_usize(pchoose.tbar, -1, 10);
    gtk_box_pack_start(GTK_BOX(vbox), pchoose.tbar, FALSE, TRUE, 0);
    gtk_progress_set_format_string(GTK_PROGRESS(pchoose.tbar), "");
    gtk_progress_set_show_text(GTK_PROGRESS(pchoose.tbar), TRUE);
  }

  table = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, TRUE, 0);
  w = gtk_check_button_new_with_label("Don't show me this again.");

  hsh_create_str(fname, FNAME_LEN, homepath, "/.hick/", 
		 HICKPRJ_DISABLE_FILE,0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), file_status(fname));
  gtk_box_pack_start(GTK_BOX(table), w, FALSE, TRUE, 0);

  gtk_signal_connect(GTK_OBJECT(w), "pressed",
		     GTK_SIGNAL_FUNC(cb_stop_timer), 0);
  gtk_signal_connect(GTK_OBJECT(w), "toggled",
		     GTK_SIGNAL_FUNC(cb_disable), (gpointer)homepath);

  hbox = gtk_hbutton_box_new();
  gtk_box_pack_end(GTK_BOX(table), hbox, FALSE, TRUE, 0);
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(hbox), 10);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(hbox), GTK_BUTTONBOX_END);

  pchoose.bOK = gtk_button_new_with_label("OK");
  gtk_box_pack_start(GTK_BOX(hbox), pchoose.bOK, FALSE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS(pchoose.bOK, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(pchoose.bOK);

  gtk_signal_connect(GTK_OBJECT(pchoose.bOK), "pressed",
		     GTK_SIGNAL_FUNC(cb_stop_timer), 0);
  gtk_signal_connect(GTK_OBJECT(pchoose.bOK), "clicked",
		     GTK_SIGNAL_FUNC(cb_quit), 0);

  pchoose.bDef = gtk_button_new_with_label("Set Default");
  gtk_box_pack_start(GTK_BOX(hbox), pchoose.bDef, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS(pchoose.bDef, GTK_CAN_DEFAULT);

  gtk_signal_connect(GTK_OBJECT(pchoose.bDef), "pressed",
		     GTK_SIGNAL_FUNC(cb_stop_timer), 0);
  gtk_signal_connect(GTK_OBJECT(pchoose.bDef), "clicked",
		     GTK_SIGNAL_FUNC(cb_set_default), 0);

  for(row = n = 0; n < prjs->projects.siz; n++) {
    if(parr_get(&prjs->projects, Project, n) == prjs->def) {
      row = n;
    }

    titles[0] = parr_get(&prjs->projects, Project, n)->name;
    gtk_clist_append(GTK_CLIST(pchoose.plist), titles);
  }

  if(!prjs->def) {
    prjs->def = prjs->current;
  }

  gtk_clist_select_row(GTK_CLIST(pchoose.plist), row, 0);
  display_info(pchoose.text, prjs->current->infofile);

  gtk_signal_connect(GTK_OBJECT(pchoose.plist), "select_row",
		     GTK_SIGNAL_FUNC(cb_stop_timer), 0);
  gtk_signal_connect(GTK_OBJECT(pchoose.plist), "select_row",
		     GTK_SIGNAL_FUNC(cb_select_row), 0);

  return pchoose.window;
}


static void
display_info(GtkWidget *text, const char *infofile)
{
  int fd;
  char *buf;
  size_t len;
  struct stat sbuf;

  gtk_text_freeze(GTK_TEXT(text));
  gtk_text_backward_delete(GTK_TEXT(text),gtk_text_get_length(GTK_TEXT(text)));

  if(infofile) {
    if((fd = open(infofile, O_RDONLY, 0))) {
      if(fstat(fd, &sbuf) != -1) {
	buf = MallocArr(char, sbuf.st_size + 1);

	if((len = read(fd, buf, (size_t)sbuf.st_size))) {
	  gtk_text_insert(GTK_TEXT(text), 0, 0, 0, buf, len);
	}

	FreeMem(buf);
      }

      close(fd);
    }
  }

  gtk_text_thaw(GTK_TEXT(text));
}
