/* HickPrj 1.0.1 -- Login project chooser for hick.
** Copyright (C) 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "hicksh.h"

#include "memd.h"
#include "project.h"

static void project_new(Projects *prjs, char *name, char *specfile, 
			char *infofile, char *homedir);
static char *parse_get_next(const char *str, size_t len, char *buf, 
			    size_t *pos);
static void parse_projects(Projects *prjs, const char *prjsstr);


void
projects_init(Projects *prjs, const char *prjsstr, char *user, char *infofile,
	      char *homedir)
{
  size_t n;
  FILE *file;
  char fname[FNAME_LEN + 1];
  const char *hicktype = getenv("HICKTYPE");

  prjs->prjfile[0] = (char)0;
  parr_init(&prjs->projects);

  if(user && homedir) {
    project_new(prjs, user, 0, infofile, homedir);
  }

  if(prjsstr) {
    parse_projects(prjs, prjsstr);
  }

  prjs->def = 0;
  prjs->current = (prjs->projects.siz ? parr_get(&prjs->projects, Project, 0) :
		   0);

  if(homedir && hicktype) {	/* ${HOME}/.hick/project-solaris */
    hsh_create_str(prjs->prjfile, FNAME_LEN, homedir, "/.hick/project-", 
		   hicktype, 0);

    if(!(file = fopen(prjs->prjfile, "r"))) {
      prjs->def = parr_get(&prjs->projects, Project, 0);
      return;
    }

    fgets(fname, FNAME_LEN, file);
    fname[strlen(fname) - 1] = (char)0;

    for(n = 0; n < prjs->projects.siz; n++) {
      if(!strcmp(parr_get(&prjs->projects, Project, n)->name, fname)) {
	prjs->def = prjs->current = parr_get(&prjs->projects, Project, n);
	break;
      }
    }

    fclose(file);
  }
}


int
project_save_default(Projects *prjs)
{
  FILE *file;

  if(prjs->prjfile[0] && (file = fopen(prjs->prjfile, "w"))) {
    fprintf(file, "%s\n", prjs->def->name);
    fclose(file);
  }
}


static void
project_new(Projects *prjs, char *name, char *prjdir, char *infofile, 
	    char *homedir)
{
  Project *prj;
  struct stat sbuf;
  char specfile[FNAME_LEN + 1];
  unsigned int can_edit = 0, can_access;

  if(prjdir) {
    hsh_create_str(specfile, FNAME_LEN, prjdir, "/", HICK_SPECFILE, 0);

    can_access = 0;

    if(hsh_stat(specfile, &sbuf, NULL) == IOW_OK) {
      if(!S_ISREG(sbuf.st_mode)) {
        return;			/* no a regular file, ignore project */
      }

      can_access = (access(specfile, R_OK) == 0 ? 1 : 0);
      can_edit = (access(specfile, W_OK) == 0 ? 1 : 0);
    }

    /* if we can't access the file cause if dosen't exist, check and see if
    ** we can write in the directory where the file is. If so setup in edit
    ** mode.
    */

    if(!can_access) {
      return;			/* no access, ignore project */
    }
  }

  prj = MallocMem(Project);
  prj->name    = name;
  prj->homedir = homedir;
  prj->prjdir  = prjdir;

  if(prjdir && !infofile) {
    prj->infofile = MallocArr(char, strlen(prjdir) + strlen(HICK_INFOFILE) +2);
    hsh_create_str(prj->infofile, FNAME_LEN, prjdir, "/", HICK_INFOFILE, 0);
  } else {
    prj->infofile = infofile;
  }

  if(prjdir) {
    prj->specfile = MallocArr(char, strlen(prjdir) + strlen(HICK_SPECFILE) +2);
    hsh_create_str(prj->specfile, FNAME_LEN, prjdir, "/", HICK_SPECFILE, 0);
  } else {
    prj->specfile = 0;
  }

  parr_add(&prjs->projects, prj);
}


static void
parse_projects(Projects *prjs, const char *prjsstr)
{
  size_t n, len;
  char *name, *prjdir, *home, *buf;

  n = 0;
  len = strlen(prjsstr);
  buf = MallocArr(char, len + 1);

  do {
    name = prjdir = home = NULL;

    if((name = parse_get_next(prjsstr, len, buf, &n))) {
      if((prjdir = parse_get_next(prjsstr, len, buf, &n))) {
	if((home = parse_get_next(prjsstr, len, buf, &n))) {
	  project_new(prjs, name, prjdir, 0, home);
	} else {
	  FreeMem(name);
	  FreeMem(prjdir);
	}
      } else {
	FreeMem(name);
      }
    }

  } while(name && prjdir && home);

  FreeMem(buf);
}


static char *
parse_get_next(const char *str, size_t len, char *buf, size_t *pos)
{
  size_t n;
  char *p;

  buf[0] = (char)0;

  for(n = 0; (*pos < len) && str[*pos] && (str[*pos] != ':'); (*pos)++) {
    if((str[*pos] == '\\') && (str[++(*pos)] != ':')) {
      buf[n++] = '\\';
    }

    buf[n++] = str[*pos];
  }

  (*pos)++;

  if(buf[0]) {
    buf[n] = (char)0;
    p = MallocArr(char, n + 1);
    strcpy(p, buf);
    return p;
  }
  
  return NULL;
}
