/* HickUP 1.1.1 -- Graphical interface to hick.
** Copyright (C) 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <time.h>

#include "hicksh.h"
#include "common.h"

#include "gui.h"
#include "avail.h"
#include "availrep.h"

#define avail_step_packet(n) \
    ((n) == INSTALL_DEFAULT ? INSTALL_NOT : INSTALL_DEFAULT)

#define avail_step_version(n) \
  ((n) == INSTALL_NOT ? INSTALL_VERSION :  \
   ((n) == INSTALL_VERSION ? INSTALL_FORCE : INSTALL_NOT))

#define avail_step_pack_packet(n) \
    ((n) == INSTALL_DEFAULT ? INSTALL_NOT : \
     ((n) == INSTALL_VERSION ? INSTALL_NOT : INSTALL_DEFAULT))

#define avail_step_pack_version(n) \
    ((n) == INSTALL_VERSION ? INSTALL_NOT : INSTALL_VERSION)


static int is_installed(const char *version, userEntryList *ulist);

static AvailPacket *apack_new(Packet *packet, NodeList *list);
static AvailVersion *aver_new(AvailPacket *apack, Version *ver, 
			      unsigned int install);

static int avail_packet_set(AvailPacket *apack, AvailVersion *aver, int flag, 
			    NodeList *avail);
static int avail_edit_set(AvailPacket *apack, AvailVersion *aver, int flag,
			  NodeList *avail);

static void avail_set_pixmaps(AvailPacket *apack, int isedit);

static int test_conflict(PArr *conflicts, unsigned int install, int isedit,
			 NodeList *avail);
static int contest(unsigned int cinst, unsigned int inst, int isedit);

static void force_conflict_test(NodeList *avail, Project *prj);
static void test_force_conflict(AvailPacket *apack, AvailVersion *aver, 
				PArr *cons, NodeList *avail, Project *prj);
static void justify_force_conflict(AvailPacket *apack, AvailVersion *aver, 
				   AvailPacket *packet, AvailVersion *version,
				   Project *prj);


void
avail_from_specvars(NodeList *packets, SpecVars *svars, const char *instfile,
		    NodeList *avail, PArr *types, Project *prj)
{
  size_t n, i;
  Node *pnode;
  Version *ver;
  time_t curtime;
  SpecType *stype;
  int candef, isdef;
  AvailPacket *apack;
  AvailVersion *aver;
  userEntry *ue, *uetmp;
  SpecVar *svar, *force;
  char fname[FNAME_LEN];
  userEntryList installed, ulist;

  time(&curtime);
  avail_free(avail);
  availrep_clear(prj);

  initUserEntryList(&ulist);
  initUserEntryList(&installed);

  for(n = 0; n < types->siz; n++) {
    if(svars) {
      if(!(stype = spec_find_type(parr_get(types, char, n), svars))) {
	continue;
      }
    } else {
      stype = NULL;
    }

    hsh_create_str(fname, FNAME_LEN, instfile, "/", HICK_DIR, "/.", 
		   parr_get(types, char, n), "/", INSTALLED_FILE, 0);

    _entryLoad(fname, &installed);

    for(pnode = packets->head; pnode; pnode = pnode->next) {
      if(((Packet *)pnode)->type != parr_get(types, char, n)) {
	continue;
      }

      _freeUserEntryList(&ulist);

      for(ue = installed.ul_head; ue; ue = uetmp) {
	uetmp = ue->ue_next;

	if(!strcasecmp(ue->ue_name, pnode->name)) {
	  _entryMove(ue, &installed, &ulist);
	}
      }

      if(svars) {
	if(!(svar = force = stype_find(pnode->name, SPEC_FORCE, stype))) {
	  svar = stype_find(pnode->name, SPEC_ENABLE, stype);
	}

	isdef = availrep_check(prj, ((Packet *)pnode), curtime, svar, 1, 
			       &ulist);

	if(!((Packet *)pnode)->info.forced && !svar) {
	  continue;		/* no access */
	}
      } else {
	svar = force = NULL;
	isdef = availrep_check(prj, ((Packet *)pnode), curtime, NULL, 0, 
			       &ulist);
      }

      apack = apack_new(((Packet *)pnode), avail);

      for(candef = CAN_NONE, i = 0; i < ((Packet *)pnode)->vers.siz; i++) {
	ver = parr_get(&((Packet *)pnode)->vers, Version, i);

	if(svars && !ver->info.forced && !specvar_is_version(svar,ver->name)) {
	  if(ver->isdef) {
	    candef = CAN_SOME;
	  }
	  continue;
	}

	if(ver->isdef && (candef != CAN_SOME)) {
	  candef = CAN_ALL;
	}

	if(((Packet *)pnode)->info.forced || force) {
	  if(ver->info.forced || force) {
	    aver = aver_new(apack, ver, INSTALL_FORCE);

	    if(is_installed(ver->name, &ulist) != INSTALL_VERSION) {
	      aver->info.rinstall = INSTALL_NOT;
	    }
	  } else {
	    aver_new(apack, ver, INSTALL_NOT);
	  }
	} else {
	  if(isdef) {
	    if(ver->isdef) {
	      aver_new(apack, ver, INSTALL_DEFAULT);
	    } else {
	      aver_new(apack, ver, INSTALL_NOT);
	    }

	    aver = parr_get(&apack->vers, AvailVersion, apack->vers.siz -1);
	  } else {
	    aver_new(apack, ver, is_installed(ver->name, &ulist));
	  }
	}
      }

      if(apack->info.install == INSTALL_FORCE) {
	continue;		/* don't go there if it's forced. */
      }

      if((apack->can_def = candef) == CAN_SOME) {
	for(isdef = FALSE, i = 0; i < apack->vers.siz; i++) {
	  aver = parr_get(&apack->vers, AvailVersion, i);

	  if(aver->version->isdef) {
	    isdef = TRUE;
	  }

	  if(aver->info.install == INSTALL_DEFAULT) {
	    aver->info.install = INSTALL_VERSION;
	  }
	}

	if(!isdef) {
	  apack->can_def = CAN_NONE;
	}

	if(apack->info.install != INSTALL_NOT) {
	  apack->info.install = INSTALL_VERSION;
	}
      }
    }

    availrep_removed(prj, parr_get(types, char, n), &installed);
  }

  force_conflict_test(avail, prj);
  availrep_show_report(prj);
}


void
avail_from_specvars_edit(NodeList *packets, SpecVars *svars, PArr *types, 
			 NodeList *avail)
{
  Node *node;
  size_t n, i;
  Version *ver;
  SpecType *stype;
  AvailPacket *apack;
  SpecVar *svar, *force;

  avail_free(avail);

  for(n = 0; n < types->siz; n++) {
    stype = spec_find_type(parr_get(types, char, n), svars);

    for(node = packets->head; node; node = node->next) {
      if(((Packet *)node)->type != parr_get(types, char, n)) {
	continue;
      }

      apack = apack_new(((Packet *)node), avail);

      if(!((Packet *)node)->info.forced) {
	if(!(svar = force = stype_find(node->name, SPEC_FORCE, stype))) {
	  svar = stype_find(node->name, SPEC_ENABLE, stype);
	}

	if(svar && !svar->versions.siz) {
	  apack->info.install = apack->info.rinstall = INSTALL_DEFAULT;
	}
      } else {
	force = 0;
	svar = NULL;
      }

      for(i = 0; i < ((Packet *)node)->vers.siz; i++) {
	ver = parr_get(&((Packet *)node)->vers, Version, i);

	if(ver->isdef) {
	  apack->can_def = 1;
	}

	if(((Packet *)node)->info.forced) {
	  if(ver->info.forced) {
	    aver_new(apack, ver, INSTALL_FORCE);
	  } else {
	    aver_new(apack, ver, INSTALL_NOT);
	  }

	  continue;
	}
	 
	if(!svar || !svar->versions.siz ||
	   !parr_find(&svar->versions, ver->name, parr_strcasecmp)) {

	  aver_new(apack, ver, INSTALL_NOT);
	  continue;
	}

	if(force) {
	  aver_new(apack, ver, INSTALL_FORCE);
	} else {
	  aver_new(apack, ver, INSTALL_VERSION);
	}
      }
    }
  }
}


void
avail_to_specvars(NodeList *avail, PArr *types, SpecVars *svars, int setreal)
{
  Node *node;
  char *type;
  PArr parr;
  size_t n, i, f;
  AvailVersion *aver;
  AvailPacket *apack;

  for(n = 0; n < types->siz; n++) {
    type = parr_get(types, char, n);

    for(node = avail->head; node; node = node->next) {
      apack = ((AvailPacket *)node);

      if((apack->packet->type != type) || (apack->packet->info.forced)) {
	continue;
      }

      switch(apack->info.install) {
        case INSTALL_DEFAULT: {	/* enable packet */
	  spec_add(type, apack->packet->name, NULL, SPEC_ENABLE, svars);

	  if(setreal) {
	    apack->info.rinstall = apack->info.install;

	    for(i = 0; i < apack->vers.siz; i++) {
	      aver = parr_get(&apack->vers, AvailVersion, i);
	      aver->info.rinstall = aver->info.install;
	    }
	  }
	  break;
	}
        case INSTALL_FORCE:
        case INSTALL_VERSION: {	/* go thru versions, enable versions */
	  parr_init(&parr);

	  f = (apack->info.install ==INSTALL_FORCE ? SPEC_FORCE : SPEC_ENABLE);

	  for(i = 0; i < apack->vers.siz; i++) {
	    aver = parr_get(&apack->vers, AvailVersion, i);

	    if(aver->info.install == apack->info.install) {
	      parr_add(&parr, aver->version->name);
	    }

	    if(setreal) {
	      aver->info.rinstall = aver->info.install;
	    }
	  }

	  spec_add(type, apack->packet->name, &parr, f, svars);
	  FreeMem(parr.arr);
	  break;
	}
      }
    }
  }
}


static int
is_installed(const char *version, userEntryList *ulist)
{
  userEntry *ue;

  for(ue = ulist->ul_head; ue; ue = ue->ue_next) {
    if(!strcasecmp(ue->ue_version, version)) {
      if((ue->ue_flags + 1) == INSTALL_FORCE) {
	return INSTALL_VERSION;
      }

      return (ue->ue_flags + 1);
    }
  }

  return INSTALL_NOT;
}


static AvailPacket *
apack_new(Packet *packet, NodeList *list)
{
  AvailPacket *apack;

  apack = MallocMem(AvailPacket);
  apack->packet        = packet;
  apack->info.install  = INSTALL_NOT;
  apack->info.rinstall = INSTALL_NOT;
  apack->can_def       = FALSE;
  apack->node.name     = packet->name;
  packet->node.flag    = NT_PACKET;

  parr_init(&apack->vers);
  parr_init(&apack->refs);

  hsh_node_prepend(&apack->node, list);
  return apack;
}


static AvailVersion *
aver_new(AvailPacket *apack, Version *ver, unsigned int install)
{
  AvailVersion *aver;

  aver = MallocMem(AvailVersion);
  aver->packet         = apack;
  aver->version        = ver;
  aver->info.rinstall  = install;
  aver->info.install   = install;

  if((install != INSTALL_NOT) && (apack->info.install == INSTALL_NOT)) {
    apack->info.install = apack->info.rinstall = install;
  }

  parr_add(&apack->vers, aver);
  return aver;
}


void
avail_free(NodeList *avail)
{
  size_t n;
  Node *node;

  for(node = avail->head; node; node = avail->head) {
    avail->head = node->next;

    for(n = 0; n < ((AvailPacket *)node)->vers.siz; n++) {
      FreeMem(parr_get(&((AvailPacket *)node)->vers, AvailVersion, n));
    }

    if(((AvailPacket *)node)->vers.arr) {
      FreeMem(((AvailPacket *)node)->vers.arr);
    }

    if(((AvailPacket *)node)->refs.arr) {
      FreeMem(((AvailPacket *)node)->refs.arr);
    }

    FreeMem(node);
  }

  hsh_nodelist_init(avail);
}


int
avail_set(AvailPacket *apack, AvailVersion *aver, int flag, int edit, 
	  NodeList *avail)
{
  if(!apack->packet->info.forced) {
    if(edit) {
      return avail_edit_set(apack, aver, flag, avail);
    }

    if(apack->info.install != INSTALL_FORCE) {
      return avail_packet_set(apack, aver, flag, avail);
    }
  }

  return FALSE;
}


static int
avail_packet_set(AvailPacket *apack, AvailVersion *aver, int flag, 
		 NodeList *avail)
{
  size_t n;
  AvailVersion *atmp;
  unsigned int install, have_conflict;

  if(!aver) {
    install = (flag == -1 ? avail_step_pack_packet(apack->info.install) :flag);

    if(install == INSTALL_DEFAULT) {
      for(have_conflict = 0, n = 0; n < apack->vers.siz; n++) {
	atmp = parr_get(&apack->vers, AvailVersion, n);
	atmp->info.install = INSTALL_NOT;

	if(atmp->version->isdef &&
	   !avail_can_enable(apack, atmp, INSTALL_VERSION, FALSE, avail)) {
	  have_conflict = 1;
	  break;
	}
      }

      if(apack->can_def == CAN_NONE) {
	if(apack->vers.siz) {
	  apack->info.install = INSTALL_VERSION;
	  parr_get(&apack->vers, AvailVersion, 0)->info.install = 
	    INSTALL_VERSION;
	} else {
	  apack->info.install = INSTALL_NOT;
	}
      } else {
	for(n = 0; n < apack->vers.siz; n++) {
	  atmp = parr_get(&apack->vers, AvailVersion, n);

	  if(atmp->version->isdef && 
	     (!have_conflict || (have_conflict && 
	      avail_can_enable(apack, atmp, INSTALL_VERSION, FALSE,avail)))) {

	    if(have_conflict || (apack->can_def == CAN_SOME)) {
	      atmp->info.install = INSTALL_VERSION;
	      apack->info.install = INSTALL_VERSION;
	    } else {
	      atmp->info.install = INSTALL_DEFAULT;
	      apack->info.install = INSTALL_DEFAULT;
	    }
	  }
	}
      }

      avail_set_pixmaps(apack, FALSE);
      return TRUE;
    }

    apack->info.install = INSTALL_NOT;

    for(n = 0; n < apack->vers.siz; n++) {
      parr_get(&apack->vers, AvailVersion, n)->info.install = INSTALL_NOT;
    }

    avail_set_pixmaps(apack, FALSE);  
    return TRUE;
  }


  install = (flag == -1 ? avail_step_pack_version(aver->info.install) : flag);

  if((have_conflict = !avail_can_enable(apack, aver, install, FALSE, avail))) {
    if(flag != -1) {
      return FALSE;
    }

    install = INSTALL_NOT;
  }

  if((apack->info.install != INSTALL_DEFAULT) && 
     (aver->info.install == install)) {
    return TRUE;
  }

  switch((aver->info.install = install)) {
    case INSTALL_NOT: {
      install = apack->info.install;
      apack->info.install = INSTALL_NOT;

      for(n = 0; n < apack->vers.siz; n++) {
	atmp = parr_get(&apack->vers, AvailVersion, n);

	if(atmp == aver) {
	  atmp->info.install = INSTALL_NOT;
	} else {
	  if(atmp->info.install == INSTALL_DEFAULT) {
	    apack->info.install = INSTALL_VERSION;
	    atmp->info.install = INSTALL_VERSION;
	  }
	}
      }
      break;
    }
    case INSTALL_VERSION: {
      apack->info.install = INSTALL_VERSION;

      if(apack->packet->multiple) {
	aver->info.install = INSTALL_VERSION;
      }

      for(n = 0; n < apack->vers.siz; n++) {
	atmp = parr_get(&apack->vers, AvailVersion, n);

	if(atmp == aver) {
	  atmp->info.install = INSTALL_VERSION;
	} else {
	  if(!apack->packet->multiple ||
	     (apack->packet->multiple && 
	      (atmp->info.install == INSTALL_DEFAULT))) {
	    atmp->info.install = INSTALL_NOT;
	  }
	}
      }
    }
  }

  avail_set_pixmaps(apack, FALSE);
  return TRUE;
}


static int
avail_edit_set(AvailPacket *apack, AvailVersion *aver, int flag,
	       NodeList *avail)
{
  size_t n;
  AvailVersion *atmp;
  unsigned int install, have_conflict;

  if(!aver) {
    install = (flag == -1 ? avail_step_packet(apack->info.install) : flag);

    if(install == INSTALL_DEFAULT) { /* enable whole packet? */
      for(have_conflict = 0, n = 0; n < apack->vers.siz; n++) {
	atmp = parr_get(&apack->vers, AvailVersion, n);

	if((have_conflict = !avail_can_enable(apack, atmp, INSTALL_VERSION, 
					      TRUE, avail))) {
	  break;
	}
      }

      if(!have_conflict || 
	 (have_conflict && (apack->info.install != INSTALL_VERSION))) {
	if(have_conflict) {
	  apack->info.install = INSTALL_NOT;

	  for(n = 0; n < apack->vers.siz; n++) {
	    atmp = parr_get(&apack->vers, AvailVersion, n);

	    if(avail_can_enable(apack, atmp, INSTALL_VERSION, TRUE, avail)) {
	      apack->info.install = install = INSTALL_VERSION;
	    } else {
	      install = INSTALL_NOT;
	    }

	    atmp->info.install = install;
	  }
	} else {
	  apack->info.install = INSTALL_DEFAULT;

	  for(n = 0; n < apack->vers.siz; n++) {
	    parr_get(&apack->vers, AvailVersion, n)->info.install =INSTALL_NOT;
	  }
	}

	avail_set_pixmaps(apack, TRUE);
	return TRUE;
      }

      if(flag != -1) {
	return FALSE;
      }
    }

    apack->info.install = INSTALL_NOT;

    for(n = 0; n < apack->vers.siz; n++) {
      parr_get(&apack->vers, AvailVersion, n)->info.install = INSTALL_NOT;
    }

    avail_set_pixmaps(apack, TRUE);
    return TRUE;
  }

  /* emable / disable / force a version? */

  install = (flag == -1 ? avail_step_version(aver->info.install) : flag);

  if((have_conflict = !avail_can_enable(apack, aver, install, TRUE, avail))) {
    if(flag != -1) {
      return FALSE;
    }

    install = INSTALL_NOT;
  }

  if((apack->info.install != INSTALL_DEFAULT) && 
     (aver->info.install == install)) {
    return TRUE;
  }

  switch((aver->info.install = install)) {
    case INSTALL_NOT: {
      install = apack->info.install;
      apack->info.install = INSTALL_NOT;

      for(n = 0; n < apack->vers.siz; n++) {
	atmp = parr_get(&apack->vers, AvailVersion, n);

	if(install != INSTALL_DEFAULT) {
	  if(atmp == aver) {
	    atmp->info.install = INSTALL_NOT;
	  } else {
	    if(atmp->info.install != INSTALL_NOT) {
	      apack->info.install = atmp->info.install;
	    }
	  }
	} else {
	  atmp->info.install = (atmp != aver ? INSTALL_VERSION : INSTALL_NOT);

	  if(atmp->info.install == INSTALL_VERSION) {
	    apack->info.install = INSTALL_VERSION;
	  }
	}
      }
      break;
    }
    case INSTALL_FORCE: {
      apack->info.install = INSTALL_FORCE;

      for(n = 0; n < apack->vers.siz; n++) {
	atmp = parr_get(&apack->vers, AvailVersion, n);

	if((atmp->info.install == INSTALL_VERSION) ||
	   (!apack->packet->multiple && (atmp != aver))) {
	  atmp->info.install = INSTALL_NOT;
	}
      }
      break;
    }
    case INSTALL_VERSION: {
      apack->info.install = INSTALL_VERSION;

      for(n = 0; n < apack->vers.siz; n++) {
	atmp = parr_get(&apack->vers, AvailVersion, n);

	if(atmp->info.install == INSTALL_FORCE) {
	  atmp->info.install = INSTALL_NOT;
	}
      }
    }
  }

  avail_set_pixmaps(apack, TRUE);
  return TRUE;
}


static void
avail_set_pixmaps(AvailPacket *apack, int isedit)
{
  size_t n, exp;
  Node *node;
  GPixmap *pixi;
  NTList *ntlist;
  RefPacket *pref;
  RefData *rdat;

  for(n = 0; n < apack->refs.siz; n++) {
    ntlist = parr_get(&apack->refs, NTList, n);

    if(ntlist->ntree.node.flag == NT_EMPTY) { /* RefData */
      rdat = ((RefData *)ntlist);

      if(rdat->packet && rdat->hview && rdat->hview->pixfunc) {
	pixi = hview_pixmap_get(rdat->hview, rdat->packet, rdat->version, 
				NULL);

	hview_pixmap_display(rdat->hview, &ntlist->ntree.node, pixi);
      }
    } else {			/* RefPacket */
      pref = ((RefPacket *)ntlist);

      if(!pref->packet || !pref->hview || !pref->hview->pixfunc) {
	continue;
      }

      exp = GTK_CTREE_ROW(pref->ntlist.ntree.ctnode)->expanded;

      if(!exp || (isedit && exp)) {
	pixi = hview_pixmap_get(pref->hview, pref->packet, NULL, (void *)exp);

	hview_pixmap_display(pref->hview, &pref->ntlist.ntree.node, pixi);
      }

      for(node = pref->ntlist.nlist.head; node; node = node->next) {
	pixi = hview_pixmap_get(pref->hview, pref->packet, 
				((RefVersion *)node)->version, NULL);

	hview_pixmap_display(pref->hview, node, pixi);
      }
    }
  }
}


AvailPacket *
avail_find_packet(const char *packet, NodeList *list)
{
  Node *node;

  for(node = list->head; node && (node->name != packet); node = node->next);

  return (AvailPacket *)node;
}


AvailVersion *
avail_find_version(const char *version, AvailPacket *apack)
{
  size_t n;

  for(n = 0; n < apack->vers.siz; n++) {
    if(parr_get(&apack->vers, AvailVersion, n)->version->name == version) {
      return parr_get(&apack->vers, AvailVersion, n);
    }
  }

  return NULL;
}


int
avail_can_enable(AvailPacket *packet, AvailVersion *version, 
		 unsigned int install, int isedit, NodeList *avail)
{
  if(!test_conflict(&packet->packet->info.conflicts, install, isedit, avail)) {
    return 0;
  }

  if(version && 
     !test_conflict(&version->version->info.conflicts, install, isedit, 
		    avail)) {
    return 0;
  }

  return 1;
}


static int
test_conflict(PArr *conflicts, unsigned int install, int isedit, 
	      NodeList *avail)
{
  size_t n, i;
  Version *ver;
  Conflict *con;
  AvailPacket *apack;
  AvailVersion *aver;

  for(n = 0; n < conflicts->siz; n++) {
    con = parr_get(conflicts, Conflict, n);

    if((apack = avail_find_packet(con->packet->name, avail))) {
      if(((apack->info.install == INSTALL_DEFAULT) && 
	  (install == INSTALL_FORCE)) || 
	  (!con->vers.siz && contest(apack->info.install, install, isedit))) {
	return 0;			/* got a conflict */
      }

      for(i = 0; i < con->vers.siz; i++) {
	ver = parr_get(&con->vers, Version, i);

	if((aver = avail_find_version(ver->name, apack)) && 
	   contest(aver->info.install, install, isedit)) {
	  return 0;		
	}
      }
    }
  }

  return 1;
}


static int
contest(unsigned int cinst, unsigned int inst, int isedit)
{
  if(isedit) {
    return (((cinst == INSTALL_FORCE) && ((inst == INSTALL_VERSION) ||
					  (inst == INSTALL_FORCE))) ||
	    ((cinst == INSTALL_VERSION) && (inst == INSTALL_FORCE)));
  } else {
    return ((inst == INSTALL_VERSION) && cinst);
  }
}


static void
force_conflict_test(NodeList *avail, Project *prj)
{
  size_t n;
  Node *node;
  AvailPacket *apack;
  AvailVersion *aver;

  for(node = avail->head; node; node = node->next) {
    apack = (AvailPacket *)node;

    if(!apack->info.install && !apack->packet->info.forced) {
      continue;
    }

    test_force_conflict(apack, NULL, &apack->packet->info.conflicts, avail, 
			prj);

    for(n = 0; n < apack->vers.siz; n++) {
      aver = parr_get(&apack->vers, AvailVersion, n);

      if(!aver->info.install && !aver->version->info.forced) {
	continue;
      }

      test_force_conflict(apack, aver, &aver->version->info.conflicts, avail,
			  prj);
    }
  }
}


static void
test_force_conflict(AvailPacket *apack, AvailVersion *aver, PArr *cons,
		    NodeList *avail, Project *prj)
{
  size_t n, i;
  Conflict *con;
  AvailPacket *packet;
  AvailVersion *version;

  for(n = 0; n < cons->siz; n++) {
    con = parr_get(cons, Conflict, n);

    if(!(packet = avail_find_packet(con->packet->name, avail)) ||
       (!packet->info.install && !packet->packet->info.forced)) {
      continue;			/* not installed or forced */
    }

    if(!con->vers.siz) {
      justify_force_conflict(apack, aver, packet, NULL, prj);
      continue;
    }

    for(i = 0; i < con->vers.siz; i++) {
      version = avail_find_version(parr_get(&con->vers, Version, i)->name, 
				   packet);

      if(version && (version->info.install || version->version->info.forced)) {
	justify_force_conflict(apack, aver, packet, version, prj);
      }
    }
  }
}


static void
justify_force_conflict(AvailPacket *apack, AvailVersion *aver, 
		       AvailPacket *packet, AvailVersion *version, 
		       Project *prj)
{
  size_t n;
  PVInfo *tpv, *opv;
  APVInfo *tapv, *oapv;

  if(apack->packet->info.forced && packet->packet->info.forced) {
    availrep_conflict(prj, apack->packet->type, apack, aver, packet, version);
    return;
  }

  tpv = (aver ? &aver->version->info : &apack->packet->info);
  tapv = (aver ? &aver->info : &apack->info);

  opv = (version ? &version->version->info : &packet->packet->info);
  oapv = (version ? &version->info : &packet->info);

  if((tpv->forced && oapv->install) || 
     ((tapv->install == INSTALL_FORCE) && 
      ((oapv->install == INSTALL_DEFAULT) ||
       (oapv->install == INSTALL_VERSION)))) {

    availrep_conflict(prj, apack->packet->type, packet, version, apack, aver);
    packet->info.install = INSTALL_NOT;

    for(n = 0; n < packet->vers.siz; n++) {
      parr_get(&packet->vers, AvailVersion, n)->info.install = INSTALL_NOT;
    }
  } else {
    availrep_conflict(prj, apack->packet->type, apack, aver, packet, version);
    apack->info.install = INSTALL_NOT;

    for(n = 0; n < apack->vers.siz; n++) {
      parr_get(&apack->vers, AvailVersion, n)->info.install = INSTALL_NOT;
    }
  }
}
