/* HickUP 1.1.1 -- Graphical interface to hick.
** Copyright (C) 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "load.h"

#include "entry.h"
#include "version.h"
#include "getpacket.h"

static void info_create(PVInfo *info, HickEnvList *conflicts, 
			HickEnvList *require, NodeList *emptyList, 
			NodeList *packList, const char *name, 
			const char *version, const char *type);

static void require_create(PVInfo *info, HickEnvList *require, 
			   NodeList *emptyList, NodeList *packList, 
			   const char *name, const char *version,
			   const char *type);
static Require *require_new_lookup(HickRequirePacket *rpac, 
				   NodeList *emptyList, NodeList *packList, 
				   const char *name, const char *version, 
				   const char *type);

static void conflicts_create(PVInfo *info, HickEnvList *conflicts, 
			     NodeList *emptyList, NodeList *packList, 
			     const char *name, const char *version,
			     const char *type);

static Packet *packet_lookup(const char *name, const char *type, 
			     NodeList *emptyList, NodeList *packList);

static void versions_lookup(ArrayInfo *versions, Packet *packet, PArr *vers, 
			    const char *name, const char *version, 
			    const char *type, int iscon);

static void force_set(HickPacket *hpac, Packet *packet);


int
projects_load_packets(Projects *prjs, const char *path)
{
  size_t n, i;
  Version *ver;
  Packet *packet;
  time_t curTime;
  PArr tmpvers;
  HickVer *hver;
  HickPacket hpac;
  HickPackets pack;
  HickVerList denied;
  userEntryList ulist;
  size_t no_packets = 0, no_versions = 0;

  loaderr_clear();

  time(&curTime);
  ulist.ul_head = NULL;
  ulist.ul_tail = NULL;
  denied.vl_head = NULL;
  denied.vl_tail = NULL;

  for(n = 0; n < prjs->types.siz; n++) {
    if(openPackets(path, NULL, parr_get(&prjs->types, char, n), NULL, &pack)) {
      continue;
    }

    initHickPacket(&hpac);

    while(nextPacketFull(&pack, &hpac)) {
      if((packet = packet_find(hpac.pac_name, parr_get(&prjs->types, char, n),
		     &prjs->packets))) {
	loaderr_add("packet: %s %s: already exists.\n", 
		    parr_get(&prjs->types, char, n), hpac.pac_name);
	continue;
      }

      _freeUserEntryList(&ulist);

      i = _versionPermissions(&hpac.pac_versions, hpac.pac_allow,
			      hpac.pac_deny, &denied);
      freeHickVerList(&denied, 1);

      if(!i) {
	continue;		/* no permissions to the packet */
      }

      parr_init(&tmpvers);

      if((packet = packet_find(hpac.pac_name, parr_get(&prjs->types, char, n),
			       &prjs->failure))) {
	parr_move(&tmpvers, &packet->vers);
      }

      packet = packet_new(&hpac, hpac.pac_name, 
			  parr_get(&prjs->types, char, n), 
			  &prjs->groups, &prjs->failure, &prjs->packets);
      hpac.pac_about = NULL;
      no_packets++;

      info_create(&packet->info, &hpac.pac_conflict, &hpac.pac_require,
		  &prjs->failure, &prjs->packets, packet->name, NULL,
		  parr_get(&prjs->types, char, n));

      for(hver = hpac.pac_versions.vl_head; hver; hver = hver->ver_next) {
	if((i = parr_find(&tmpvers, hver->ver_number, _versioncmp))) {
	  parr_add(&packet->vers, parr_remove(&tmpvers, i - 1));
	}

	ver = version_new(hver, hver->ver_number, packet);

	info_create(&ver->info, &HVER_CONFLICT(hver), &HVER_REQUIRE(hver),
		    &prjs->failure, &prjs->packets, packet->name, ver->name, 
		    parr_get(&prjs->types, char, n));
	hver->ver_about = NULL;
	no_versions++;
      }

      parr_move(&packet->vers, &tmpvers);

      force_set(&hpac, packet);
    }

    closePackets(&pack);
  }

  _freeUserEntryList(&ulist);

  loaderr_check_information(&prjs->packets);
  loaderr_show();
  printf("%d packets / %d versions loaded.\n", no_packets, no_versions);
  return 1;
}


static void
info_create(PVInfo *info, HickEnvList *conflicts, HickEnvList *require,
	    NodeList *emptyList, NodeList *packList, const char *name,
	    const char *version, const char *type)
{
  conflicts_create(info, conflicts, emptyList, packList, name, version, type);

  require_create(info, require, emptyList, packList, name, version, type);
}


static void
require_create(PVInfo *info, HickEnvList *require, NodeList *emptyList,
	       NodeList *packList, const char *name, const char *version,
	       const char *type)
{
  HickEnv *env;
  HickRequirePacket *rpac;
  Require *req, *lastor = NULL, *tail = NULL;

  if(!require->el_head) {
    return;
  }

  for(env = require->el_head; env; env = env->env_next) {
    req = require_new_lookup(env->u.env_rpacs, emptyList, packList, name,
			     version, type);

    if(!info->require) {
      info->require = tail = req;
    } else {
      tail->next = req;
      tail = req;
    }

    for(rpac = env->u.env_rpacs->rpac_or; rpac; rpac = rpac->rpac_or) {
      req = require_new_lookup(rpac, emptyList, packList, name, version, type);

      if(!tail->ror) {
	lastor = tail->ror = req;
      } else {
	lastor->next = req;
	lastor = req;
      }
    }
  }
}


static Require *
require_new_lookup(HickRequirePacket *rpac, NodeList *emptyList,
		   NodeList *packList, const char *name, const char *version,
		   const char *type)
{
  Require *req;

  req = MallocMem(Require);
  req->next = NULL;
  req->ror = NULL;
  parr_init(&req->info.vers);

  req->info.packet = packet_lookup(rpac->rpac_name, type, emptyList, packList);

  if(rpac->rpac_versions.siz) {
    versions_lookup(&rpac->rpac_versions, req->info.packet,
		    &req->info.vers, name, version, type, FALSE);
  }

  return req;
}


static void
conflicts_create(PVInfo *info, HickEnvList *conflicts, NodeList *emptyList,
		 NodeList *packList, const char *name, const char *version,
		 const char *type)
{
  size_t n;
  HickEnv *env;
  Conflict *con;

  if(!conflicts->el_head) {
    return;
  }

  for(n = 0, env = conflicts->el_head; env; env = env->env_next, n++) {
    con = MallocMem(Conflict);
    parr_init(&con->vers);
    parr_add(&info->conflicts, con);

    con->packet = packet_lookup(env->env_name, type, emptyList, packList);

    if(env->u.env_versions->siz) {
      versions_lookup(env->u.env_versions, con->packet, &con->vers, name,
		      version, type, TRUE);
    }
  }
}


static Packet *
packet_lookup(const char *name, const char *type, NodeList *emptyList,
	      NodeList *packList)
{
  Packet *packet;

  if((packet = packet_find(name, type, packList))) {
    return packet;
  }

  if((packet = packet_find(name, type, emptyList))) {
    return packet;
  }

  return packet_new(NULL, name, type, NULL, NULL, emptyList);
}


static void
versions_lookup(ArrayInfo *versions, Packet *packet, PArr *vers,
		const char *name, const char *version, const char *type,
		int iscon)
{
  size_t n;
  Version *ver;

  if(packet->node.flag == NT_PACKET) {
    for(n = 0; n < versions->siz; n++) {
      if((ver = version_find(&packet->vers, versions->arr[n]))) {
	parr_add(vers, ver);
      } else {
	loaderr_add("%s: %s %s %s: missing version %s in %s.\n",
		    (iscon ? "conflict" : "require"), type, name,
		    (version ? version : "(whole)"), versions->arr[n],
		    packet->name);
      }
    }
  } else {
    for(n = 0; n < versions->siz; n++) {
      if((ver = version_find(&packet->vers, versions->arr[n]))) {
	parr_add(vers, ver);
      } else {
	ver = version_new(NULL, versions->arr[n], packet);
	parr_add(vers, ver);
      }
    }
  }
}


static void
force_set(HickPacket *hpac, Packet *packet)
{
  size_t n;
  Version *ver;
  DoubleArray forced;

  _checkDirection(hpac->pac_force, PERM_FORCE, &forced);

  for(n = 0; n < forced.arr1.siz; n++) {
    if((ver = version_find(&packet->vers, forced.arr1.arr[n]))) {
      packet->info.forced = TRUE;
      ver->info.forced = TRUE;
    } else {
      loaderr_add("force: %s %s: missing version %s.\n", packet->type,
		  packet->name, forced.arr1.arr[n]);
    }
  }

  hsh_array_remove(&forced.arr1);
}
