/* HickUP 1.1.1 -- Graphical interface to hick.
** Copyright (C) 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <pwd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#include <gtk/gtkrc.h>
#include <gtk/gtkmain.h>

#include "hickmk.h"
#include "hickup.h"
#include "common.h"		/* from hicklib 1.3 :: for HICK_DIR define */

#include "gui.h"
#include "pref.h"
#include "dialog.h"
#include "buttons.h"

GUI_Main _gmain;

extern int check_rescue(GUI_Main *gmain); /* from rescue.c */
extern void run_rescue(GUI_Main *gmain);

static int usage(const char *filename);

extern char *optarg;


int
main(int argc, char **argv)
{
  GtkWidget *window;
  struct passwd *pwd;
  char *p = 0, *desc = 0, *update_prj = 0;
  int n, update_exit = 0, rescue_exit = 0, need_rescue;

  gtk_init(&argc, &argv);
  gtk_rc_init();

  while((n = getopt(argc, argv, "hvruU:fp:d:")) != EOF) {
    switch(n) {
      case 'h': {		/* show usage */
	return usage(argv[0]);
      }
      case 'v': {
	printf("%s %s\n", PROGRAM, VERSION_STRING);
	return 0;
      }
      case 'r': {		/* check need of rescue then exit */
	rescue_exit = 1;
	break;
      }
      case 'p': {		/* project spec argument */
	p = optarg;
	break;
      }
      case 'u': {		/* update user project */
	update_exit = 1;
	break;
      }
      case 'U': {		/* update specified project */
	update_prj = optarg;
	update_exit = 2;
	break;
      }
      case 'f': {		/* Update all projects */
	update_exit = 3;
	break;
      }
      case 'd': {		/* user Description */
	desc = optarg;
	break;
      }
      default: {
	return usage(argv[0]);
      }
    }
  }

  if(!(pwd = getpwuid(geteuid()))) {
    dialog_err_exit_new(100, "No Passwd Entry",
			"Failed to obtain your user entry (UID #%d).", 
			geteuid());
  }

  if(!pwd->pw_name || !*pwd->pw_name) {
    dialog_err_exit_new(102, "No User Name",
			"User name is missing in password entry (UID #%d).",
			geteuid());
  }

  if(!pwd->pw_dir || !*pwd->pw_dir) {
    dialog_err_exit_new(101, "No Home Specified",
			"No home directory specified for user '%s'.",
			pwd->pw_name);
  }

  _pref.pfile = PREF_FILE;
  _pref.hickdir = HICK_DIR;
  hsh_create_str(_pref.path, FNAME_LEN, pwd->pw_dir, "/.hick", 0);

  if(pref_load()) {
    pref_init(TRUE, HICKMK_DEFAULT_WIDTH);
  }

  gui_main_init(&_gmain, p, pwd->pw_name, pwd->pw_dir, desc);

  if(!_gmain.projects.current) {
    dialog_err_exit_new(112, "No Project", "No projects have been loaded.");
  }

  if((n = projects_get_types(&_gmain.projects, PACKET_PATH))) {
    if(n == -1) {
      dialog_err_exit_new(110, "Open Packet Path",
			    "Failed to open packet path '%s'.", PACKET_PATH);
    } else {
      dialog_err_exit_new(110, "No Hick Types",
			    "No avalible hick types in '%s'.", PACKET_PATH);
    }
  }

  if(!(need_rescue = check_rescue(&_gmain)) && rescue_exit) {
    return 0;
  }

  if((_gmain.projects.hicktype = getenv("HICKTYPE"))) {
    _gmain.projects.hicktype = parr_find_value(&_gmain.projects.types,
					       char, _gmain.projects.hicktype,
					       parr_strcasecmp);
  }

  if(!projects_load_packets(&_gmain.projects, PACKET_PATH)) {
    dialog_err_exit_new(111, "Load Packets",
			"Failed to load packets from '%s'.", PACKET_PATH);
  }

  hsh_create_str(_gmain.menufile, FNAME_LEN, pwd->pw_dir, "/.hick/menubar", 0);

  window = gui_main_create(&_gmain);

  if(update_exit) {
    update_and_exit(&_gmain.projects, update_exit, update_prj);
  }

  if(need_rescue) {
    run_rescue(&_gmain);

    if(!rescue_exit) {
      cb_reload_projects(0, 0);
    }
  }

  if(rescue_exit) {
    return 0;
  }

  gtk_widget_show(window);
  gtk_main();

#ifdef MEMDEBUG
  /*  printf("Clearing up\n");

  projects_clear_all(&_gmain.projects);
  pref_clear();

  ShowMem;
  g_mem_profile(); */
#endif

  return 0;
}

static int
usage(const char *filename)
{
  printf("Usage: %s [ OPTIONS ]\n\n"
	 "\t-h\t\tPrint this help test.\n"
	 "\t-v\t\tShow version information.\n"
	 "\t-r\t\tCheck if rescue is needed, and exit.\n"
	 "\t-p STRING\tProject specifications\n"
	 "\t-u\t\tUpdate user packets\n"
	 "\t-U PROJECT\tUpdate specified project packets\n"
	 "\t-f\t\tUpdate all projects, including user\n"
	 "\t-d\t\tFile containing user description text\n\n", filename);
  return 0;
}
