/* HickUP 1.1.1 -- Graphical interface to hick.
** Copyright (C) 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtknotebook.h>

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "gui.h"
#include "filter.h"
#include "search.h"

extern void viewlog_update_hicktypes(void); /* viewlog.c */

static void parse_projects(Projects *prjs, const char *prjsstr);
static char *parse_get_next(const char *str, size_t len, char *buf, 
			    size_t *pos);
static void project_new(Projects *prjs, char *name, char *prjdir, 
			char *infofile, char *homedir);


void
projects_init(Projects *prjs, const char *prjsstr, char *user, char *homedir,
	      char *userdesc)
{
  parr_init(&prjs->types);
  parr_init(&prjs->groups);
  hsh_nodelist_init(&prjs->packets);
  hsh_nodelist_init(&prjs->failure);
  parr_init(&prjs->names);
  parr_init(&prjs->projects);

  if(user && homedir) {		/* add user default project? */
    project_new(prjs, user, 0, userdesc, homedir);
  }

  if(prjsstr) {			/* add projects in string */
    parse_projects(prjs, prjsstr);
  }

  prjs->current=(Project *)(prjs->projects.siz ? prjs->projects.arr[0] : NULL);
}


int
projects_get_types(Projects *prjs, const char *path)
{
  size_t n;
  ArrayInfo arri;

  parr_clear(&prjs->types);

  if(getHickTypes(path, &arri)) {
    return -1;
  }

  if(!arri.siz) {
    return -2;
  }

  for(n = 0; n < arri.siz; n++) {
    parr_add(&prjs->types, arri.arr[n]);
  }

  FreeMem(arri.arr);
  return 0;
}


void
project_init_packets(Project *prj, NodeList *packets, PArr *types)
{
  SpecVars svars;

  prj->packets.has_init = 1;

  if(prj->specfile) {
    spec_load(prj->specfile, &svars);

    avail_from_specvars(packets, &svars, prj->homedir, &prj->packets.avail,
			types, prj);
    spec_free(&svars);
  } else {
    avail_from_specvars(packets, NULL, prj->homedir, &prj->packets.avail, 
			types, prj);
  }

  filter_get(&prj->packets.filter, &prj->packets.hview);

  hview_freeze(&prj->packets.hview);
  refs_display_tree(hview_display_get(&prj->packets.hview), 
		    &prj->packets.hview, NULL, FALSE);
  hview_thaw(&prj->packets.hview);

  if(_gmain.projects.hicktype) {
    gui_find_expand(NULL, NULL, _gmain.projects.hicktype, prj, 0);
  }
}


void
project_init_edit(Project *prj, NodeList *packets, PArr *types)
{
  SpecVars svars;

  prj->edit->has_init = 1;

  spec_load(prj->specfile, &svars);
  avail_from_specvars_edit(packets, &svars, types, &prj->edit->avail);
  spec_free(&svars);

  filter_get(&prj->edit->filter, &prj->edit->hview);

  hview_freeze(&prj->edit->hview);
  refs_display_tree(hview_display_get(&prj->edit->hview), &prj->edit->hview, 
		    NULL, FALSE);
  hview_thaw(&prj->edit->hview);

  if(_gmain.projects.hicktype) {
    gui_find_expand(NULL, NULL, _gmain.projects.hicktype, prj, 1);
  }
}


static void
project_new(Projects *prjs, char *name, char *prjdir, char *infofile,
	    char *homedir)
{
  Project *prj;
  struct stat sbuf;
  char specfile[FNAME_LEN + 1];
  unsigned int can_edit = 0, can_access;

  if(prjdir) {
    hsh_create_str(specfile, FNAME_LEN, prjdir, "/", HICK_SPECFILE, 0);
    can_access = 0;

    if(hsh_stat(specfile, &sbuf, NULL) == IOW_OK) {
      if(!S_ISREG(sbuf.st_mode)) {
	dialog_err_return_new(0, "Open Project",
			      "Can't open '%s', is not a regular file.", name);
	return;
      }

      can_access = (access(specfile, R_OK) == 0 ? 1 : 0);
      can_edit = (access(specfile, W_OK) == 0 ? 1 : 0);
    } else {
      return;
    }

    if(!can_access) {
      dialog_err_return_new(0, "Open Project",
			    "Failed to get read access to project '%s'.", 
			    name);
      return;
    }
  }

  parr_add(&prjs->names, name);

  prj = MallocMem(Project);
  prj->name     = name;
  prj->prjdir   = prjdir;
  prj->homedir  = homedir;
  prj->can_edit = can_edit;
  prj->edit     = NULL;

  if(prjdir && !infofile) {
    prj->infofile = MallocArr(char, strlen(prjdir) + strlen(HICK_INFOFILE) +2);
    hsh_create_str(prj->infofile, FNAME_LEN, prjdir, "/", HICK_INFOFILE, 0);
  } else {
    prj->infofile = infofile;
  }

  if(prjdir) {
    prj->specfile = MallocArr(char, strlen(prjdir) + strlen(HICK_SPECFILE) +2);
    hsh_create_str(prj->specfile, FNAME_LEN, prjdir, "/", HICK_SPECFILE, 0);
  } else {
    prj->specfile = 0;
  }

  prj->packets.hview.view = NULL;
  prj->packets.has_init = 0;

  parr_init(&prj->removed);
  dialog_init(&prj->report);
  dialog_init(&prj->update);

  parr_add(&prjs->projects, prj);
}


static void
parse_projects(Projects *prjs, const char *prjsstr)
{
  size_t n, len;
  char *name, *prjdir, *home, *buf;

  n = 0;
  len = strlen(prjsstr);
  buf = MallocArr(char, len + 1);

  do {
    name = prjdir = home = NULL;

    if((name = parse_get_next(prjsstr, len, buf, &n))) {
      if((prjdir = parse_get_next(prjsstr, len, buf, &n))) {
        if((home = parse_get_next(prjsstr, len, buf, &n))) {
          project_new(prjs, name, prjdir, 0, home);
        } else {
          FreeMem(name);
          FreeMem(prjdir);
        }
      } else {
        FreeMem(name);
      }
    }
  } while(name && prjdir && home);

  FreeMem(buf);
}


static char *
parse_get_next(const char *str, size_t len, char *buf, size_t *pos)
{
  size_t n;
  char *p;

  buf[0] = (char)0;

  for(n = 0; (*pos < len) && str[*pos] && (str[*pos] != ':'); (*pos)++) {
    if((str[*pos] == '\\') && (str[++(*pos)] != ':')) {
      buf[n++] = '\\';
    }

    buf[n++] = str[*pos];
  }

  (*pos)++;

  if(buf[0]) {
    buf[n] = (char)0;
    p = MallocArr(char, n + 1);
    strcpy(p, buf);
    return p;
  }
  
  return NULL;
}


void
project_removed(Project *prj, char *packet, char *version, char *type)
{
  Removed *remd;

  if((remd = MallocMem(Removed))) {
    strcpy(remd->packet, packet);
    strcpy(remd->version, version);
    remd->type = type;

    parr_add(&prj->removed, remd);
  }
}


void
cb_reload_projects(GtkWidget *w, gpointer data)
{
  int r;
  size_t n;
  Project *prj;

  prj = _gmain.projects.current;
  n = gtk_notebook_get_current_page(GTK_NOTEBOOK(prj->notebook));
  search_reset(n);

  projects_clear_all(&_gmain.projects);

  if((r = projects_get_types(&_gmain.projects, PACKET_PATH))) {
    if(r == -1) {
      dialog_err_return_new(0, "Open Packet Path",
			    "Failed to open packet path '%s'.", PACKET_PATH);
    } else {
      dialog_err_return_new(0, "No Hick Types",
			    "No avalible hick types in '%s'.", PACKET_PATH);
    }

    return;
  }

  if((_gmain.projects.hicktype = getenv("HICKTYPE"))) {
    _gmain.projects.hicktype = parr_find_value(&_gmain.projects.types,
                                               char, _gmain.projects.hicktype,
                                               parr_strcasecmp);
  }

  viewlog_update_hicktypes();

  if(!projects_load_packets(&_gmain.projects, PACKET_PATH)) {
    dialog_err_return_new(0, "Load Packets",
                        "Failed to load packets from '%s'.", PACKET_PATH);
    return;
  }

  if(n == 0) {
    project_init_packets(_gmain.projects.current, &_gmain.projects.packets, 
			 &_gmain.projects.types);
  } else {
    if(n == 2) {
      project_init_edit(_gmain.projects.current, &_gmain.projects.packets, 
			   &_gmain.projects.types);
    }
  }

  search_reset(n);
}


void
projects_clear_all(Projects *prjs)
{  
  size_t n;
  Project *prj;

  for(n = 0; n < prjs->projects.siz; n++) { /* clear projects */
    prj = parr_get(&prjs->projects, Project, n);

    parr_clear(&prj->removed);

    if(prj->packets.has_init) {
      select_clear(&prj->packets);
      reqchk_close(&prj->packets.reqchk);

      hview_display_destroy(&prj->packets.hview);
      avail_free(&prj->packets.avail);

      prj->packets.has_init = 0;
    }

    if(prj->edit) {
      select_clear(prj->edit);
      reqchk_close(&prj->edit->reqchk);

      hview_display_destroy(&prj->edit->hview);
      avail_free(&prj->edit->avail);

      prj->edit->has_init = 0;
    }
  }

  packets_clear(&prjs->packets);
  packets_clear(&prjs->failure);

  prjs->hicktype = NULL;
  parr_clear(&prjs->groups);
  parr_clear(&prjs->types);
}
