/* HickUP 1.1.1 -- Graphical interface to hick.
** Copyright (C) 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <time.h>

#include <gtk/gtktext.h>
#include <gtk/gtkpixmap.h>

#include "gui.h"
#include "wwrap.h"

#define CONMAX_LEN  (PNAME_LEN + PVER_LEN + 5)

static void select_clear_lists(TabPackets *ptab);

static void select_display(TabPackets *ptab);
static void select_display_header(TabPackets *ptab, Packet *packet, 
				  Version *version, char **s1, char **s2);
static void select_display_about(TabPackets *ptab, char *s1, char *s2);

static void select_display_conflict(TabPackets *ptab, Packet *packet,
				    Version *version);
static void conflict_create(TabPackets *ptab, PVInfo *info);

static void select_conflict_create(const char *name, HView *hview, 
				   AvailPacket *apack, AvailVersion *aver);

static void select_display_require(TabPackets *ptab, Packet *packet,
				   Version *version);


void
cb_select(GtkCTree *ctree, GtkCTreeNode *ctnode, gint column, 
	  TabPackets *ptab)
{
  NodeTree *ntree;

  if(ptab->ctnode != ctnode) {
    ptab->ctnode = ctnode;

    ntree = (NodeTree *)gtk_ctree_node_get_row_data(ctree, ctnode);

    if(!hview_select_set(&ptab->hview, ntree) || !ctnode) {
      select_clear(ptab);
    } else {
      select_display(ptab);
    }
  }
}


void
select_clear(TabPackets *ptab)
{
  hview_select_clear(&ptab->hview);
  ptab->ctnode = NULL;
  ptab->old_s1 = NULL;
  ptab->old_s2 = NULL;

  gtk_widget_hide(ptab->pixmap);
  gtk_label_set_text(GTK_LABEL(ptab->packet), "");
  gtk_label_set_text(GTK_LABEL(ptab->version), "");
  gtk_label_set_text(GTK_LABEL(ptab->info), "");

  ww_text_clear(ptab->about);
  select_clear_lists(ptab);
}


static void
select_clear_lists(TabPackets *ptab)
{
  reqchk_set(&ptab->require.display, FALSE, TRUE);
  reqchk_set(&ptab->conflict.display, FALSE, FALSE);
  hview_display_destroy(&ptab->require);
  hview_display_destroy(&ptab->conflict);
}


static void
select_display(TabPackets *ptab)
{
  char *s1, *s2;
  Packet *packet;
  Version *version;
  unsigned int flag = ptab->hview.select.ntree->node.flag;

  packet = (ptab->hview.select.packet ? ptab->hview.select.packet->packet : 
	    NULL);

  version = (ptab->hview.select.version ? ptab->hview.select.version->version :
	     NULL);

  if(!packet && !version) {
    s1 = s2 = NULL;
    gtk_label_set_text(GTK_LABEL(ptab->version), "");
    gtk_label_set_text(GTK_LABEL(ptab->packet), 
		       ptab->hview.select.ntree->node.name);
    ww_text_clear(ptab->about);
  }

  if(flag && _gmain.pixmaps.pixmaps[flag - 1]) {
    gtk_pixmap_set(GTK_PIXMAP(ptab->pixmap),
		   _gmain.pixmaps.pixmaps[flag - 1]->close.pix,
		   _gmain.pixmaps.pixmaps[flag - 1]->close.mask);
    gtk_widget_show(ptab->pixmap);
  } else {
    gtk_widget_hide(ptab->pixmap);
  }

  if(packet) {
    select_display_header(ptab, packet, version, &s1, &s2);
  } else {
    gtk_label_set_text(GTK_LABEL(ptab->info), "");
  }

  if((ptab->old_s1 != s1) || (ptab->old_s2 != s2)) {
    select_display_about(ptab, s1, s2);
    ptab->old_s1 = s1;
    ptab->old_s2 = s2;
  }

  if(packet) {
    select_display_conflict(ptab, packet, version);
    select_display_require(ptab, packet, version);
  } else {
    select_clear_lists(ptab);
  }
}


static void
select_display_header(TabPackets *ptab, Packet *packet, Version *version, 
		      char **s1, char **s2)
{
  GString *gs;
  const char *p;
  time_t curtime;
  APVInfo *apvinfo;
  AvailPacket *apack;
  AvailVersion *aver;

  *s1 = packet->info.about;
  gtk_label_set_text(GTK_LABEL(ptab->packet), packet->name);

  if(!version) {
    *s2 = NULL;
    gtk_label_set_text(GTK_LABEL(ptab->version), "");
  }

  if(version) {
    gtk_label_set_text(GTK_LABEL(ptab->version), version->name);
    gtk_widget_show(ptab->version);

    if(version->override) {
      *s1 = version->info.about;
      *s2 = NULL;
    } else {
      *s2 = version->info.about;
    }
  } else {
    gtk_widget_hide(ptab->version);
  }

  gs = NULL;

  if(packet->info.forced || (version && version->info.forced)) {
    gs = g_string_new("(forced) ");
  } else {
    if(!ptab->is_edit) {	/* check project spec. for force */
      if((apack = avail_find_packet(packet->name, &ptab->avail))) {
	if(!version) {
	  apvinfo = &apack->info;
	} else {
	  if((aver = avail_find_version(version->name, apack))) {
	    apvinfo = &apack->info;
	  } else {
	    apvinfo = NULL;
	  }
	}

	if(apvinfo && apvinfo->install == INSTALL_FORCE) {
	  gs = g_string_new("(forced) ");
	}
      }
    }
  }

  if(packet->info.expire || (version && version->info.expire)) {
    time(&curtime);

    if((packet && packet->info.expire <= curtime) || 
       (version && version->info.expire <= curtime)) {
      p = "(expired) ";
    } else {
      p = "(will expire) ";
    }

    gs = (gs ? g_string_append(gs, p) : g_string_new(p));
  }

  if(packet->required && !version) {
    p = "(required)";
    gs = (gs ? g_string_append(gs, p) : g_string_new(p));
  }

  if(version && version->isdef) {
    p = "(default)";
    gs = (gs ? g_string_append(gs, p) : g_string_new(p));
  }

  if(gs) {
    gtk_label_set_text(GTK_LABEL(ptab->info), gs->str);
    g_string_free(gs, TRUE);
  } else {
    gtk_label_set_text(GTK_LABEL(ptab->info), "");
  }
}


static void
select_display_about(TabPackets *ptab, char *s1, char *s2)
{
  guint len;

  gtk_text_freeze(GTK_TEXT(ptab->about));

  if((len = gtk_text_get_length(GTK_TEXT(ptab->about)))) {
    gtk_text_backward_delete(GTK_TEXT(ptab->about), len);
  }

  if(s1) {
    gtk_text_insert(GTK_TEXT(ptab->about), NULL, NULL, NULL, s1, -1);
  }

  if(s2) {
    gtk_text_insert(GTK_TEXT(ptab->about), NULL, NULL, NULL, s2, -1);
  }

  gtk_text_thaw(GTK_TEXT(ptab->about));
}


static void
select_display_conflict(TabPackets *ptab, Packet *packet, Version *version)
{
  hview_freeze(&ptab->conflict);
  reqchk_set(&ptab->conflict.display, FALSE, FALSE);
  hview_display_destroy(&ptab->conflict);

  if(packet && packet->info.conflicts.siz) {
    conflict_create(ptab, &packet->info);
  }

  if(version && version->info.conflicts.siz) {
    conflict_create(ptab, &version->info);
  }

  hview_thaw(&ptab->conflict);
}


static void
conflict_create(TabPackets *ptab, PVInfo *info)
{
  size_t n, i;
  Conflict *con;
  AvailPacket *apack;
  AvailVersion *aver;
  char buf[CONMAX_LEN];

  for(n = 0; n < info->conflicts.siz; n++) {
    con = parr_get(&info->conflicts, Conflict, n);

    if(!con->vers.siz) {
      if((apack = avail_find_packet(con->packet->name, &ptab->avail))) {
	select_conflict_create(con->packet->name, &ptab->conflict, apack, 0);
      }
    } else {
      for(i = 0; i < con->vers.siz; i++) {
	hsh_create_str(buf, CONMAX_LEN, con->packet->name, " ", 
		       parr_get(&con->vers, Version , i)->name, 0);


	if((apack = avail_find_packet(con->packet->name, &ptab->avail))) {
	  if((aver = avail_find_version(parr_get(&con->vers, Version, i)->name,
					apack))) {
	    select_conflict_create(buf, &ptab->conflict, apack, aver);
	  }
	}
      }
    }
  }
}


static void
select_conflict_create(const char *name, HView *hview, AvailPacket *apack, 
		       AvailVersion *aver)
{
  GPixmap *pixi;
  RefData *rdat;
  gchar *text[2];

  rdat = refs_data_new(NULL, apack, aver, &hview->display);
  rdat->hview = hview;
  parr_add(&apack->refs, &rdat->ntlist);

  text[0] = (char)0;
  text[1] = (char *)name;

  rdat->row = gtk_clist_append(hview_clist_get(hview), text);
  gtk_clist_set_row_data(hview_clist_get(hview), rdat->row, &rdat->ntlist);

  pixi = hview_pixmap_get(hview, apack, aver, NULL);
  gtk_clist_set_pixmap(hview_clist_get(hview), rdat->row, 0,
		       pixi->pix, pixi->mask);
}


static void
select_display_require(TabPackets *ptab, Packet *packet, Version *version)
{
  Require *req;

  hview_freeze(&ptab->require);
  reqchk_set(&ptab->require.display, FALSE, TRUE);
  hview_display_destroy(&ptab->require);

  if(packet && packet->info.require) {
    for(req = packet->info.require; req; req = req->next) {
      reqchk_create_require(&ptab->require, req, &ptab->avail, 
			    &ptab->require.display, FALSE);
    }
  }

  if(version && version->info.require) {
    for(req = version->info.require; req; req = req->next) {
      reqchk_create_require(&ptab->require, req, &ptab->avail, 
			    &ptab->require.display, FALSE);
    }
  }

  refs_display_tree(&ptab->require.display, &ptab->require, NULL, TRUE);
  reqchk_set(&ptab->require.display, TRUE, TRUE);
  hview_thaw(&ptab->require);
}


GPixmap *
select_pixmap_conflict(AvailPacket *apack, AvailVersion *aver, void *data)
{
  if(aver) {
    if(aver->info.install) {
      return &_gmain.pixmaps.no;
    }
  } else {
    if(apack && apack->info.install) {
      return &_gmain.pixmaps.no;
    }
  }

  return &_gmain.pixmaps.yes;
}


GPixmap *
select_pixmap_require(AvailPacket *apack, AvailVersion *aver, void *data)
{
  if(aver) {
    if(aver->info.install) {
      return &_gmain.pixmaps.yes;
    }
  } else {
    if(apack && apack->info.install) {
      return &_gmain.pixmaps.yes;
    }
  }

  return &_gmain.pixmaps.no;
}
