/* HickUP 1.1.1 -- Graphical interface to hick.
** Copyright (C) 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtkmain.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtksignal.h>
#include <gtk/gtknotebook.h>
#include <gtk/gtkprogressbar.h>

#include "hick.h"
#include "hickmk.h"
#include "hicksh.h"

#include "gui.h"
#include "pref.h"
#include "report.h"
#include "update.h"

#define TAB_REQUIRE   0
#define TAB_CONFLICT  1
#define TAB_FAILURE   2
#define TAB_FORCED    3

#define IS_DOT_OR_DOTDOT(e) (((e)->d_name[0] == '.') && \
    (((e)->d_name[1] == (char)0) || \
     (((e)->d_name[1] == '.') && ((e)->d_name[2] == (char)0))))

static size_t count_packets_type(NodeList *list, const char *type);
static size_t count_dirs(const char *dirname);

static void hcb_update(HickEvent *event, Progress *pbar);
static void parse_gtk_events(void);
static void set_label(GtkWidget *label, const char *msg, const char *type);
static void update_report(Project *prj, HickResult *result, const char *type);
static void update_report_new(Project *prj);
static void update_clear_report(Project *prj);
static void hcb_update(HickEvent *event, Progress *pbar);


static const char *fail_reasons[13] = {
  "Packet was not found.",
  "Version was not found.",
  "No default version in packet.",
  "Couldn't install, force override.",

  "Couldn't find packet.",
  "Couldn't find version.",
  "Missing default version.",
  "Removed by force.",

  "Can't remove force installed packet.",
  "Failed to remove packet.",
  "Failed to remove version.",

  "Packet have expired.",
  "Version have expired."
};

static int _update_running = FALSE;


int
update_is_running(void)
{
  return _update_running;
}


void
update_set_running(int state)
{
  _update_running = state;
}


int
update_show_report(Project *prj)
{
  int page;
  Report *r;

  if(prj->update.window) {
    r = ((Report *)prj->update.arg);

    report_show_pixmaps(r);
    page = report_display(r, ((1 << TAB_REQUIRE) |
			      (1 << TAB_CONFLICT) |
			      (1 << TAB_FAILURE) |
			      (_pref.report_forced << TAB_FORCED)));

    if(page != -1) {
      gtk_notebook_set_page(GTK_NOTEBOOK(r->notebook), page);
      gtk_widget_show(prj->update.window);
      return 1;
    }
  }

  return 0;
}


void
update_show(Project *prj)
{
  Report *r;

  if(!prj->update.window) {
    update_report_new(prj);
  }

  r = ((Report *)prj->update.arg);
  report_show_pixmaps(r);

  gtk_notebook_set_page(GTK_NOTEBOOK(r->notebook), 0);
  gtk_widget_show(prj->update.window);
}


static gint
cb_delete(gpointer data) { 
  return 1; 
}


void
update_progress_bar_new(const char *title, size_t total, Progress *pbar)
{
  GtkWidget *w, *vbox;
  char stitle[256], slabel[256];

  pbar->current = 0;
  pbar->total = total;

  hsh_create_str(stitle, 255, "Running ", title, 0);
  hsh_create_str(slabel, 255, stitle, ":", 0);

  if(!pbar->dialog.window) {
    w = pbar->dialog.window = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_widget_set_usize(w, 300, 72);
    gtk_window_set_title(GTK_WINDOW(w), stitle);
    gtk_window_set_position(GTK_WINDOW(w), GTK_WIN_POS_CENTER);
    gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, TRUE);

    gtk_signal_connect(GTK_OBJECT(w), "destroy",
		  GTK_SIGNAL_FUNC(gtk_widget_destroyed), &pbar->dialog.window);
    gtk_signal_connect(GTK_OBJECT(w), "delete_event",
                       GTK_SIGNAL_FUNC(cb_delete), NULL);

    vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(w), vbox);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);

    pbar->label = gtk_label_new(slabel);
    gtk_box_pack_start(GTK_BOX(vbox), pbar->label, TRUE, TRUE, 0);

    pbar->status = gtk_label_new("");
    gtk_box_pack_start(GTK_BOX(vbox), pbar->status, FALSE, TRUE, 0);

    pbar->pbar = gtk_progress_bar_new();
    gtk_progress_set_format_string(GTK_PROGRESS(pbar->pbar), "%p%%");
    gtk_progress_set_show_text(GTK_PROGRESS(pbar->pbar), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox), pbar->pbar, FALSE, TRUE, 0);
    gtk_widget_show_all(vbox);
  } else {
    gtk_window_set_title(GTK_WINDOW(pbar->dialog.window), stitle);
    gtk_label_set_text(GTK_LABEL(pbar->label), slabel);
    gtk_progress_bar_update(GTK_PROGRESS_BAR(pbar->pbar), 0);
  }

  gtk_widget_show(pbar->dialog.window);
}



void
update_run(Project *prj, Projects *prjs, Progress *pbar)
{
  char *type;
  Node *node;
  Removed *remd;
  HickPackets pack;
  HickResult result;
  AvailVersion *aver;
  size_t n, i, j, need_update, run_update;

  update_clear_report(prj);

  for(n = 0; n < prjs->types.siz; n++) {
    type = parr_get(&prjs->types, char, n);

    if(openPackets(PACKET_PATH, prj->homedir, type, NULL, &pack)) {
      dialog_err_return_new(0, "Open Packet Path",
			    "Failed to open packet path '%s'.", PACKET_PATH);
      return;
    }

    pbar->current = 0;
    pbar->total = count_packets_type(&prjs->packets, type);

    set_label(pbar->status, "Updating packets for ", type);

    run_update = FALSE;

    for(i = 0; i < prj->removed.siz; i++) {
      remd = parr_get(&prj->removed, Removed, i);

      if(remd->type == type) {
	parr_remove(&prj->removed, parr_find(&prj->removed, remd, NULL) - 1);

	removePacket(&pack, remd->packet, remd->version, 1, 1);

	for(node = prj->packets.avail.head; node; node = node->next) {
	  if((((AvailPacket *)node)->packet->type != type) ||
	     strcasecmp(remd->packet, node->name)) {
	    continue;
	  }

	  for(j = 0; j < ((AvailPacket *)node)->vers.siz; j++) {
	    aver = parr_get(&((AvailPacket *)node)->vers, AvailVersion, j);

	    if(aver->info.install != INSTALL_NOT) {
	      addPacket(&pack, node->name, aver->version->name, 1,
			(aver->info.install == INSTALL_DEFAULT));
	    }	  
	  }

	  break;
	}

	FreeMem(remd);
	run_update = TRUE;
      }
    }

    for(node = prj->packets.avail.head; node; node = node->next) {
      if(((AvailPacket *)node)->packet->type != type) {
	continue;
      }

      run_update = TRUE;

      for(need_update = 0, i = 0; i < ((AvailPacket *)node)->vers.siz; i++) {
	aver = parr_get(&((AvailPacket *)node)->vers, AvailVersion, i);

	if(aver->info.rinstall != aver->info.install) {
	  need_update = 1;
	  break;
	}
      }

      if(need_update) {
	for(i = 0; i < ((AvailPacket *)node)->vers.siz; i++) {
	  aver = parr_get(&((AvailPacket *)node)->vers, AvailVersion, i);

	  if(aver->info.install != INSTALL_NOT) {
	    addPacket(&pack, node->name, aver->version->name, 1,
		      (aver->info.install == INSTALL_DEFAULT));
	  } else {
	    if((aver->info.install == INSTALL_NOT) && 
	       (aver->info.rinstall != INSTALL_NOT)) {
	      removePacket(&pack, node->name, aver->version->name, 1, 1);
	    }
	  }
	}
      }
    }

    if(run_update) {
      updatePackets(&pack, &result, (prjs->hicktype != type),  
		    (HICKEVENT_NOTUSED | HICKEVENT_INSTALL |HICKEVENT_UPDATE | 
		     HICKEVENT_REMOVE), (HickEventFunction)hcb_update, pbar);
      
      update_report(prj, &result, type);
    }


    closePackets(&pack);
  }
}


void
update_run_mk(Project *prj, Projects *prjs, Progress *pbar)
{
  size_t n;
  char *type;

  pbar->current = 0;
  pbar->total = (1 + _pref.info_page + _pref.html_page +
		 (_pref.scripts.arri.siz || _pref.scripts.all) +
		 (_pref.menus.arri.siz || _pref.menus.all));

  hcb_update(NULL, pbar);

  for(n = 0; n < prjs->types.siz; n++) {
    pbar->current = 0;
    type = parr_get(&prjs->types, char, n);

    set_label(pbar->status, "Removing old information for ", type);
    hcb_update(NULL, pbar);

    hickmk_html_remove(prj->homedir, type);
    hickmk_info_remove(prj->homedir, type);
    hickmk_menus_remove(prj->homedir, type);
    hickmk_scripts_remove(prj->homedir, type);
    hcb_update(NULL, pbar);

    if(_pref.info_page) {
      set_label(pbar->status, "Generating GNU Info for ", type);

      hickmk_html_create(NULL, 0, HTML_PATH, prj->homedir, type);
      hcb_update(NULL, pbar);
    }

    if(_pref.html_page) {
      set_label(pbar->status, "Generating HTML Index for ", type);

      hickmk_info_create(INFO_PATH, prj->homedir, type, _pref.info_width);
      hcb_update(NULL, pbar);
    }

    if(_pref.scripts.arri.siz || _pref.scripts.all) {
      set_label(pbar->status, "Generating scripts for ", type);

      if(_pref.scripts.all) {
	hickmk_scripts_create(NULL, 0, SCRIPT_PATH, prj->homedir, type);
      } else {
	hickmk_scripts_create((const char **)_pref.scripts.arri.arr,
			      _pref.scripts.arri.siz, SCRIPT_PATH,
			      prj->homedir, type);
      }

      hcb_update(NULL, pbar);
    }

    if(_pref.menus.arri.siz || _pref.menus.all) {
      set_label(pbar->status, "Generating menus for ", type);

      if(_pref.menus.all) {
	hickmk_menus_create(NULL, 0, MENUS_PATH, prj->homedir, type, 
			    _pref.menu_sub);
      } else {
	hickmk_menus_create((const char **)_pref.menus.arri.arr, 
			    _pref.menus.arri.siz, MENUS_PATH, 
			    prj->homedir, type, _pref.menu_sub);
      }

      hcb_update(NULL, pbar);
    }
  }
}


void
rescue_update(Project *prj, const char *type)
{
  HickPackets pack;
  HickResult results;
  char fname[FNAME_LEN + 1];
  Progress pbar = { { NULL, NULL }, 0, 0, NULL, NULL, NULL };

  hsh_create_str(fname, FNAME_LEN, prj->homedir, "/.hick/", type, 0);

  if(!prj->packets.has_init) {
    project_init_packets(prj, &_gmain.projects.packets,&_gmain.projects.types);
  }

  pbar.current = 0;
  update_progress_bar_new("Rescue", count_dirs(fname), &pbar);
  parse_gtk_events();
  HickRescueOpen(&pack, PACKET_PATH, prj->homedir, type);

  set_label(pbar.status, "Preparing for rescue on ", type);
  HickRescuePrepare(&pack, (HickEventFunction)hcb_update, &pbar);

  pbar.current = 0;
  pbar.total = count_packets_type(&_gmain.projects.packets, type);
  set_label(pbar.status, "Updating packets for ", type);

  HickRescueUpdate(&pack, &results, (_gmain.projects.hicktype == type),
		   (HICKEVENT_NOTUSED | HICKEVENT_INSTALL |
		    HICKEVENT_UPDATE | HICKEVENT_REMOVE), 
		   (HickEventFunction)hcb_update, &pbar);

  update_run_mk(prj, &_gmain.projects, &pbar);
  gtk_widget_hide(pbar.dialog.window);
}


static size_t
count_packets_type(NodeList *list, const char *type)
{
  Node *node;
  size_t n = 0;

  for(node = list->head; node; node = node->next) {
    if(((Packet *)node)->type == type) {
      n++;
    }
  }

  return n;
}


static size_t
count_dirs(const char *dirname)
{
  DIR *dirp;
  size_t n = 0;
  struct stat sbuf;
  struct dirent *entry;

  if(!(dirp = opendir(dirname)) || (hsh_chdir(dirname, NULL) == IOW_FAIL)) {
    return 0;
  }

  n++;

  while((entry = readdir(dirp))) {
    if(IS_DOT_OR_DOTDOT(entry)) {
      continue;
    }

    if((hsh_lstat(entry->d_name, &sbuf, NULL) == IOW_OK) &&
       S_ISDIR(sbuf.st_mode)) {

      n += count_dirs(entry->d_name);
    }
  }

  closedir(dirp);
  hsh_chdir("..", NULL);
  return n;
}


static void
hcb_update(HickEvent *event, Progress *pbar)
{
  gfloat r;

  if(event) {
    if(!event->ev_dir && (!event->ev_entry ||
			  ((event->ev_event == HICKEVENT_NOTUSED) &&
			   event->ev_hpac &&
			   !event->ev_hpac->pac_versions.vl_head))) {
      return;
    }
  }

  pbar->current++;

  r = (gfloat)pbar->current / (gfloat)pbar->total;
  gtk_progress_bar_update(GTK_PROGRESS_BAR(pbar->pbar), (r < 1.0 ? r : 1.0));

  parse_gtk_events();
}


static void
parse_gtk_events(void)
{
  while(gtk_events_pending()) {
    gtk_main_iteration();
  }
}


static void
set_label(GtkWidget *label, const char *msg, const char *type)
{
  char buf[FNAME_LEN + 1];

  hsh_create_str(buf, FNAME_LEN, msg, type, ".", 0);
  gtk_label_set_text(GTK_LABEL(label), buf);
  parse_gtk_events();
}


static void
update_report(Project *prj, HickResult *result, const char *type)
{
  size_t n;
  HickForced *fo;
  HickRequire *req;
  HickFailure *fail;
  HickConflict *con;
  HickRequirePacket *rpac, *rpor;

  if(!result->hr_conflicts && !result->hr_forced && !result->hr_failure &&
     !result->hr_required) {
    return;
  }

  if(!prj->update.window) {
    update_report_new(prj);
  }

  for(con = result->hr_conflicts; con; con = con->con_next) {
    report_add_clist_item((Report *)prj->update.arg, TAB_CONFLICT, 
			  type, con->con_name, con->con_version, 
			  con->con_wName, con->con_wVersion);
  }

  for(fo = result->hr_forced; fo; fo = fo->fo_next) {
    if(!fo->fo_hidden) {
      report_add_clist_item((Report *)prj->update.arg, TAB_FORCED, 
			    type, fo->fo_name, fo->fo_version);
    }
  }

  for(fail = result->hr_failure; fail; fail = fail->fail_next) {
    report_add_clist_item((Report *)prj->update.arg, TAB_FAILURE,
			  type, fail->fail_name, fail->fail_version,
			  fail_reasons[fail->fail_reason]);
  }

  if(result->hr_required) {
    printf("  Required:\n");
 
    for(req = result->hr_required; req; req = req->req_next) {
      printf("    %s %s requires\n", req->req_name, req->req_version);

      for(rpac = req->req_list; rpac; rpac = rpac->rpac_next) {
	if(!rpac->rpac_versions.siz) {
	  printf("      %s\n", rpac->rpac_name);
	} else {
	  printf("      %s %s\n", rpac->rpac_name, rpac->rpac_versions.arr[0]);

	  for(n = 1; n < rpac->rpac_versions.siz; n++) {
	    printf("        or %s %s\n", rpac->rpac_name, 
		   rpac->rpac_versions.arr[n]);
	  }
	}

	for(rpor = rpac->rpac_or; rpor; rpor = rpor->rpac_next) {
	  if(!rpor->rpac_versions.siz) {
	    printf("        or %s\n", rpor->rpac_name);
	  } else {
	    for(n = 0; n < rpor->rpac_versions.siz; n++) {
	      printf("        or %s %s\n", rpor->rpac_name, 
		     rpor->rpac_versions.arr[n]);
	    }
	  }
	}
      }
    }
  }

  freeHickResult(result);
}


static void
update_clear_report(Project *prj)
{
  if(prj->update.window) {
    report_clear((Report *)prj->update.arg);
  }
}


static void
update_report_new(Project *prj)
{
  Report *r;
  char title[128];
  GtkWidget *vbox, *w;

  hsh_create_str(title, 127, "Packet Update Report For ", prj->name, 0);

  dialog_new(&prj->update, title,
	     "OK", BUTTON_ISDEF, NULL, NULL, NULL, 0);

  gtk_widget_set_usize(prj->update.window, 530, 300);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(prj->update.window)->vbox), vbox, 
                     TRUE, TRUE, 0);

  gui_pixmap_create_dot(&_gmain.pixmaps, prj->update.window);

  r = report_new(4, _gmain.pixmaps.dot.pix, _gmain.pixmaps.dot.mask);
  prj->update.arg = (void *)r;

  report_tab_set_clist(r, TAB_REQUIRE, "Require", "Required packets.",
		       NULL, prj, 1);
  report_tab_clist_col_set(r, TAB_REQUIRE, 0, "Not Implemented", -1);

  report_tab_set_clist(r, TAB_CONFLICT, "Conflict", "Conflicting packets.",
		       NULL, prj, 5);
  report_tab_clist_col_set(r, TAB_CONFLICT, 0, "Type", 60);
  report_tab_clist_col_set(r, TAB_CONFLICT, 1, "Packet", 105);
  report_tab_clist_col_set(r, TAB_CONFLICT, 2, "Version", 105);
  report_tab_clist_col_set(r, TAB_CONFLICT, 3, "With Packet", 105);
  report_tab_clist_col_set(r, TAB_CONFLICT, 4, "With Version", -1);

  report_tab_set_clist(r, TAB_FAILURE, "Failures",
		       "Packets failed to install.", NULL, prj, 4);
  report_tab_clist_col_set(r, TAB_FAILURE, 0, "Type", 60);
  report_tab_clist_col_set(r, TAB_FAILURE, 1, "Packet", 100);
  report_tab_clist_col_set(r, TAB_FAILURE, 2, "Version", 100);
  report_tab_clist_col_set(r, TAB_FAILURE, 3, "Reason", -1);

  report_tab_set_clist(r, TAB_FORCED, "Forced", "Force installed packets.",
                       NULL, prj, 3);
  report_tab_clist_col_set(r, TAB_FORCED, 0, "Type", 130);
  report_tab_clist_col_set(r, TAB_FORCED, 1, "Packet", 160);
  report_tab_clist_col_set(r, TAB_FORCED, 2, "Version", -1);

  w = report_create(r);
  gtk_box_pack_start(GTK_BOX(vbox), w, TRUE, TRUE, 0);
  gtk_widget_show_all(vbox);  
}
