/* HMerge 1.0.1 -- Packet merger.
** Copyright (C) 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>

#include "nodes.h"
#include "array.h"
#include "getpacket.h"

#define PROGRAM_NAME     "hmerge"
#define PROGRAM_VERSION  "1.0.1"


typedef struct {
  NodeNum num;

  char *filename;

  HickPacket pkg;
} MergePacket;


static void default_set(HickPacket *pkg, PArr *defs);
static int merge_packets(HickPacket *pkg, NodeList *packets);
static int load_file(MergePacket *mpkg);
static int get_weight(char *filename);

static int
usage(const char *filename)
{
  fprintf(stderr, "Usage: %s [ OPTIONS ] FILE [ FILE ... ]\n\n"
	  "\t-h\t\tThis help text.\n"
	  "\t-v\t\tVersion information.\n"
	  "\t-d VERSION\tSpecifies default version(s). The option can "
	  "be used\n\t\t\tmultiple times, If no default set first version "
	  "will be \n\t\t\tset as the default.\n\n", filename);
  return 0;
}


int
main(int argc, char **argv)
{
  int n;
  int weight;
  Node *node;
  NodeList list;
  MergePacket *mpkg;
  PArr defvers;

  parr_init(&defvers);

  while((n = getopt(argc, argv, "hvd:")) != EOF) {
    switch(n) {
      case 'v': {		/* version information */
	printf(PROGRAM_NAME " " PROGRAM_VERSION "\n");
	return 0;
      }
      case 'd': {		/* default version */
	parr_add(&defvers, optarg);
	break;
      }
      case 'h':			/* usage info */
      default: {		/* ERROR! */
        return usage(argv[0]);
      }
    }
  }

  if(optind >= argc) {
    return usage(argv[0]);
  }

  hsh_nodelist_init(&list);

  for(n = optind; n < argc; n++) {
    if((weight = get_weight(argv[n])) == -1) {
      fprintf(stderr, "error: not a valid filename '%s'.\n", argv[n]);
      continue;
    }
      
    for(node = list.head; node; node = node->next) {
      if(((NodeNum *)node)->num == weight) {
	fprintf(stderr, "error: Same weight '%d' on one or more files.\n", 
		weight);
	return 1;
      }
    }

    if(!(mpkg = MallocMem(MergePacket))) {
      fprintf(stderr, "error: Out of memory!\n");
      return 1;
    }

    mpkg->num.num = weight;
    mpkg->filename = argv[n];
    initHickPacket(&mpkg->pkg);

    if(!load_file(mpkg)) {
      FreeMem(mpkg);
      continue;
    }

    hsh_node_num_insert(&mpkg->num, &list, 0);
  }

  if(!list.head) {
    fprintf(stderr, "error: no packets loaded -- exiting.\n");
    return 1;
  }

  mpkg = ((MergePacket *)list.head);
  hsh_node_remove(list.head, &list);

  if(!merge_packets(&mpkg->pkg, &list)) {
    default_set(&mpkg->pkg, &defvers);

    savePacket(stdout, &mpkg->pkg, 0, HTIME_DATE);
  }

  return 0;
}


static void
default_set(HickPacket *pkg, PArr *defs)
{
  HickVer *ver;
  int defset = 0;

  for(ver = pkg->pac_versions.vl_head; ver; ver = ver->ver_next) {
    ver->ver_default = 0;

    if(parr_find_value(defs, char, ver->ver_number, parr_strcasecmp)) {
      if(defset && !pkg->pac_multiple) {
	continue;
      }

      defset = 1;
      ver->ver_default = 1;
    }
  }

  if(!defs->siz && pkg->pac_versions.vl_head) {
    pkg->pac_versions.vl_head->ver_default = 1;
  }
}


static int
merge_packets(HickPacket *pkg, NodeList *packets)
{
  Node *node;
  MergePacket *mpkg;
  HickVer *ver, *nver;

  for(node = packets->head; node; node = node->next) {
    mpkg = ((MergePacket *)node);

    for(ver = mpkg->pkg.pac_versions.vl_head; ver; ver = nver) {
      nver = ver->ver_next;

      if(_versionFind(ver, &pkg->pac_versions)) {
	fprintf(stderr, "duplicate copies of version '%s'\n", ver->ver_number);
	continue;
      }

      _versionAppend(ver, &pkg->pac_versions);
    }
  }

  return 0;
}


static int
load_file(MergePacket *mpkg)
{
  FILE *file;
  HickEntry entry;

  if(!(file = fopen(mpkg->filename, "r"))) {
    fprintf(stderr, "Failed to open '%s'.\n", mpkg->filename);
    return 0;
  }

  initHickEntry(&entry);
  
  if(!_getPacketHead(file, &entry)) {
    freeHickEntry(&entry);
    fclose(file);
    fprintf(stderr, "Failed to load '%s': packet header error.\n", 
	    mpkg->filename);
    return 0;
  }

  strcpy(mpkg->pkg.pac_name, entry.he_name);

  mpkg->pkg.pac_groups.arr = entry.he_groups.arr;
  mpkg->pkg.pac_groups.siz = entry.he_groups.siz;
  mpkg->pkg.pac_groups.alloc = entry.he_groups.alloc;

  hsh_array_init(&entry.he_groups);

  _getPacket(file, 0, &mpkg->pkg, 1);
  fclose(file);
  return 1;
}


static int
get_weight(char *filename) 
{
  int n;
  char *ptr;

  for(n = 0; filename[n] && (filename[n] != '-'); n++);

  if(!filename[n]) {
    return -1;
  }

  ptr = filename + n + 1;

  while(filename[++n]) {
    if(!isdigit(filename[n])) {
      return -1;
    }
  }

  return atoi(ptr);
}
