/* HMod 1.0.0 -- Hick modules emulation.
** Copyright (C) 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "entry.h"
#include "version.h"
#include "hmodadd.h"
#include "hmodlist.h"
#include "hmodload.h"

static int hmod_add_installed(HickPacket *hpkg, PArr *loaded, int tryadd);

static int hmod_add_conflict_test(HickPacket *hpkg, HickPackets *pkg, 
				  PArr *loaded, int tryadd, int confchk);

static int hmod_add_parse_versions(HickPacket *hpkg, PArr *loaded, int tryadd,
				   PArr *vers, int isdef, HickVerList *nouse);


void
hmod_add_get(HModEnv *henv, NodeList *scripts, HickPackets *pkg, PArr *loaded, 
	     const char *shell, int tryadd, char **packets, size_t siz,
	     int confchk)
{
  HModLoadUpdate update;

  update.func_ver       = hmod_add_parse_versions;
  update.func_insted    = hmod_add_installed;
  update.func_conflict  = hmod_add_conflict_test;
  update.func_env       = hmod_add_environ;
  update.func_loaded    = hmod_list_add;
  update.conflict_chk   = confchk;
  update.environ_rescue = 0;

  hmod_load_update(&update, henv, pkg, loaded, tryadd, packets, siz);
}


static int
hmod_add_installed(HickPacket *hpkg, PArr *loaded, int tryadd)
{
  if(!hpkg->pac_multiple && hmod_list_pkg_find(loaded, hpkg->pac_name)) {
    if(!tryadd) {
      fprintf(stderr, "module: version of '%s' is already loaded.\n",
	      hpkg->pac_name);
    }

    return 1;
  }

  return 0;
}


static void
conflict_print(const char *name, const char *version, const char *cname, 
	       const char *cversion)
{
  if(version) {
    fprintf(stderr, "module: %s %s conflicts with %s %s\n", name, version, 
	    cname, cversion);
  } else {
    fprintf(stderr, "module: %s conflicts with %s %s\n", name, cname,cversion);
  }
}


static int
hmod_add_conflict(const char *name, const char *version, HickEnv *head, 
		  userEntryList *insted, PArr *loaded, int tryadd, int confchk)
{
  size_t n;
  HickEnv *env;
  userEntry *ue;
  int conflict = 0;
  HModLoaded *litem;

  for(env = head; env; env = env->env_next) {
    if(!env->u.env_versions->siz) {
      if((litem = hmod_list_pkg_find(loaded, env->env_name))) {
	if(!tryadd) {
	  conflict_print(name, 0, litem->name, litem->version);
	}

	conflict = 1;
	continue;
      }

      if(confchk) {
	if((ue = _entryFind(env->env_name, insted))) {
	  if(!tryadd) {
	    conflict_print(name, 0, ue->ue_name, ue->ue_version);
	  }

	  conflict = 1;
	}
      }
    } else {
      for(n = 0; n < env->u.env_versions->siz; n++) {
	if((litem = hmod_list_find(loaded, env->env_name,
				   env->u.env_versions->arr[n]))) {
	  if(!tryadd) {
	    conflict_print(name, version, litem->name, litem->version);
	  }

	  conflict = 1;
	  continue;
	}

	if(confchk) {
	  ue = _entryFindTwo(env->env_name, env->u.env_versions->arr[n], 
			     insted);

	  if(ue) {
	    if(!tryadd) {
	      conflict_print(name, version, ue->ue_name, ue->ue_version);
	    }

	    conflict = 1;
	  }
	}
      }
    }
  }

  return conflict;
}


static int
hmod_add_conflict_test(HickPacket *hpkg, HickPackets *pkg, PArr *loaded, 
		       int tryadd, int confchk)
{
  HickVer *ver;
  int conflict = 0;

  if(hmod_add_conflict(hpkg->pac_name, 0, hpkg->pac_conflict.el_head, 
		       &pkg->hp_installed, loaded, tryadd, confchk)) {
    conflict = 1;
  }

  for(ver = hpkg->pac_versions.vl_head; ver; ver = ver->ver_next) {
    if(hmod_add_conflict(hpkg->pac_name, ver->ver_number,
			 HVER_CONFLICT(ver).el_head, &pkg->hp_installed,
			 loaded, tryadd, confchk)) {
      conflict = 1;
    }
  }

  return conflict;
}


static int
hmod_add_parse_versions(HickPacket *hpkg, PArr *loaded, int tryadd, PArr *vers,
			int isdef, HickVerList *nouse)
{
  HickVer *hver, *rver, *dver;

  if(!isdef) {			/* get matching version */
    for(hver = hpkg->pac_versions.vl_head; hver; ) {
      if(!hmod_list_ver_find(vers, hver->ver_number)) {
	rver = hver;
	hver = hver->ver_next;
	_versionMove(rver, &hpkg->pac_versions, nouse);
	continue;
      }

      hver = hver->ver_next;
    }

    if(hpkg->pac_versions.vl_head) {
      return 1;
    }

    return 0;
  }

  for(hver = hpkg->pac_versions.vl_head; hver; hver = hver->ver_next) {
    if(hver->ver_default) {
      break;
    }
  }

  if(hver) {			/* save default version(s) */
    dver = 0;
    for(hver = hpkg->pac_versions.vl_head; hver; ) {
      if(hver->ver_default && (!dver || hpkg->pac_multiple)) {
	dver = hver;
	hver = hver->ver_next;
      } else {
	rver = hver;
	hver = hver->ver_next;
	_versionMove(rver, &hpkg->pac_versions, nouse);
      }
    }
  } else {			/* save first version */
    for(hver = hpkg->pac_versions.vl_head->ver_next; hver; ) {
      _versionMove(rver, &hpkg->pac_versions, nouse);
      hver = hpkg->pac_versions.vl_head->ver_next;
    }
  }

  return 1;
}


void
hmod_add_environ(HickEnv *env, HModEnv *henv)
{
  switch(env->env_flags) {
    case HFLAG_ENV_PREPEND: {
      hmod_env_prepend(henv, env->env_name, env->env_value);
      break;
    }
    case HFLAG_ENV_APPEND: {
      hmod_env_append(henv, env->env_name, env->env_value);
      break;
    }
    case HFLAG_ENV_SET: {
      hmod_env_set(henv, env->env_name, env->env_value);
      break;
    }
    case HFLAG_ENV_UNSET: {
      hmod_env_unset(henv, env->env_name);
      break;
    }
  }
}
