/* HICK Lib 1.3.2 - User Profiler Library.
** Copyright (C) 97, 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>

#include "common.h"


int
_convertString(const char *s, char *buf, size_t maxlen)
{
  size_t i, n;

  for(i = n = 0; s[n] && (i < maxlen - 2); ++i, ++n) {
    switch(s[n]) {
      case '\n': {
	buf[i] = '\\';
	buf[++i] = 'n';
	break;
      }
      case '\\':
      case '"': {
	buf[i++] = '\\';
      }
      default: {
	buf[i] = s[n];
      }
    }
  }

  buf[i] = (char)0;
  return n;
}


struct dirent *
_getNextEntry(HickPackets *pack)
{
  size_t len;
  struct stat sbuf;
  char fname[FNAME_LEN+1];

  do {
    if(!pack->hp_packdir || !pack->hp_curpack) {
      while((pack->hp_curtype = readdir(pack->hp_typedir))) {
	if(IS_DOT_OR_DOTDOT(pack->hp_curtype)) {
	  continue;
	}

	if(pack->hp_type && strcasecmp(pack->hp_curtype->d_name, COMMON_DIR) &&
	   strcasecmp(pack->hp_curtype->d_name, pack->hp_type)) {
	  continue;
	}

	hsh_create_str(fname, FNAME_LEN, pack->hp_packpath, "/",
		       pack->hp_curtype->d_name, 0);

	if((pack->hp_packdir = hsh_opendir(fname, NULL))) {
	  break;
	}
      }

      if(!pack->hp_curtype) {
	return 0;
      }
    }

    while((pack->hp_curpack = readdir(pack->hp_packdir))) {
      hsh_create_str(fname, FNAME_LEN, pack->hp_packpath, "/",
		     pack->hp_curtype->d_name, "/",
		     pack->hp_curpack->d_name, 0);

      if((hsh_stat(fname, &sbuf,NULL) == IOW_FAIL) || !S_ISREG(sbuf.st_mode)) {
	continue;
      }

      len = NAMLEN(pack->hp_curpack) - 1;

      if(IS_EMACS_BACKUP(pack->hp_curpack, len)) {
	continue;
      }

      if((pack->hp_file = fopen(fname, "r"))) {
	pack->hp_isopen = 1;
	return pack->hp_curpack;
      }
    }

    closedir(pack->hp_packdir);
    pack->hp_packdir = 0;
  } while(pack->hp_curtype);

  return 0;
}


void
bflush(Buffert *b)
{
  if(b->pos) {
    b->buf[b->pos] = (char)0;
    fwrite(b->buf, b->pos, 1, b->file);
    b->pos = 0;
  }
}


void
_convertBstring(const char *s, int c, Buffert *b)
{
  if(!s) {
    return;
  }

  for( ; *s; s++) {
    if((b->pos + 2) >= BUFFERTSIZ) {
      fwrite(b->buf, b->pos, 1, b->file);
      b->pos = 0;
    }

    if(*s == '\n') {
      b->buf[b->pos++] = '\\';
      b->buf[b->pos++] = 'n';
      continue;
    }

    if((*s == '\\') || (*s == c)) {
      b->buf[b->pos++] = '\\';
    }

    b->buf[b->pos++] = *s;
  }
}


void
_bputs(const char *s, size_t len, Buffert *b)
{
  /* greater then the buffert size twice? flush buffert, and write string */

  if(len >= BUFFERTSIZ * 2) {
    fwrite(b->buf, b->pos, 1, b->file);
    b->pos = 0;
    fwrite(s, len, 1, b->file);
    return;
  }

  /* is it less then the buffert with what's stored in it? copy it over. */

  if(len < BUFFERTSIZ - b->pos) {
    for( ; *s; ++b->pos, ++s) {
      b->buf[b->pos] = *s;
    }
    return;
  }

  /* It's less the the buffert twice, and current buffert with the string
   * length is greater then the max buffert size. Copy over the buffert as
   * much as it can, the flush, and copy the rest over.
   */

  for( ; *s && (b->pos < BUFFERTSIZ); ++b->pos, ++s) {
    b->buf[b->pos] = *s;
  }

  fwrite(b->buf, b->pos, 1, b->file);
  b->pos = 0;

  for( ; *s; ++b->pos, ++s) {
    b->buf[b->pos] = *s;
  }
}

/*   %% -- %
**   %s -- string
**   %S -- special string converts (\ -> \\, '"' -> \'"', newline -> \n)
**   %C -- special string converts (\ -> \\, : -> \:, newline -> \n)
**   %d -- integer
**   %o -- octal
*/

void
bprint(Buffert *b, const char *fmt, ...)
{
  va_list pvar;
  char *p, *t, buf[BUFLEN+1];

  va_start(pvar, fmt);

  for( ; *fmt; ++fmt) {
    if(*fmt == '%') {
      switch(*(++fmt)) {
	case '%': {
	  bput(*fmt, b);
	  break;
	}
	case 's': {		/* string nothing converted */
	  if((p = va_arg(pvar, char *))) {
	    bputs(p, b);
	  }
	  break;
	}
	case 'S': {		/* get '"' converted */
	  t = p = va_arg(pvar, char *);

	  if(!p || (*p == (char)0)) {
	    _bputs("\"\"", 2, b);
	    break;
	  }

	  /* first we check if we need to put '"' around it. */

	  for( ; *t && (*t != '"') && (*t != '}') && (*t != '#') &&
		 !isspace((int)*t); ++t);

	  if(*t) {
	    bput('"', b);
	  }

	  _convertBstring(p, '"', b);

	  if(*t) {
	    bput('"', b);
	  }
	  break;
	}
	case 'C': {		/* get ':' converted */
	  _convertBstring(va_arg(pvar, char *), ':', b);
	  break;
	}
	case 'd': {		/* a digit (written as a unsigned long). */
#ifdef HAVE_SNPRINTF
	  snprintf(buf, BUFLEN, "%lu", va_arg(pvar, unsigned long));
#else
	  sprintf(buf, "%lu", va_arg(pvar, unsigned long));
#endif
	  for(p = buf; *p; ++p) {
	    bput(*p, b);
	  }
	  break;
	}
	case 'o': {		/* write in octal */
#ifdef HAVE_SNPRINTF
	  snprintf(buf, BUFLEN, "%o", va_arg(pvar, unsigned int));
#else
	  sprintf(buf, "%o", va_arg(pvar, unsigned int));
#endif
	  for(p = buf; *p; ++p) {
	    bput(*p, b);
	  }
	  break;
	}
      }
    } else {			/* if not a %, then we'll add it as it was */
      bput(*fmt, b);
    }
  }

  va_end(pvar);
}
