/* HICK Lib 1.3.2 - User Profiler Library.
** Copyright (C) 97, 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _HICK_H
#define _HICK_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <time.h>
#include <stdlib.h>

#ifdef STDC_HEADERS
# include <string.h>
#endif

#include <hicksh.h>

#undef HICKLIB_MAJOR_VERSION
#define HICKLIB_MAJOR_VERSION 1

#undef HICKLIB_MINOR_VERSION
#define HICKLIB_MINOR_VERSION 3

#undef HICKLIB_MICRO_VERSION
#define HICKLIB_MICRO_VERSION 2

#ifndef PVER_LEN
#define PVER_LEN   16           /* max verison length of a packet */
#endif

#ifndef PNAME_LEN
#define PNAME_LEN  32           /* max name length of a packet */
#endif

#ifndef PGROUP_LEN
#define PGROUP_LEN 32           /* max group len of a packet */
#endif

/* savePacket() date flag */

#define HTIME_TIMET    0	/* write the time_t for date format */
#define HTIME_DATE     1	/* write date (dd/mm/yyyy) */
#define HTIME_HOURS    2	/* write date with hour (dd/mm/yyyy,hh:mm) */
#define HTIME_SECONDS  3	/* write date with sec (dd/mm/yyyy,hh:mm.ss) */

/* Hick funtions error codes */

#define HICKERR_HOMEDIR       -1 /* can't enter home path */
#define HICKERR_HICKDIR       -2 /* can't enter hick dir. */
#define HICKERR_TYPEDIR       -3 /* can't enter type dir. */
#define HICKERR_TYPECONF      -4 /* can't enter type configure dir. */
#define HICKERR_INSTPACK      -5 /* can't enter installed packets dir. */
#define HICKERR_SCRIPTDIR     -6 /* can't enter inline shell script dir. */

#define HICKERR_PACKDIR      -20 /* can't enter packet types dir. */

#define HICKERR_LOGFILE      -30 /* can't open runtime log file for writting.*/
#define HICKERR_ENVFILE      -31 /* can't open environ file for writting. */
#define HICKERR_INFOFILE     -32 /* can't open GNU info file for writting. */
#define HICKERR_MENUFILE     -33 /* can't open WM menus file for writting. */
#define HICKERR_XRESFILE     -34 /* can't open Xresources file for writting. */
#define HICKERR_INSTFILE     -35 /* can't open installed file for writting */
#define HICKERR_CALLFILE     -36 /* can't open call script for appending */
#define HICKERR_HTMLFILE     -37 /* can't open html file for writting. */
#define HICKERR_RENAME       -38 /* can't rename the .installed to installed */
#define HICKERR_CREATEDFILE  -39 /* can't open created file for writting */

#define HICKERR_OUTOFMEMORY  -50 /* Out of memory! */
#define HICKERR_NOTFOUND     -51 /* Packet was not found. */

/* Hick flags  */

#define HFLAG_ENV_PREPEND    0
#define HFLAG_ENV_APPEND     1
#define HFLAG_ENV_SET        2
#define HFLAG_ENV_UNSET      3

#define HFLAG_LINK_NORMAL    0
#define HFLAG_LINK_RENAME    1

#define HFLAG_XRES_INLINE    0
#define HFLAG_XRES_FILENAME  1

#define HFLAG_CREATE_NEW     0
#define HFLAG_CREATE_APPEND  1

#define HFLAG_CALL_INSTALL   (1<<0)
#define HFLAG_CALL_UPDATE    (1<<1)
#define HFLAG_CALL_REMOVE    (1<<2)

#define HED_ENV_INFO       1	/* info      - GNU Info             */
#define HED_ENV_CALL       2	/* call      - Script calling       */
#define HED_ENV_CONFLICT   3	/* conflict  - Conflicts            */
#define HED_ENV_CREATE     4	/* create    - Create file          */
#define HED_ENV_ENV        5	/* env       - Enviroment variables */
#define HED_ENV_LINK       6	/* link      - File linking         */
#define HED_ENV_MITEM      7	/* menuItem  - Menu item for WM     */
#define HED_ENV_REQUIRE    8	/* require   - Required packets     */
#define HED_ENV_SCRIPT     9	/* script    - Inline scripts       */
#define HED_ENV_XRES      10	/* xres      - X resources          */
#define HED_ENV_HTML      11	/* html      - HTML Item            */

#define HED_DEF_PERM     493	/* Default file permission. (755 -rwxr-xr-x) */

/* default priority levels */

#define HICKPRI_SCRIPTS     50  /* default priority level for scripts */
#define HICKPRI_MENUITEM    50  /* default priority for menu items */
#define HICKPRI_ENVIRON     50  /* default priority level for env variables */

/* Hick event defines */

#define HICKEVENT_NONE		0       /* no events */
#define HICKEVENT_NOTUSED	(1<<1)	/* Packtes which isn't used
					 * (note: packets only */
#define HICKEVENT_INSTALL	(1<<3)	/* Packet or version is installing */
#define HICKEVENT_UPDATE	(1<<2)	/* Packet or version is updated */
#define HICKEVENT_REMOVE	(1<<4)	/* Packet or version is removed */
#define HICKEVENT_VERSION	(1<<5)	/* Call event for each version,
					 * insted of for the whole packet. */
#define HICKEVENT_BEFORE	(1<<6)	/* Do the call before the changes is
					 * made instead of after. */
#define HICKEVENT_RDIR          (1<<7)  /* Event when removing a directory,
					 * used in HickRemoveType() and
					 * HickRescuePrepare().
					 */

enum uentryFlags {		/* userEntry flags */
  UE_NONE    = 0,		/* No flag. */
  UE_DEFAULT = 1,		/* Default version. */
  UE_FORCE   = 2,		/* Force installed packet. */
  UE_RWHOLE  = 3		/* Remove whole packets. */
};

/* userEntry *************************************************************** */

typedef struct _userEntry {
  struct _userEntry *ue_next;	/* pointer to next packet. */
  struct _userEntry *ue_prev;	/* pointer to previous packet. */

  char ue_name[PNAME_LEN+1];	/* packet name. */
  char ue_version[PVER_LEN+1];	/* packet version. */
  time_t ue_date;		/* date information. */
  unsigned int ue_flags;	/* flags */
} userEntry;

/* userEntryList *********************************************************** */

typedef struct {
  userEntry *ul_head;		/* Pointer to first userEntry in list */
  userEntry *ul_tail;		/* Pointer to last userEntry in list */
} userEntryList;

/* HickPackets ************************************************************* */

typedef struct {
  DIR *hp_typedir;              /* directory containing the different types. */
  struct dirent *hp_curtype;    /* current directory entry of the types. */

  DIR *hp_packdir;              /* directory containing the type packets. */
  struct dirent *hp_curpack;    /* current packent entry we are poking in. */

  FILE *hp_file;                /* pointer to file which we are using. */
  int hp_isopen;                /* we got an open file pointer? */

  char *hp_packpath;            /* path to where the packets can be found. */
  char *hp_homepath;            /* the users homepath, where to install. */

  char *hp_type;                /* which packet type are we working on. */
  char hp_group[PGROUP_LEN+1];  /* which kind of groups do we want? */

  userEntryList hp_remove;
  userEntryList hp_install;
  userEntryList hp_installed;
} HickPackets;

/* HickVer ***************************************************************** */

typedef struct _hickver {
  struct _hickver *ver_next;	/* Pointer to next version in list. */
  struct _hickver *ver_prev;	/* Pointer to previous version in list. */

  char ver_number[PVER_LEN+1];	/* the version number */

  char *ver_about;		/* Extra about message for version */
  int ver_override;		/* If this about message will override the
				 * main about message. */

  time_t ver_created;		/* date of creation of this version */
  time_t ver_modified;		/* last time modified */
  time_t ver_expire;		/* Date when the version is expireing */

  int ver_default;		/* is it the default version. */
} HickVer;

/* HickVerList ************************************************************* */

typedef struct {
  HickVer *vl_head;		/* Pointer to first HickVer in list. */
  HickVer *vl_tail;		/* Pointer to last HickVer in list. */
} HickVerList;

/* HickEnvItem ************************************************************* */

typedef struct _hickenvitem {
  struct _hickenvitem *eit_next; /* Pointer to next item in list. */

  char *eit_name;		/* Name of the variable. */
  char *eit_value;		/* Value of the variable. */
} HickEnvItem;

/* HickSubGroup ************************************************************ */

typedef struct _hicksubgroup {
  char hsg_group[PGROUP_LEN+1];	/* On which group is this subgroup. */
  char hsg_name[PGROUP_LEN+1];	/* Name on this subgroup. */

  ArrayInfo hsg_versions;	/* Array with valid version numbers */

  struct _hicksubgroup *hsg_down; /* Pointer to subgroup below this one */
  struct _hicksubgroup *hsg_next; /* Pointer to next subgroup in list */
} HickSubGroup;

/* HickRequirePacket ******************************************************* */

typedef struct _hickreqpac {
  struct _hickreqpac *rpac_next; /* Pointer to next packet required. */

  struct _hickreqpac *rpac_or;  /* Pointer to another packet which can be
                                 * installed insted of this one.
                                 */

  char rpac_name[PNAME_LEN+1];  /* Required packet name. */

  ArrayInfo rpac_versions;	/* Array of versions, at least one required */
} HickRequirePacket;

/* HickEnv ***************************************************************** */

typedef struct _hickenv {
  struct _hickenv *env_next;	/* pointer to next env in list. */
  struct _hickenv *env_prev;	/* pointer to previous env in list. */

  unsigned int env_flags;	/*   Create: new, append
				 *      Env: prepend, append, set
				 *     Link: normal, replace
				 *     Xres: inline, filename
				 *     Call: install, update, remove
				 * Conflict: size of versions array
				 */

  unsigned int env_priority;	/*   Create: mode
				 *      Env: priority
				 * MenuItem: priority
				 *   Script: priority
				 */

  char *env_name;		/*     Call: filename
				 * Conflict: name
				 *   Create: filename
				 *      Env: name
				 *     Info: name
				 *     Link: to
				 * MenuItem: menu name
				 *  Require: name
				 *   Script: type
				 *     Xres: filename or name
				 *     HTML: html code
				 */

  char *env_value;		/*     Call: options
				 * Conflict: version
				 *   Create: contents
				 *      Env: value
				 *     Info: filename
				 *     Link: from
				 * MenuItem: command
				 *  Require: version
				 *   Script: script
				 *     Xres: value (if not filename)
				 *     HTML: topic
				 */

  char *env_itemTitle;		/*     Info: sub menu
				 * MenuItem: item title
				 */

  char *env_about;		/*     Info: about */

  union {
    char *env_topic;		/*     Info: topic */
    HickRequirePacket *env_rpacs; /*Require: List of require or stmt */
    HickEnvItem *env_items;	/* MenuItem: List of menu variables */
    ArrayInfo *env_versions;	/* Conflict: versions */
  } u;

  ArrayInfo env_types;		/* to which OSes this belongs to. */
} HickEnv;

/* HickEnvList ************************************************************* */

typedef struct {
  HickEnv *el_head;		/* pointer to first env in list. */
  HickEnv *el_tail;		/* pointer to last env in list. */
} HickEnvList;

/* HickVersion ************************************************************* */

typedef struct {
  HickVer hver_version;		/* Version information. */

  HickEnvList hver_info;	/* Info variables. */
  HickEnvList hver_call;	/* Call variables. */
  HickEnvList hver_conflict;	/* Conflict variables. */
  HickEnvList hver_create;	/* Create variables. */
  HickEnvList hver_env;		/* Env variables. */
  HickEnvList hver_link;	/* Link variables. */
  HickEnvList hver_mitem;	/* MenuItem variables. */
  HickEnvList hver_require;	/* Require varaibles. */
  HickEnvList hver_script;	/* Script variables. */
  HickEnvList hver_xres;	/* Xres variables. */
  HickEnvList hver_html;	/* HTML variables. */

  ArrayInfo hver_types;		/* in what OS(es) does this version belong? */
} HickVersion;

#define HVER_VERSION(v)  (((HickVersion *)(v))->hver_version)
#define HVER_INFO(v)     (((HickVersion *)(v))->hver_info)
#define HVER_CALL(v)     (((HickVersion *)(v))->hver_call)
#define HVER_CONFLICT(v) (((HickVersion *)(v))->hver_conflict)
#define HVER_CREATE(v)   (((HickVersion *)(v))->hver_create)
#define HVER_ENV(v)      (((HickVersion *)(v))->hver_env)
#define HVER_LINK(v)     (((HickVersion *)(v))->hver_link)
#define HVER_MITEM(v)    (((HickVersion *)(v))->hver_mitem)
#define HVER_REQUIRE(v)  (((HickVersion *)(v))->hver_require)
#define HVER_SCRIPT(v)   (((HickVersion *)(v))->hver_script)
#define HVER_XRES(v)     (((HickVersion *)(v))->hver_xres)
#define HVER_HTML(v)     (((HickVersion *)(v))->hver_html)
#define HVER_TYPES(v)    (((HickVersion *)(v))->hver_types)

/* HickEntry *************************************************************** */

typedef struct {
  char he_name[PNAME_LEN+1];	/* name of the packet */

  ArrayInfo he_groups;		/* array of groups this packet belongs to */

  char *he_about;		/* about message */

  time_t he_created;		/* date of creation of this packet */
  time_t he_modified;		/* last time modified */
  time_t he_expire;		/* Date when the packet is expiering */

  HickVerList he_versions;	/* List of versions */
  HickSubGroup *he_subGroups;	/* List of subgroups */

  unsigned int he_hidden   : 1; /* should this packet be hidden from view */
  unsigned int he_required : 1; /* is this packet required to be installed? */
  unsigned int he_multiple : 1; /* can multiple versions be installed? */

  ArrayInfo he_using;		/* Array of versions in use. */
  ArrayInfo he_force;		/* Array of versions which is forced on this
				 * user.
				 */
} HickEntry;

/* HickRequire ************************************************************* */

typedef struct _hickrequire {
  struct _hickrequire *req_next; /* pointer to next HickRequire in list. */
  
  char req_name[PNAME_LEN+1];	/* Name of the packet requireing the packet */
  char req_version[PVER_LEN+1]; /* Version number of the packet requireing
				 * the packet */

  HickRequirePacket *req_list;	/* List of packets required by this program. */
} HickRequire;

/* HickConflict ************************************************************ */

typedef struct _hickconflict {
  struct _hickconflict *con_next; /* pointer to next HickConflict in list */

  char con_name[PNAME_LEN+1];   /* name of packet in conflict */
  char con_version[PVER_LEN+1]; /* version of packet in conflict */

  char con_wName[PNAME_LEN+1];  /* name of packet it's in conflict with. */
  char con_wVersion[PVER_LEN+1]; /* version of packet it's in conflict with. */
} HickConflict;

/* HickFailure ************************************************************* */

enum failreason {
  FAIL_NOTFOUND_PACKET,		/* install: packet couldn't be found */
  FAIL_NOTFOUND_VERSION,	/* install: version couldn't be found */
  FAIL_NOTFOUND_DEFAULT,	/* install: default version wasn't found */
  FAIL_FORCE_INSTALLED,		/* install: force overrides install */

  FAIL_MISSING_PACKET,		/* installed: the packet is missing */
  FAIL_MISSING_VERSION,		/* installed: the version is missing */
  FAIL_MISSING_DEFAULT,		/* installed: default version is missing. */
  FAIL_FORCE_REMOVED,		/* installed: was removed by force */

  FAIL_REMOVE_FORCE,		/* remove: can't remove, force installed */
  FAIL_REMOVE_PACKET,		/* remove: missing packet */
  FAIL_REMOVE_VERSION,		/* remove; missing version */

  FAIL_EXPIRED_PACKET,		/* Packet has expired */
  FAIL_EXPIRED_VERSION		/* Version has expired */
};

typedef struct _hickfailure {
  struct _hickfailure *fail_next; /* pointer to next HickFailure in list */

  char fail_name[PNAME_LEN+1];    /* name of packet which failed */
  char fail_version[PVER_LEN+1];  /* version of packet which failed */

  enum failreason fail_reason;	  /* reason why it failed to install. */
} HickFailure;

/* HickForced ************************************************************** */

typedef struct _hickforced {
  struct _hickforced *fo_next;	/* pointer to next HickForced in list. */

  unsigned int fo_hidden : 1;	/* is this a hidden packet */

  char fo_name[PNAME_LEN+1];	/* name of forced packet */
  char fo_version[PVER_LEN+1];	/* version of forced packet */
} HickForced;

/* HickResult ************************************************************** */

typedef struct {
  HickRequire *hr_required;	/* List of required packets. */
  HickConflict *hr_conflicts;	/* List of packets in conflict */
  HickFailure *hr_failure;	/* Packets who was failed to be installed */
  HickForced *hr_forced;	/* List of packets force installed */
} HickResult;

/* HickDirection *********************************************************** */

typedef struct _hickdirect {
  struct _hickdirect *hd_next;	/* Pointer to next HickDirection in list*/

  ArrayInfo hd_names;		/* Names which on these directions aplies to */

  unsigned int hd_group : 1;	/* group - TRUE, user - FALSE */

  ArrayInfo hd_versions;	/* On what versions this direction aplies on */
} HickDirection;

/* HickPacket ************************************************************** */

typedef struct {
  char pac_name[PNAME_LEN+1];	/* packet name */

  ArrayInfo pac_groups;

  char *pac_about;		/* about message of the packet. */

  unsigned int pac_hidden   : 1; /* should this packet be hidden from view */
  unsigned int pac_required : 1; /* is this packet required to be installed? */
  unsigned int pac_multiple : 1; /* can multiple versions be installed? */

  time_t pac_created;		/* date of creation of this version. */
  time_t pac_modified;		/* last time modified. */
  time_t pac_expire;		/* Date when the packet is expireing */

  unsigned int pac_priority;	/* packet priority level. */

  HickDirection *pac_allow;	/* List of allow permissions. */
  HickDirection *pac_deny;	/* List of deny permissions. */
  HickDirection *pac_force;	/* List of force settings. */

  HickSubGroup *pac_subGroups;	/* List of subgroups. */

  HickVerList pac_versions;	/* List of versions. */

  HickEnvList pac_info;		/* Info variables. */
  HickEnvList pac_call;		/* Call variables. */
  HickEnvList pac_conflict;	/* Conflict variables. */
  HickEnvList pac_create;	/* Create variables. */
  HickEnvList pac_env;		/* Env variables. */
  HickEnvList pac_link;		/* Link variables. */
  HickEnvList pac_mitem;	/* MenuItem variables. */
  HickEnvList pac_require;	/* Require varaibles. */
  HickEnvList pac_script;	/* Script variables. */
  HickEnvList pac_xres;		/* Xres variables. */
  HickEnvList pac_html;		/* HTML variables. */
} HickPacket;

/* HickEvent *************************************************************** */

typedef struct {
  HickPackets *ev_pack;		/* pointer to the current HickPackets used. */
  HickPacket *ev_hpac;		/* packet in use, can be NULL. */
  HickEntry *ev_entry;		/* entry in use, can be NULL. */
  userEntry *ev_ue;		/* userEntry in use, can be NULL. */
  HickVer *ev_ver;		/* version in use, can be NULL. */
  const char *ev_dir;		/* directory name, can be NULL */

  int ev_event;			/* which kind of event was this call? */
} HickEvent;

/* HickEventFunction ******************************************************* */

typedef void (*HickEventFunction)(HickEvent *event, void *arg);

/* Externs ***************************************************************** */

#ifdef __cplusplus
extern "C" {
#endif

/* defined in: packet.c */

extern int openPackets(const char *packpath, const char *homepath,
		       const char *hicktype, const char *group,
		       HickPackets *result);

extern void initHickEntry(HickEntry *entry);

extern void closePackets(HickPackets *pack);

extern HickEntry *freeHickEntry(HickEntry *entry);

extern int nextPacketHeader(HickPackets *pack, HickEntry *result);
extern int nextPacketFull(HickPackets *pack, HickPacket *result);
extern int nextPacket(HickPackets *pack, HickEntry *result, int sabout);

extern HickResult *freeHickResult(HickResult *result);

/* defined in: entry.c */

extern int addPacket(HickPackets *pack, const char *name,  const char *version,
		     int chkVer, int def);

extern int removePacket(HickPackets *pack, const char *name,
			const char *version, int chkVer, int whole);

/* defined in: update.c */

extern int updatePackets(HickPackets *pack, HickResult *result, int doscript,
			 unsigned int mask, HickEventFunction call, void *arg);

/* defined in: gettype.c */

extern int getHickTypes(const char *path, ArrayInfo *result);

/* defined in: rescue.c */

extern int HickCheckRescue(const char *homepath, const char *hicktype);

extern int HickRescueOpen(HickPackets *pack, const char *packpath,
			  const char *homepath, const char *hicktype);

extern void HickRescuePrepare(HickPackets *pack, HickEventFunction call,
			      void *arg);

extern int HickRescueUpdate(HickPackets *pack, HickResult *result,
			    int doscript, unsigned int mask,
			    HickEventFunction call, void *arg);

extern void HickRemoveType(const char *homepath, const char *hicktype,
			   HickEventFunction call, void *arg);

/* defined in: hicked.c  */

extern int loadPacket(const char *name, const char *packpath,
		      const char *hicktype, ArrayInfo *types,
		      HickPacket *result, int sabout);

extern void savePacket(FILE *file, HickPacket *hpac, const char *hicktype,
		       int tlevl);

/* defined in: getpacket.c */

extern void initHickPacket(HickPacket *hpac);

extern HickPacket *freeHickPacket(HickPacket *hpac);
extern void freeHickDirection(HickDirection *head);
extern void freeHickSubGroup(HickSubGroup *head);

/* defined in: version.c */

extern void freeHickVerList(HickVerList *list, int hver);
extern HickVer *freeHickVer(HickVer *ver, int hver);

/* defined in: hickenv.c */

extern HickEnvList *freeHickEnvList(HickEnvList *list, int type);
extern HickEnv *freeHickEnv(HickEnv *env, int type);

#ifdef __cplusplus
}
#endif

#endif /* _HICK_H */
