/* HICK Lib 1.3.2 - User Profiler Library.
** Copyright (C) 97, 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <pwd.h>
#include <errno.h>

#ifdef STDC_HEADERS
# include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FNMATCH
#include <fnmatch.h>
#else
#undef FNM_PATHNAME	/* some sytems got this defined in their unistd.h */
#define FNM_PATHNAME (1 << 0)
#endif

#include "hicksh.h"

#include "common.h"
#include "linker.h"

#define LOGIN_MAX 12

static int linkPath(const char *from, const char *pattern, const char *to,
		    const char *userhome, const char *basedir, FILE *pathFile,
		    FILE *logFile);

static void doSymlink(const char *from, const char *to, const char *todir,
		      const char *toexp, FILE *pathFile, FILE *logFile);

static int isFnmatch(const char *fname);


int
_createLinks(const char *to, const char *userhome, const char *basedir,
	     const char *from, int flag, FILE *pathFile, FILE *logFile)
{
  unsigned int fromLen, toLen;
  char fromName[FNAME_LEN+1], fromDir[FNAME_LEN+1];
  char toName[FNAME_LEN+1], toDir[FNAME_LEN+1], toExp[FNAME_LEN+1];

  if((fromLen = strlen(from)) >= FNAME_LEN) {
    return -1;			/* from path to long. */
  }

  if((toLen = strlen(to)) >= FNAME_LEN) {
    return -2;			/* to path to long. */
  }

  _getFilename(fromName, fromDir, from, (int)fromLen);

  if((flag == LINK_RENAME) || !isFnmatch(fromName)) {
    if(flag == LINK_RENAME) {
      _getFilename(toName, toDir, to, (int)toLen);
    } else {
      strncpy(toDir, to, FNAME_LEN);
      toDir[FNAME_LEN] = (char)0;

      strncpy(toName, fromName, FNAME_LEN);
    }

    if(!_expandPath(toExp, FNAME_LEN, toDir, userhome, basedir)) {
      fprintf(logFile, "link: can't expand path '%s'.\n", toDir);
      return -21;
    }

    if((hsh_path_create(toExp, logFile) == IOW_FAIL) ||
       (hsh_chdir(toExp, NULL) == IOW_FAIL)) {
      fprintf(logFile, "link: can't change/create '%s'.\n", toExp);
      return -22;
    }

    doSymlink(from, toName, toDir, toExp, pathFile, logFile);
  } else {			/* link a file pattern */
    return linkPath(fromDir, fromName, to, userhome, basedir, pathFile,
		    logFile);
  }

  return 0;
}


static int
linkPath(const char *from, const char *pattern, const char *to,
	 const char *userhome, const char *basedir, FILE *pathFile,
	 FILE *logFile)
{
  DIR *dirp;
  struct stat sbuf;
  struct dirent *entry;
  char toFile[FNAME_LEN+1], toDir[FNAME_LEN+1], fname[FNAME_LEN+1];

  if(!(dirp = hsh_opendir(from, logFile))) {
    fprintf(logFile, "link: can't open source directory '%s' for reading.\n",
	    from);
    return -20;
  }

  if(!_expandPath(toFile, FNAME_LEN, to, userhome, basedir)) {
    fprintf(logFile, "link: can't expand path '%s'.\n", to);
    return -21;
  }

  if((hsh_path_create(toFile, logFile) == IOW_FAIL) ||
     (hsh_chdir(toFile, NULL) == IOW_FAIL)) {
    fprintf(logFile, "link: can't change/create '%s'.\n", toFile);
    return -22;
  }

  while((entry = readdir(dirp))) {
    if(IS_DOT_OR_DOTDOT(entry) ||
       fnmatch(pattern, entry->d_name, FNM_PATHNAME)) {
      continue;
    }

    hsh_create_str(fname, FNAME_LEN, from, "/", entry->d_name, 0);

    if(hsh_stat(fname, &sbuf, NULL) == IOW_FAIL) {
      continue;
    }

    if(S_ISDIR(sbuf.st_mode)) {
      hsh_create_str(toDir, FNAME_LEN, to, "/", entry->d_name, 0);

      linkPath(fname, "*", toDir, userhome, basedir, pathFile, logFile);

      hsh_chdir(toFile, NULL);	/* go back to the old directory. */
      continue;
    }

    if(!S_ISREG(sbuf.st_mode)) { /* not a regular file? ignore it.. */
      continue;
    }

    doSymlink(fname, entry->d_name, to, toFile, pathFile, logFile);
  }

  closedir(dirp);
  return 0;
}


static void
doSymlink(const char *from, const char *to, const char *todir,
	  const char *toexp, FILE *pathFile, FILE *logFile)
{
  if(hsh_symlink(from, to, logFile) == IOW_FAIL) {
    fprintf(logFile, "link: failed to link to %s/%s from %s.\n",
	    ((*todir == '/') || (*todir == '~') ? toexp : todir), to,
	    from);
  } else {
    fprintf(pathFile, "%s/%s\n",
	    ((*todir == '/') || (*todir == '~') ? toexp : todir), to);
  }
}


void
_getFilename(char *filename, char *path, const char *s, int len)
{
  int n;

  for(n = len; (n >= 0) && (s[n] != '/'); n--);

  if(n < 0) {
    n = 0;
  }

  if(s[n] != '/') {
    if(filename) {
      strncpy(filename, s, FNAME_LEN);
    }

    if(path) {
      path[0] = (char)0;
    }
  } else {
    if(filename) {
      strncpy(filename, s + n + 1, FNAME_LEN);
      filename[len - n - 1] = (char)0;
    }

    if(path) {
      strncpy(path, s, (size_t)n);
      path[n] = (char)0;
    }
  }
}


int
_removeLinks(FILE *logfile, const char *filename, const char *basedir)
{
  FILE *file;
  unsigned int len;
  char buf[FNAME_LEN+1], fname[FNAME_LEN+1];

  if(!(file = fopen(filename, "r"))) {
    return -1;
  }

  while(fgets(buf, FNAME_LEN, file)) {
    if(buf[(len = strlen(buf)) - 1] == '\n') {
      buf[len - 1] = (char)0;
    }

    if(buf[0] == '/') {		/* got full path specified */
      if(hsh_unlink(buf, logfile) == IOW_FAIL) {
	fprintf(logfile, "link: can't remove '%s' in packet file '%s.\n",
		buf, filename);
      }
    } else {			/* add the basedir */
      hsh_create_str(fname, FNAME_LEN, basedir, "/", buf, 0);

      if(hsh_unlink(fname, logfile) == IOW_FAIL) {
	fprintf(logfile, "link: can't remove '%s' in packet file '%s.\n",
		fname, filename);
      }
    }
  }

  fclose(file);
  return 0;
}


char *
_expandPath(char *buf, unsigned int size, const char *path,
	    const char *userhome, const char *stdpath)
{
  struct passwd *pwd;
  unsigned int i, j, len;
  char logname[LOGIN_MAX+1];

  if(path[0] == '/') {		/* Got a '/', copy it stright over */
    strncpy(buf, path, size);
    buf[size] = (char)0;
    return buf;
  }

  if(path[0] != '~') {		/* other */
    if(!stdpath || !*stdpath) {	/* copy it over else */
      strncpy(buf, path, size);
    } else {			/* we got a standard path, prepend it */
      hsh_create_str(buf, size, stdpath, "/", path, 0);
    }

    buf[size] = (char)0;
    return buf;
  }

  if((path[1] == '/') || (path[1] == (char)0)) {
    /* if we got a ~/ or only a ~, we prepend the userhome. */

    if(!userhome || !*userhome) { /* try to get user dir from passwd */
      if(!(pwd = getpwuid(geteuid())) || !pwd->pw_dir) {
	return NULL;		/* oops, can't find the user, or no dir. */
      }

      if(path[1]) {
	hsh_create_str(buf, size, pwd->pw_dir, "/", (char *)(path + 2), 0);
      } else {
	strncpy(buf, pwd->pw_dir, size);
	buf[size] = (char)0;
      }
    } else {			/* got a specified user home? */
      if(path[1]) {
	hsh_create_str(buf, size, userhome, "/", (char *)(path + 2), 0);
      } else {
	strncpy(buf, userhome, size);
	buf[size] = (char)0;
      }
    }
  } else {			/* get user name */
    len = strlen(path);

    for(j = 0, i = 1; (i < len) && (i < LOGIN_MAX) && (path[i] != '/'); ) {
      logname[j++] = path[i++];
    }

    logname[j] = (char)0;

    if(!(pwd = getpwnam(logname)) || !pwd->pw_dir) {
      return NULL;		/* failed to find the user, or no dir. */
    }

    hsh_create_str(buf, size, pwd->pw_dir, "/", (char *)(path + i + 1), 0);
  }

  buf[size] = (char)0;
  return buf;
}


static int
isFnmatch(const char *fname)
{
  for( ; *fname; fname++) {
    if(*fname == '\\') {
      if(*(++fname) == (char)0) {
	break;
      }
    }

    if((*fname == '*') || (*fname == '?') || (*fname == '[')) {
      return 1;
    }
  }

  return 0;
}
