/* HICK Lib 1.3.2 - User Profiler Library.
** Copyright (C) 97, 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "entry.h"
#include "common.h"

static void _removeType(const char *homepath, const char *hicktype,
			HickEventFunction call, void *arg, int rresc);


int
HickCheckRescue(const char *homepath, const char *hicktype)
{
  struct stat sbuf;
  char fname[FNAME_LEN+1];

  if(!homepath || !hicktype) {
    return 0;
  }

  hsh_create_str(fname, FNAME_LEN, homepath, "/", HICK_DIR, "/.", hicktype,
		 "/.", INSTALLED_FILE, 0);

  if(hsh_stat(fname, &sbuf, NULL) == IOW_OK) {
    return 1;
  }

  hsh_create_str(fname, FNAME_LEN, homepath, "/", HICK_DIR, "/.",
		 INSTALLED_FILE, "-", hicktype, 0);

  return (hsh_stat(fname, &sbuf, NULL) == IOW_OK ? 2 : 0);
}


int
HickRescueOpen(HickPackets *pack, const char *packpath, const char *homepath,
	       const char *hicktype)
{
  int i, f = 0;
  struct stat sbuf;
  char fname[FNAME_LEN+1], fold[FNAME_LEN+1];

  if(!homepath || !hicktype) {
    return 0;
  }

  hsh_create_str(fname, FNAME_LEN, homepath, "/", HICK_DIR, "/.",
		 INSTALLED_FILE, "-", hicktype, 0);

  if(hsh_stat(fname, &sbuf, NULL) == IOW_FAIL) {
    hsh_create_str(fold, FNAME_LEN, homepath, "/", HICK_DIR, "/.", hicktype,
		   "/.", INSTALLED_FILE, 0);

    if(hsh_stat(fold, &sbuf, NULL) == IOW_FAIL) {
      return 1;
    }

    f = 1;
  }

  i = openPackets(packpath, homepath, hicktype, 0, pack);

  if(!f) {
    _freeUserEntryList(&pack->hp_installed);
    _entryLoad(fname, &pack->hp_install);
  } else {
    _entryMoveList(&pack->hp_installed, &pack->hp_install);

    /* move installed file to ~/.hick/.installed-<hicktype> */
    hsh_create_str(fold, FNAME_LEN, homepath, "/", HICK_DIR, "/.", hicktype,
		   "/", INSTALLED_FILE, 0);

    hsh_rename(fold, fname, NULL);
  }

  return i;
}


void
HickRescuePrepare(HickPackets *pack, HickEventFunction call, void *arg)
{
  _removeType(pack->hp_homepath, pack->hp_type, call, arg, 0);
}


int
HickRescueUpdate(HickPackets *pack, HickResult *result, int doscript,
		 unsigned int mask, HickEventFunction call, void *arg)
{
  int i;
  char fname[PNAME_LEN+1];

  hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		 INSTALLED_FILE, "-", pack->hp_type, 0);

  i = updatePackets(pack, result, doscript, mask, call, arg);

  closePackets(pack);

  if(i) {
    return i;
  }

  hsh_unlink(fname, NULL);
  return 0;
}


void
HickRemoveType(const char *homepath, const char *hicktype,
	       HickEventFunction call, void *arg)
{
  _removeType(homepath, hicktype, call, arg, 1);
}


static void
_removeType(const char *homepath, const char *hicktype,
	    HickEventFunction call, void *arg, int rresc)
{
  HickEvent ev;
  char basedir[FNAME_LEN+1], fname[FNAME_LEN+1];

  ev.ev_pack  = NULL;
  ev.ev_hpac  = NULL;
  ev.ev_entry = NULL;
  ev.ev_ue    = NULL;
  ev.ev_ver   = NULL;
  ev.ev_event = HICKEVENT_RDIR;

  hsh_create_str(basedir, FNAME_LEN, homepath, "/", HICK_DIR, 0);

  if(rresc) {			/* remove rescue backupfile? */
    /* .installed-<hicktype> */
    hsh_create_str(fname, FNAME_LEN, basedir, "/.", INSTALLED_FILE, "-",
		   hicktype, 0);
    hsh_unlink(fname, NULL);
  }

  /* Xresources-<hicktype> */
  hsh_create_str(fname, FNAME_LEN, basedir, "/", XRES_FILE, "-",
		 hicktype, 0);
  hsh_unlink(fname, NULL);

  /* call-<hicktype> */
  hsh_create_str(fname, FNAME_LEN, basedir, "/", CALL_SCRIPT, "-",
		 hicktype, 0);
  hsh_unlink(fname, NULL);

  /* <hicktype> */
  hsh_create_str(fname, FNAME_LEN, basedir, "/", hicktype, 0);
  _unlinkRecursive(fname, &ev, call, arg);
  hsh_chdir(basedir, NULL);
  hsh_rmdir(fname, NULL);

  /* .<hicktype> */
  hsh_create_str(fname, FNAME_LEN, basedir, "/.", hicktype, 0);
  _unlinkRecursive(fname, &ev, call, arg);
  hsh_chdir(basedir, NULL);
  hsh_rmdir(fname, NULL);
}


int
_unlinkRecursive(const char *dirname, HickEvent *ev, HickEventFunction call,
		 void *arg)
{
  DIR *dirp;
  int ret = 0;
  struct stat sbuf;
  struct dirent *entry;

  if(!(dirp = hsh_opendir(dirname, NULL))) {
    return 1;
  }

  if(hsh_chdir(dirname, NULL) == IOW_FAIL) {
    return 2;
  }

  if(call) {
    ev->ev_dir = dirname;
    (*call)(ev, arg);
  }

  while((entry = readdir(dirp))) {
    if(IS_DOT_OR_DOTDOT(entry)) {
      continue;
    }

    if((entry->d_name[0] == '.') && (entry->d_name[1] == 'n') &&
       (entry->d_name[2] == 'f') && (entry->d_name[3] == 's')) {
      continue;			/* ignore .nfsXXXX files */
    }

    if((hsh_lstat(entry->d_name, &sbuf, NULL) == IOW_OK) &&
       S_ISDIR(sbuf.st_mode)) {

      _unlinkRecursive(entry->d_name, ev, call, arg);

      if(hsh_rmdir(entry->d_name, NULL) == IOW_FAIL) {
        ret = 3;
      } 
    } else {
      if(hsh_unlink(entry->d_name, NULL) == IOW_FAIL) {
        ret = 3;
      }
    }

    rewinddir(dirp);            /* FIXME: very ugly solution..  */
  }

  closedir(dirp);
  hsh_chdir("..", NULL);
  return ret;
}
