/* HICK Lib 1.3.2 - User Profiler Library.
** Copyright (C) 97, 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <ctype.h>
#include <stdarg.h>

#include "event.h"
#include "entry.h"
#include "common.h"
#include "linker.h"
#include "dofuncs.h"
#include "getfile.h"
#include "version.h"
#include "getpacket.h"

#define DEF_NOT_SET     0
#define DEF_NO_DEFS     1
#define DEF_UPDATE_DEF  2

enum checkFlags {
  DIR_HOME       = HICKERR_HOMEDIR,  /* <homedir>                           */
  DIR_HICKDIR    = HICKERR_HICKDIR,  /* <homedir>/.hick                     */
  DIR_TYPEDIR    = HICKERR_TYPEDIR,  /* <homedir>/.hick/<hicktype>          */
  DIR_HIDDENTYPE = HICKERR_TYPECONF, /* <homedir>/.hick/.<hicktype>         */
  DIR_PACKETS    = HICKERR_INSTPACK, /* <homedir>/.hick/.<hicktype>/packets */
  DIR_SCRIPTS    = HICKERR_SCRIPTDIR /* <homedir>/.hick/.<hicktype>/scripts */
};

static int checkDir(HickPackets *pack, enum checkFlags type);
static void closeFiles(FILE *file, ...);

static void addFailure(userEntry *ue, enum failreason reason,
		       HickResult *result, FILE *logFile);
static void addForced(const char *name, const char *version, 
		      unsigned int hidden, HickResult *result);
static void addRequired(const char *name, HickVerList *vers, 
			userEntryList *inst, unsigned int multiple, 
			FILE *logFile);

extern HickRequire *checkRequire(HickRequire *reqList, userEntryList *install,
				 userEntryList *installed, const char *id, 
				 FILE *logFile);

static userEntry *getValidEntry(userEntryList *list, HickVer *def,
				HickVerList *vers, enum failreason rver,
				enum failreason rdef, HickResult *result,
				FILE *logFile, HickPackets *pack,
				unsigned int mask, HickEventFunction call,
				void *arg);

static void set_entry_flag(enum uentryFlags flag, userEntryList *list);
static void install_default(int flag, const char *name, userEntryList *install,
			    userEntryList *installed, userEntryList *remove,
			    HickVerList *vers);


int
updatePackets(HickPackets *pack, HickResult *result, int doscript,
	      unsigned int mask, HickEventFunction call, void *arg)
{
  time_t curTime;
  int n, instdef;
  unsigned int i;
  HickEntry entry;
  HickPacket hpac;
  struct stat sbuf;
  Buffert instBuf;
  DoubleArray forceArr;
  DIR *typedir, *packdir;
  HickRequire *reqList = 0;
  HickVer *ver, *vertmp, *defver;
  struct dirent *tentry, *pentry;
  userEntry *ue, *uetmp, *uedefI;
  HickVerList instHead, instedHead;
  char fname[FNAME_LEN+1], basedir[FNAME_LEN+1], cwd[FNAME_LEN+1];
  userEntryList isInstalled, installed, install, remove;
  FILE *logFile, *envFile, *infoFile, *menuFile, *xresFile, *packFile;
  FILE *htmlFile, *createdFile, *callFile = NULL;

  if(result) {
    result->hr_required = NULL;
    result->hr_conflicts = NULL;
    result->hr_failure = NULL;
    result->hr_forced = NULL;
  }

  if(hsh_getcwd(cwd, FNAME_LEN, NULL)) {
    cwd[0] = (char)0;
  }

  if((n = checkDir(pack, DIR_HOME)) || (n = checkDir(pack, DIR_HICKDIR)) ||
     (n = checkDir(pack, DIR_TYPEDIR)) || (n=checkDir(pack, DIR_HIDDENTYPE)) ||
     (n = checkDir(pack, DIR_PACKETS)) || (n = checkDir(pack, DIR_SCRIPTS))) {
    return n;
  }

  /* open the log file. */

  hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		 pack->hp_type, "/", LOG_FILE, 0);

  if(!(logFile = fopen(fname, "w"))) {
    return HICKERR_LOGFILE;
  }

  /* clear out the 'scripts' dir from files. */

  hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		 pack->hp_type, "/", SCRIPTS_DIR, 0);

  packdir = hsh_opendir(fname, logFile);

  while((pentry = readdir(packdir))) {
    if(IS_DOT_OR_DOTDOT(pentry)) {
      continue;
    }

    hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		   pack->hp_type, "/", SCRIPTS_DIR, "/", pentry->d_name, 0);

    if(hsh_unlink(fname, logFile) == IOW_FAIL) {
      fprintf(logFile, "update: can't remove '%s' in scripts directory.\n",
	      pentry->d_name);
    }
  }

  closedir(packdir);

  /* Open output files, and initialize the files. */

  hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		 pack->hp_type, "/", ENVIRON_FILE, 0);

  if(!(envFile = fopen(fname, "w"))) {
    fclose(logFile);
    return HICKERR_ENVFILE;
  }

  hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		 pack->hp_type, "/", MENUS_FILE, 0);

  if(!(menuFile = fopen(fname, "w"))) {
    closeFiles(logFile, envFile, 0);
    return HICKERR_MENUFILE;
  }

  hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		 pack->hp_type, "/", INFO_FILE, 0);

  if(!(infoFile = fopen(fname, "w"))) {
    closeFiles(logFile, envFile, menuFile, 0);
    return HICKERR_INFOFILE;
  }

  hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		 pack->hp_type, "/", CREATED_FILE, 0);

  hsh_create_str(basedir, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/",
		 pack->hp_type, 0);

  _removeLinks(logFile, fname, basedir);

  if(!(createdFile = fopen(fname, "w"))) {
    closeFiles(logFile, envFile, menuFile, infoFile, 0);
    return HICKERR_CREATEDFILE;
  }

  hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		 pack->hp_type, "/", HTML_FILE, 0);

  if(!(htmlFile = fopen(fname, "w"))) {
    closeFiles(logFile, envFile, menuFile, infoFile, createdFile, 0);
    return HICKERR_HTMLFILE;
  }

  hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/",
		 XRES_FILE, "-", pack->hp_type, 0);

  if(!(xresFile = fopen(fname, "w"))) {
    closeFiles(logFile, envFile, menuFile, infoFile, htmlFile, createdFile, 0);
    return HICKERR_XRESFILE;
  }

  fputs(XRES_HEADER, xresFile);

  if(doscript) {
    hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/",
		   CALL_SCRIPT, "-", pack->hp_type, 0);

    i = hsh_stat(fname, &sbuf, logFile); /* store file stat. */

    if(!(callFile = fopen(fname, "a"))) {
      closeFiles(logFile, envFile, menuFile, infoFile, xresFile, htmlFile, 
		 createdFile, 0);
      return HICKERR_CALLFILE;
    }

    hsh_chmod(fname, 493, logFile); /* set mode 755 -rwxr-xr-x */

    /* If the file already exist we don't want to write the header again. */
    if(i == IOW_FAIL) {
      fprintf(logFile, "call: creating call file '%s'.\n", fname);
      fputs(CALL_HEADER, callFile);
    } else {
      fprintf(logFile, "call: appending to call file '%s'.\n", fname);
    }
  }

  /* Try to open packet types directory */

  if(!(typedir = hsh_opendir(pack->hp_packpath, logFile))) {
    closeFiles(envFile, menuFile, infoFile, xresFile, logFile, htmlFile,
	       createdFile, callFile, 0);
    return HICKERR_PACKDIR;
  }

  hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		 pack->hp_type, "/.", INSTALLED_FILE, 0);

  if(!(instBuf.file = fopen(fname, "w"))) {
    closeFiles(logFile, envFile, menuFile, infoFile, xresFile, htmlFile,
	       createdFile, callFile, 0);
    return HICKERR_INSTFILE;
  }

  instBuf.pos = 0;

  /* Start reading the packet info and parse the information. */

  time(&curTime);

  initHickEntry(&entry);
  initHickPacket(&hpac);

  initHickVerList(&instHead);
  initHickVerList(&instedHead);

  initUserEntryList(&remove);
  initUserEntryList(&install);
  initUserEntryList(&installed);
  initUserEntryList(&isInstalled);

  while(((tentry = readdir(typedir)))) {
    if(IS_DOT_OR_DOTDOT(tentry) ||
       (strcasecmp(tentry->d_name, COMMON_DIR) &&
	strcasecmp(tentry->d_name, pack->hp_type))) {
      continue;
    }

    hsh_create_str(fname, FNAME_LEN, pack->hp_packpath, "/", tentry->d_name,0);

    if(!(packdir = hsh_opendir(fname, logFile))) {
      continue;
    }

    while((pentry = readdir(packdir))) {
      hsh_create_str(fname, FNAME_LEN, pack->hp_packpath, "/", tentry->d_name,
		     "/", pentry->d_name, 0);

      if((hsh_stat(fname, &sbuf, logFile) == IOW_FAIL) ||
	 !S_ISREG(sbuf.st_mode)) {
	continue;
      }

      i = NAMLEN(pentry) - 1;

      if(IS_EMACS_BACKUP(pentry, i) || !(packFile = fopen(fname, "r"))) {
	continue;
      }

      while(!feof(packFile)) {
	freeHickPacket(&hpac);
	freeHickEntry(&entry);

	if(!_getPacketHead(packFile, &entry)) {
	  break;
	}

	_getPacket(packFile, 0, &hpac, 0);

	freeHickVerList(&instHead, 1);
	freeHickVerList(&instedHead, 1);

	/* initialize the remove, install, and installed lists */

	_freeUserEntryList(&remove);
	_freeUserEntryList(&install);
	_freeUserEntryList(&installed);

	_entryMoveName(entry.he_name, &pack->hp_remove, &remove);
	_entryMoveName(entry.he_name, &pack->hp_install, &install);
	_entryMoveName(entry.he_name, &pack->hp_installed, &installed);

	/* Check if the packet or any version has expired? */

	if(hpac.pac_expire && (hpac.pac_expire < curTime)) {
	  _entryRemoveMatchVersion(&remove, &install);
	  _entryRemoveMatchVersion(&remove, &installed);
	  _entryRemoveMatchVersion(&install, &installed);

	  for(ue = remove.ul_head; ue; ue = ue->ue_next) {
	    addFailure(ue, FAIL_EXPIRED_PACKET, result, logFile);
	  }

	  for(ue = install.ul_head; ue; ue = ue->ue_next) {
	    addFailure(ue, FAIL_EXPIRED_PACKET, result, logFile);
	  }

	  for(ue = installed.ul_head; ue; ue = ue->ue_next) {
	    addFailure(ue, FAIL_EXPIRED_PACKET, result, logFile);

	    if((ver = _versionFind(ue->ue_version, &hpac.pac_versions))) {
	      _doRemove(NULL, &hpac, ver, &entry, pack, logFile, 1, callFile,
			mask, call, arg);
	    } else {
	      _doRemove(ue, NULL, NULL, NULL, pack, logFile, 1, callFile,
			mask, call, arg);
	    }
	  }

	  _callEvent(mask, HICKEVENT_NOTUSED, call, arg, pack, &hpac, &entry,
		     NULL, NULL);
	  continue;		/* the packet has expired, do nothing. */
	}

	for(ver = hpac.pac_versions.vl_head; ver; ) {
	  if(ver->ver_expire && (ver->ver_expire < curTime)) {
	    for(ue = remove.ul_head; ue; ) {
	      uetmp = ue->ue_next;
	      if(!strcasecmp(ue->ue_version, ver->ver_number)) {
		addFailure(ue, FAIL_EXPIRED_VERSION, result, logFile);
		FreeMem(_entryRemove(ue, &remove));
	      }
	      ue = uetmp;
	    }

	    for(ue = install.ul_head; ue; ) {
	      uetmp = ue->ue_next;
	      if(!strcasecmp(ue->ue_version, ver->ver_number)) {
		addFailure(ue, FAIL_EXPIRED_VERSION, result, logFile);
		FreeMem(_entryRemove(ue, &remove));
	      }
	      ue = uetmp;
	    }

	    for(ue = installed.ul_head; ue; ) {
	      uetmp = ue->ue_next;
	      if(!strcasecmp(ue->ue_version, ver->ver_number)) {
		addFailure(ue, FAIL_EXPIRED_VERSION, result, logFile);
		_doRemove(NULL, &hpac, ver, &entry, pack, logFile, 1,
			  callFile, mask, call, arg);
		FreeMem(_entryRemove(ue, &remove));
	      }
	      ue = uetmp;
	    }

	    vertmp = ver->ver_next;
	    FreeMem(freeHickVer(_versionRemove(ver, &hpac.pac_versions), 1));
	    ver = vertmp;
	  } else {
	    ver = ver->ver_next;
	  }
	}

	/* Is there any force information applying for this user? */

	_checkDirection(hpac.pac_force, PERM_FORCE, &forceArr);

	if(forceArr.arr1.siz) {
	  /* FORCE INSTALL ************************************************* */

	  /* Check install list for this packet, and give errors. */

	  for(ue = install.ul_head; ue; ue = install.ul_head) {
	    install.ul_head = ue->ue_next;

	    if(!hsh_array_find(ue->ue_version, &forceArr.arr1)) {
	      addFailure(ue, FAIL_FORCE_INSTALLED, result, logFile);
	    }

	    FreeMem(ue);
	  }

	  /* set all the force versions to install into the install list. */

	  for(n = 0; n < forceArr.arr1.siz; n++) {
	    if(!_versionFind(forceArr.arr1.arr[n], &hpac.pac_versions)) {
	      continue;
	    }

	    addForced(entry.he_name, forceArr.arr1.arr[n], hpac.pac_hidden,
		      result);

	    _entryNew(entry.he_name, forceArr.arr1.arr[n], curTime, UE_FORCE,
		      &install);

	    if(!hpac.pac_multiple) {
	      break;
	    }
	  }

	  /* Go trough remove list and remove packets. */

	  for(ue = remove.ul_head; ue; ue = ue->ue_next) {
	    if((ue->ue_flags == UE_RWHOLE) ||
	       hsh_array_find(ue->ue_version, &forceArr.arr1)) {
	      addFailure(ue, FAIL_REMOVE_FORCE, result, logFile);
	    } else {
	      if((ver = _versionFind(ue->ue_version, &hpac.pac_versions))) {
		_doRemove(NULL, &hpac, ver, &entry, pack, logFile, 1,
			  callFile, mask, call, arg);
	      } else {		/* missing version */
		addFailure(ue, FAIL_REMOVE_VERSION, result, logFile);
		_doRemove(ue, NULL, NULL, &entry, pack, logFile, 1, callFile,
			  mask, call, arg);
	      }
	    }
	  }

	  _freeUserEntryList(&remove);

	  /* Check for other versions installed of this packet. */

	  for(ue = installed.ul_head; ue; ) {
	    uetmp = ue->ue_next;

	    if(!(ver = _versionFind(ue->ue_version, &hpac.pac_versions))) {
	      _doRemove(ue, NULL, NULL, NULL, pack, logFile, 1, callFile, mask,
			call, arg);
	      FreeMem(_entryRemove(ue, &installed));
	    } else {
	      uedefI =_entryFindTwo(ue->ue_name, ue->ue_version, &install);

	      if(uedefI) {
		if((hpac.pac_modified && (hpac.pac_modified > ue->ue_date)) ||
		   (ver->ver_modified && (ver->ver_modified > ue->ue_date))) {
		  _doRemove(NULL, &hpac, ver, &entry, pack, logFile, 1,
			    callFile, mask, call, arg);
		  FreeMem(_entryRemove(ue, &installed));
		} else {
		  FreeMem(_entryRemove(uedefI, &install));
		}
	      } else {		/* force remove */
		_doRemove(NULL, &hpac, ver, &entry, pack, logFile, 1,
			  callFile, mask, call, arg);

		if(ue->ue_flags != UE_FORCE) {
		  addFailure(ue, FAIL_FORCE_REMOVED, result, logFile);
		}

		FreeMem(_entryRemove(ue, &installed));
	      }
	    }
		
	    ue = uetmp;
	  }
	} else {
	  /* remove packets user dosen't have access to. */

	  n = _versionPermissions(&hpac.pac_versions, hpac.pac_allow,
				  hpac.pac_deny, &instHead);

	  if(instHead.vl_head) {
	    for(ue = installed.ul_head; ue; ) {
	      uetmp = ue->ue_next;

	      if((ver = _versionFind(ue->ue_version, &instHead))) {
		if(ue->ue_flags != UE_FORCE) {
		  addFailure(ue, FAIL_MISSING_VERSION, result, logFile);
		}

		_doRemove(NULL, &hpac, ver, &entry, pack, logFile, 1,
			  callFile, mask, call, arg);
		FreeMem(_entryRemove(ue, &installed));
	      }
	      ue = uetmp;
	    }
	  }

	  freeHickVerList(&instHead, 1); /* free the versions. */

	  if(!n) {		/* no access availible, get next packet */
	    _callEvent(mask, HICKEVENT_NOTUSED, call, arg, pack, &hpac, &entry,
		       NULL, NULL);
	    continue;
	  }

	  /* find out which versions is the default one. */

	  for(ver = hpac.pac_versions.vl_head; ver && !ver->ver_default;
	      ver = ver->ver_next);

	  defver = ver;

	  /* if we got remove whole packet, we'll move those versions now
	   * into the remove list.
	   */

	  for(ue = remove.ul_head; ue; ue = uetmp) {
	    uetmp = ue->ue_next;

	    if(ue->ue_flags == UE_RWHOLE) {
	      FreeMem(_entryRemove(ue, &remove));
	      _entryMoveList(&installed, &remove);
	      break;		/* nothing more needed to do ;) */
	    }
	  }

	  /* remove packet which was force installed */

	  for(ue = installed.ul_head; ue; ue = uetmp) {
	    uetmp = ue->ue_next;

	    if(ue->ue_flags == UE_FORCE) {
	      _entryMove(ue, &installed, &remove);
	    }
	  }

	  /* ONE PACKET **************************************************** */

	  if(!hpac.pac_multiple) { /* Only to install ONE packet? */
	    _entryRemoveMatchVersion(&remove, &installed);

	    ue = getValidEntry(&install, defver, &hpac.pac_versions,
			       FAIL_NOTFOUND_VERSION, FAIL_NOTFOUND_DEFAULT,
			       result, logFile, NULL, mask, call, arg);

	    _freeUserEntryList(&install);

	    if(ue) {		/* We got a valid version in install list? */
	      _entryMoveList(&installed, &remove);

	      if(ue->ue_flags == UE_DEFAULT) {
		strcpy(ue->ue_version, defver->ver_number);
	      }

	      _entryPrepend(ue, &install);
	    } else {		/* check installed packet */
	      ue = getValidEntry(&installed, defver, &hpac.pac_versions,
				 FAIL_MISSING_VERSION, FAIL_MISSING_DEFAULT,
				 result, logFile, pack, mask, call, arg);

	      if(ue) {		/* We got a valid version in installed list? */
		_entryMoveList(&installed, &remove);

		if(ue->ue_flags == UE_DEFAULT) {
		  if(strcasecmp(defver->ver_number, ue->ue_version)) {
		      _entryPrepend(ue, &remove);
		      _entryNew(entry.he_name, defver->ver_number, curTime, 1,
				&install);
		  } else {
		    if((hpac.pac_modified &&(hpac.pac_modified >ue->ue_date))||
		       (defver->ver_modified &&
			(defver->ver_modified > ue->ue_date))) {
		      _entryPrepend(ue, &remove);
		      _entryNew(entry.he_name, defver->ver_number,
				(hpac.pac_modified > defver->ver_modified ?
				 hpac.pac_modified : defver->ver_modified), 1,
				&install);
		    } else {
		      _entryPrepend(ue, &installed);
		    }
		  }
		} else {
		  ver = _versionFind(ue->ue_version, &hpac.pac_versions);

		  if((hpac.pac_modified && (hpac.pac_modified > ue->ue_date))||
		     (ver->ver_modified && (ver->ver_modified > ue->ue_date))){
		    _entryPrepend(ue, &remove);
		    _entryNew(entry.he_name, ue->ue_version,
			      (hpac.pac_modified > ver->ver_modified ?
			       hpac.pac_modified : ver->ver_modified), 0,
			      &install);
		  } else {
		    _entryPrepend(ue, &installed);
		  }
		}
	      }
	    }
	  } else {
	  /* MULTIPLE PACKET *********************************************** */

	    if(install.ul_head) {
	      instdef = 0;

	      for(ue = install.ul_head; ue; ue = uetmp) {
		uetmp = ue->ue_next;

		if(ue->ue_flags == UE_DEFAULT) {
		  if(!defver) {
		    addFailure(ue, FAIL_NOTFOUND_DEFAULT, result, logFile);
		  }

		  FreeMem(_entryRemove(ue, &install));
		  instdef = 1;
		} else {
		  ver = _versionFind(ue->ue_version, &hpac.pac_versions);

		  if(!ver) {
		    addFailure(ue, FAIL_NOTFOUND_VERSION, result, logFile);
		    FreeMem(_entryRemove(ue, &install));
		  } else {
		    _entryMove(_entryFindVersion(ue->ue_version, &installed),
			       &installed, &remove);
		  }
		}
	      }

	      install_default(instdef, entry.he_name, &install, &installed,
			      &remove, &hpac.pac_versions);
	    }

	    if(installed.ul_head) {
	      instdef = 0;

	      for(ue = installed.ul_head; ue; ue = uetmp) {
		uetmp = ue->ue_next;

		ver = _versionFind(ue->ue_version, &hpac.pac_versions);

		if(!ver) {
		  addFailure(ue, FAIL_MISSING_VERSION, result, logFile);
		  _entryMove(ue, &installed, &remove);
		  continue;
		} else {
		  if((ue->ue_flags == UE_DEFAULT) && (instdef != -1)) {
		    instdef = 1;

		    if(!defver) {
		      addFailure(ue, FAIL_MISSING_DEFAULT, result, logFile);
		      _entryMove(ue, &installed, &remove);
		      continue;
		    }

		    if(!ver->ver_default) {
		      _entryMove(ue, &installed, &remove);
		      continue;
		    }
		  } else {
		    instdef = -1;
		  }

		  if((hpac.pac_modified > ue->ue_date) ||
		     (ver->ver_modified > ue->ue_date)) {
		    _entryMove(ue, &installed, &remove);
		    _entryNew(entry.he_name, ue->ue_version,
			      (hpac.pac_modified > ver->ver_modified ?
			       hpac.pac_modified : ver->ver_modified),
			      ue->ue_flags, &install);
		  }
		}
	      }

	      if(instdef == -1) {
		set_entry_flag(UE_NONE, &installed);
	      } else {
		if(instdef == 1) {
		  for(ver = hpac.pac_versions.vl_head; ver;
		      ver = ver->ver_next) {

		    if(ver->ver_default &&
		       !_entryFindVersion(ver->ver_number, &installed) &&
		       !_entryFindVersion(ver->ver_number, &install)) {
		      _entryNew(entry.he_name, ver->ver_number, 0, UE_DEFAULT,
				&install);
		    }
		  }
		}
	      }
	    }
	  }

	  /* ONE PACKET, and MULTIPLE PACKETS ****************************** */

	  for(ue = remove.ul_head; ue; ue = ue->ue_next) {
	    if((ver = _versionFind(ue->ue_version, &hpac.pac_versions))) {
	      _doRemove(NULL, &hpac, ver, &entry, pack, logFile, 1, callFile,
			mask, call, arg);
	    } else {
	      if((ue->ue_flags == UE_RWHOLE) || (ue->ue_flags == UE_NONE)) {
		addFailure(ue, FAIL_REMOVE_VERSION, result, logFile);
	      }

	      _doRemove(ue, NULL, NULL, NULL, pack, logFile, 1, callFile, mask,
			call, arg);
	    }
	  }

	  /* Got a required packet? and no version selected? */
	  if(hpac.pac_required && !install.ul_head && !installed.ul_head) {
	    addRequired(entry.he_name, &hpac.pac_versions, &install,
			hpac.pac_multiple, logFile);
	  }
	}

	/* FORCE, ONE PACKET, and MULTIPLE PACKETS ************************* */

	if(!install.ul_head && !installed.ul_head) {
	  _callEvent(mask, HICKEVENT_NOTUSED, call, arg, pack, &hpac, &entry,
		     NULL, NULL);
	  continue;		/* nothing to do with the packet. */
	}

	for(ue = install.ul_head; ue; ue = ue->ue_next) {
	  if((ver = _versionFind(ue->ue_version, &hpac.pac_versions))) {
	    _versionMove(ver, &hpac.pac_versions, &instHead);
	  }
	}

	for(ue = installed.ul_head; ue; ue = ue->ue_next) {
	  if((ver = _versionFind(ue->ue_version, &hpac.pac_versions))) {
	    _versionMove(ver, &hpac.pac_versions, &instedHead);
	  }
	}

	if(_doConflict(&hpac, &entry, &instHead, &instedHead, &isInstalled,
		       result, logFile)) {
	  _callEvent(mask, HICKEVENT_NOTUSED, call, arg, pack, &hpac, &entry,
		     NULL, NULL);
	  continue;		/* Got a conflict. */
	}

	_entryMoveList(&install, &isInstalled);
	_entryMoveList(&installed, &isInstalled);

	if(mask & HICKEVENT_BEFORE) {
	  if(mask & HICKEVENT_VERSION) {
	    _callVersionEvents(&instHead, mask, HICKEVENT_INSTALL, call, arg,
			       pack, &hpac, &entry);
	    _callVersionEvents(&instedHead, mask, HICKEVENT_UPDATE, call, arg,
			       pack, &hpac, &entry);
	  } else {
	    if(instHead.vl_head) { /* this might not be that good with multi */
	      _callEvent(mask, HICKEVENT_INSTALL, call, arg, pack, &hpac,
			 &entry, NULL, NULL);
	    } else {
	      _callEvent(mask, HICKEVENT_UPDATE, call, arg, pack, &hpac,
			 &entry, NULL, NULL);
	    }
	  }
	}

	_doRequire(&hpac, &entry, &instHead, &instedHead, &reqList);

	_doCommand(_doEnviron, &hpac, pack, &instHead, &instedHead, logFile,
		   envFile, NULL);

	_doCommand(_doInfo, &hpac, pack, &instHead, &instedHead, logFile,
		   infoFile, NULL);

	_doCommand(_doHTML, &hpac, pack, &instHead, &instedHead, logFile,
		   htmlFile, NULL);

	_doCommand(_doMenuItem, &hpac, pack, &instHead, &instedHead, logFile,
		   menuFile, NULL);

	_doCommand(_doXRes, &hpac, pack, &instHead, &instedHead, logFile,
		   xresFile, NULL);

	_doCommand(_doScript, &hpac, pack, &instHead, &instedHead, logFile,
		   NULL,  NULL);

	_doCommand(_doCreate, &hpac, pack, &instHead, &instedHead, logFile,
		   createdFile, basedir);

	/* Is this if the first time we install this, or are we about to
	 * update packet information, it's the same thing anyway.
	 */

	if(instHead.vl_head) {
	  _doLink(&hpac, pack, &entry, &instHead, logFile);

	  _doCall(&hpac, &entry, &instHead, logFile, callFile,
		  HFLAG_CALL_INSTALL, pack->hp_homepath);
	} else {
	  _doCall(&hpac, &entry, &instedHead, logFile, callFile,
		  HFLAG_CALL_UPDATE, pack->hp_homepath);
	}

	if(!(mask & HICKEVENT_BEFORE)) {
	  if(mask & HICKEVENT_VERSION) {
	    _callVersionEvents(&instHead, mask, HICKEVENT_INSTALL, call, arg,
			       pack, &hpac, &entry);
	    _callVersionEvents(&instedHead, mask, HICKEVENT_UPDATE, call, arg,
			       pack, &hpac, &entry);
	  } else {
	    if(instHead.vl_head) { /* this might not be that god with multi. */
	      _callEvent(mask, HICKEVENT_INSTALL, call, arg, pack, &hpac,
			 &entry, NULL, NULL);
	    } else {
	      _callEvent(mask, HICKEVENT_UPDATE, call, arg, pack, &hpac,
			 &entry, NULL, NULL);
	    }
	  }
	}
      }

      fclose(packFile);		/* close the current packet file. */
    }

    closedir(packdir); /* close the current packet directory. */
  }

  freeHickPacket(&hpac);
  freeHickEntry(&entry);
  freeHickVerList(&instHead, 1);
  freeHickVerList(&instedHead, 1);
  _freeUserEntryList(&remove);
  _freeUserEntryList(&install);
  _freeUserEntryList(&installed);

  closedir(typedir); /* close the Hick type directory. */

  /* Remove packets we haven't yet removed, no script will be called */

  for(ue = pack->hp_remove.ul_head; ue; ue = ue->ue_next) {
    addFailure(ue, FAIL_REMOVE_PACKET, result, logFile);
    _doRemove(ue, NULL, NULL, NULL, pack, logFile, 0, callFile, mask, call,
	      arg);
  }

  for(ue = pack->hp_install.ul_head; ue; ue = ue->ue_next) {
    addFailure(ue, FAIL_NOTFOUND_PACKET, result, logFile);
  }

  for(ue = pack->hp_installed.ul_head; ue; ue = ue->ue_next) {
    addFailure(ue, FAIL_MISSING_PACKET, result, logFile);
    _doRemove(ue, NULL, NULL, NULL, pack, logFile, 0, callFile, mask, call,
	      arg);
  }

  /* Store a list of packets which we have successfully installed. */

  for(ue = isInstalled.ul_head; ue; ue = ue->ue_next) {
    bprint(&instBuf, "%d:%d:%C:%C\n", (!ue->ue_date ? curTime : ue->ue_date),
	   ue->ue_flags, ue->ue_name, ue->ue_version);
  }

  bflush(&instBuf);

  _freeUserEntryList(&pack->hp_remove);
  _freeUserEntryList(&pack->hp_install);
  _freeUserEntryList(&pack->hp_installed);

  _entryMoveList(&isInstalled, &pack->hp_installed);

  /* Close the open files. */

  closeFiles(envFile, menuFile, infoFile, xresFile, instBuf.file, htmlFile,
	     createdFile, callFile, 0);

  /* Now we create a list of required packets. */

  reqList = checkRequire(reqList, &pack->hp_install, &pack->hp_installed,
			 "require:", logFile);


  if(result) {
    result->hr_required = reqList;
  }

  if(cwd[0]) {			/* change back to the stored current dir. */
    hsh_chdir(cwd, logFile);
  }

  hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		 pack->hp_type, "/.", INSTALLED_FILE, 0);

  hsh_create_str(basedir, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/.",
		 pack->hp_type, "/", INSTALLED_FILE, 0);

  if(hsh_rename(fname, basedir, logFile) == IOW_FAIL) {
    fprintf(logFile, "update: failed to move '%s' to '%s'.\n", fname, basedir);
    return HICKERR_RENAME;
  }

  fprintf(logFile, "Last runned at %s\n", asctime(localtime(&curTime)));
  fclose(logFile);
  return 0;
}


static int
checkDir(HickPackets *pack, enum checkFlags type)
{
  struct stat sbuf;
  char fname[FNAME_LEN+1];

  switch(type) {
    case DIR_HOME: {		/* <homedir> */
      strncpy(fname, pack->hp_homepath, FNAME_LEN);
      fname[FNAME_LEN] = (char)0;
      break;
    }
    case DIR_HICKDIR: {		/* <homedir>/.hick */
      hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, 0);
      break;
    }
    case DIR_TYPEDIR: {		/* <homedir>/.hick/<hicktype> */
      hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR, "/",
		     pack->hp_type, 0);
      break;
    }
    case DIR_HIDDENTYPE: {	/* <homedir>/.hick/.<hicktype> */
      hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR,
		     "/.", pack->hp_type, 0);
      break;
    }
    case DIR_PACKETS: {		/* <homedir>/.hick/.<hicktype>/packets */
      hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR,
		     "/.", pack->hp_type, "/", PACKETS_DIR, 0);
      break;
    }
    case DIR_SCRIPTS: {		/* <homedir>/.hick/.<hicktype>/scripts */
      hsh_create_str(fname, FNAME_LEN, pack->hp_homepath, "/", HICK_DIR,
		     "/.", pack->hp_type, "/", SCRIPTS_DIR, 0);
      break;
    }
  }

  if(hsh_stat(fname, &sbuf, NULL) == IOW_OK) {
    if(!S_ISDIR(sbuf.st_mode)) {
      return type;		/* Oh, bugger it's not a directory. */
    }

    if(((sbuf.st_uid == geteuid()) &&
	(sbuf.st_mode & (S_IRUSR | S_IXUSR | S_IWUSR))) || 
       ((sbuf.st_gid == getegid()) &&
	(sbuf.st_mode & (S_IRGRP | S_IXGRP | S_IWGRP))) ||
       ((sbuf.st_mode & (S_IROTH | S_IXOTH | S_IWOTH)))) {
      return 0;			/* We can access it! with read and write :) */
    }

    return (hsh_chmod(fname, MKDIR_MODE, NULL) == IOW_OK ? 0 : type);
  }

  return (hsh_mkdir(fname, MKDIR_MODE, NULL) == IOW_OK ? 0 : type);
}


static void
closeFiles(FILE *file, ...)
{
  va_list pvar;

  fclose(file);

  va_start(pvar, file);

  while((file = va_arg(pvar, FILE *))) {
    fclose(file);
  }

  va_end(pvar);
}


static void
addFailure(userEntry *ue, enum failreason reason, HickResult *result,
	   FILE *logFile)
{
  HickFailure *fail;

  fprintf(logFile, "fail: can't install %s %s.\n", ue->ue_name,
	  ue->ue_version);

  if(result && (fail = MallocMem(HickFailure))) {
    fail->fail_next = result->hr_failure;
    result->hr_failure = fail;

    strcpy(fail->fail_name, ue->ue_name);
    strcpy(fail->fail_version, ue->ue_version);
    fail->fail_reason = reason;
  }
}


static void
addForced(const char *name, const char *version, unsigned int hidden, 
	  HickResult *result)
{
  HickForced *fo;

  if(result && (fo = MallocMem(HickForced))) {
    fo->fo_next = result->hr_forced;
    result->hr_forced = fo;

    fo->fo_hidden = hidden;
    strcpy(fo->fo_name, name);
    strcpy(fo->fo_version, version);
  }
}


static void
addRequired(const char *name, HickVerList *vers, userEntryList *inst, 
	    unsigned int multiple, FILE *logFile)
{
  HickVer *ver;
  int found_default = 0;

  for(ver = vers->vl_head; ver; ver = ver->ver_next) {
    if(ver->ver_default) {
      fprintf(logFile, "update: added required '%s' version '%s'.\n",
	      name, ver->ver_number);

      _entryNew(name, ver->ver_number, 0, UE_DEFAULT, inst);

      if(!multiple) {
	return;
      }

      found_default = 1;
    }
  }

  if(!found_default && vers->vl_head) {
    ver = vers->vl_head;
    _entryNew(name, ver->ver_number, 0, UE_DEFAULT, inst);
  }
}


static int
testRequire(HickRequirePacket *rpac, userEntryList *install,
	    userEntryList *installed)
{
  int i;
  userEntry *ue;

  if((ue = _entryFind(rpac->rpac_name, install)) ||
     (ue = _entryFind(rpac->rpac_name, installed))) {
    if(!rpac->rpac_versions.siz) { /* any version? */
      return 1;
    }

    for(i = 0; i < rpac->rpac_versions.siz; i++) {
      if(!strcasecmp(rpac->rpac_versions.arr[i], ue->ue_version)) {
	return 1;		/* versions matches! */
      }
    }
  }

  if(rpac->rpac_or) {		/* if we got an OR statement, check it to. */
    return testRequire(rpac->rpac_or, install, installed);
  }

  return 0;
}


HickRequire *
checkRequire(HickRequire *reqList, userEntryList *install, 
	     userEntryList *installed, const char *id, FILE *logFile)
{
  int i;
  HickRequire *req, *rold;
  HickRequirePacket *rpac, *rpold, *rptmp, *ror;

  for(rold = 0, req = reqList; req; ) {
    for(rpold = 0, rpac = req->req_list; rpac; ) {
      if(testRequire(rpac, install, installed)) {
	if(rpold) {
	  rpold->rpac_next = rpac->rpac_next;
	} else {
	  req->req_list = rpac->rpac_next;
	}

	/* Free the packet information. */

	for(rptmp = rpac; rptmp; rptmp = rpac) {
	  rpac = rptmp->rpac_or;

	  hsh_array_remove(&rptmp->rpac_versions);
	  FreeMem(rptmp);
	}

	rpac = (rpold ? rpold->rpac_next : req->req_list);
      } else {
        rpold = rpac;
        rpac = rpac->rpac_next;
      }
    }

    /* No required packets left? Remove it then. */

    if(!req->req_list) {
      if(rold) {
	rold->req_next = req->req_next;
      } else {
	reqList = req->req_next;
      }

      FreeMem(req);

      req = (rold ? rold : reqList);
      continue;
    }

    rold = req;
    req = req->req_next;
  }

  /* print the require information to the logfile. */

  for(req = reqList; req; req = req->req_next) {
    for(rpac = req->req_list; rpac; rpac = rpac->rpac_next) {
      fprintf(logFile, id);

      for(ror = rpac; ror; ror = ror->rpac_or) {
	fprintf(logFile, " %s", ror->rpac_name);

	if(ror->rpac_versions.siz == 0) {
	  fprintf(logFile, " any version");
	} else {
	  if(ror->rpac_versions.siz == 1) {
	    fprintf(logFile, " %s", ror->rpac_versions.arr[0]);
	  } else {
	    fprintf(logFile, " {");
	    for(i = 0; i < ror->rpac_versions.siz; i++) {
	      fprintf(logFile, " %s", ror->rpac_versions.arr[i]);
	    }
	    fprintf(logFile, " } ");
	  }
	}

	if(ror->rpac_or) {
	  fprintf(logFile, " or");
	}
      }

      fprintf(logFile, " is required by %s %s.\n", req->req_name,
	      req->req_version);
    }
  }


  return reqList;
}


static userEntry *
getValidEntry(userEntryList *list, HickVer *def, HickVerList *vers,
	      enum failreason rver, enum failreason rdef, HickResult *result,
	      FILE *logFile, HickPackets *pack, unsigned int mask,
	      HickEventFunction call, void *arg)
{
  HickVer *ver;
  userEntry *ue, *ret = NULL;

  for(ue = list->ul_head; ue; ue = list->ul_head) {
    list->ul_head = ue->ue_next;

    if(ue->ue_flags == UE_DEFAULT) {
      if(!def) {
	if(pack) {
	  _doRemove(ue, NULL, NULL, NULL, pack, logFile, 1, 0, mask, call,arg);
	}

	addFailure(ue, rdef, result, logFile);
	FreeMem(ue);
	continue;
      }
    } else {
      if(!(ver = _versionFind(ue->ue_version, vers))) {
	if(pack) {
	  _doRemove(ue, NULL, NULL, NULL, pack, logFile, 1, 0, mask, call,arg);
	}

	addFailure(ue, rver, result, logFile);
	FreeMem(ue);
	continue;
      }
    }

    if(ret) {
      if(pack) {
	_doRemove(ue, NULL, NULL, NULL, pack, logFile, 1, 0, mask, call, arg);
      }
      FreeMem(ue);
    } else {
      ret = ue;
    }
  }

  return ret;
}


static void
set_entry_flag(enum uentryFlags flag, userEntryList *list)
{
  userEntry *ue;

  for(ue = list->ul_head; ue; ue = ue->ue_next) {
    ue->ue_flags = flag;
  }
}


static void
install_default(int flag, const char *name, userEntryList *install,
		userEntryList *installed, userEntryList *remove,
		HickVerList *vers)
{
  int hasinst;
  HickVer *ver;

  if(flag) {
    _entryMoveList(installed, remove);

    hasinst = (install->ul_head ? 1 : 0);

    for(ver = vers->vl_head; ver; ver = ver->ver_next){
      if(ver->ver_default) {
	_entryNew(name, ver->ver_number, 0, (hasinst ? UE_NONE : UE_DEFAULT),
		  install);
      }
    }
  } else {
    set_entry_flag(UE_NONE, installed);
  }
}
