/* HICKMK Lib 1.2.1 - HICK Make Utility
** Copyright (C) 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>

#include "common.h"
#include "hickmk.h"
#include "utils.h"
#include "mkinterp.h"

typedef struct {
  NodeName nname;		/* item name information */
} HtmlItem;


typedef struct {
  NodeName nname;		/* topic name information */

  NodeList items;		/* list of items for this topic */
} HtmlTopic;


static void loop_topic(MKLoop *ldata, MKInterp *interp, NodeList *stmts);
static void loop_htmlitem(MKLoop *ldata, MKInterp *interp, NodeList *stmts);
static void loop_item(MKLoop *ldata, MKInterp *interp, NodeList *stmts);

static void html_load(const char *filename, NodeList *list);
static void html_free(NodeList *list);


char **
hickmk_html_get(const char *htmlpath, size_t *siz)
{
  return class_get(htmlpath, HTML_FILE, MKC_HTML, siz);
}

void
hickmk_html_remove(const char *homepath, const char *hicktype)
{
  char fname[FNAME_LEN+1];

  hsh_create_str(fname, FNAME_LEN, homepath, "/", HICK_DIR, "/", HTML_DIR,
                 "/", hicktype, 0);

  _unlinkRecursive(fname, NULL, NULL, NULL);
}

int
hickmk_html_create(const char **names, unsigned int nameSize,
		   const char *htmlpath, const char *homepath,
		   const char *hicktype)
{
  size_t n;
  Node *node;
  MKInterp interp;
  NodeList topics, funcs;
  char fname[FNAME_LEN+1], path[FNAME_LEN+1];

  hsh_nodelist_init(&funcs);
  hsh_nodelist_init(&topics);
  mkinterp_init(&interp);
 
  hsh_create_str(fname, FNAME_LEN, htmlpath, "/", HTML_FILE, 0);

  if((n = mklang_functions_load(fname, &funcs))) {
    return n;
  }

  hsh_create_str(path, FNAME_LEN, homepath, "/", HICK_DIR, "/", HTML_DIR,
                 "/", hicktype, 0);

  if(hsh_path_create(path, NULL) == IOW_FAIL) {
    return -1;
  }

  hsh_create_str(fname, FNAME_LEN, homepath, "/", HICK_DIR, "/.", hicktype,
		 "/", HTML_FILE, 0);

  html_load(fname, &topics);

  hsh_create_str(fname, FNAME_LEN, htmlpath, "/", hicktype, 0);

  interp.userhome = homepath;
  interp.stdpath  = fname;
  
  for(node = funcs.head; node; node = node->next) {
    if(((MKFunction *)node)->fclass == MKC_HTML) {
      if(names) {
        for(n = 0; n < nameSize; n++) {
          if(!strcasecmp(names[n], ((MKFunction *)node)->name)) {
            break;
          }
        }

        if(n >= nameSize) {
          continue;
        }
      }

      mkloop_set(&interp, "topic",    loop_topic,    &topics);
      mkloop_set(&interp, "item",     loop_item,     NULL);
      mkloop_set(&interp, "htmlitem", loop_htmlitem, &topics);

      mkvar_set(&interp, "ITEM",     "",         NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "TOPIC",    "",         NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "HICKTYPE", hicktype,   NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "USER",     hmk_getlogin(), NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "PATH",     path,       NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "FILE",     "",         NULL, NULL, 1, 0, 0);
      mkvar_set(&interp, "FIRST",    "",         NULL, NULL, 1, 0, 1);
      mkvar_set(&interp, "LAST",     "",         NULL, NULL, 1, 0, 1);

      mkinterp_exec_stmt(&((MKFunction *)node)->stmt, &interp);
      mkinterp_clear(&interp);
    }
  }

  html_free(&topics);
  mklang_functions_free(&funcs);
  return 0;
}


static void
loop_topic(MKLoop *ldata, MKInterp *interp, NodeList *stmts)
{
  Node *node;

  for(node = ((NodeList *)ldata->data)->head; node; node = node->next) {
    mkloop_set(interp, "item", loop_item, &((HtmlTopic *)node)->items);
    mkvar_fl_set(interp, (NodeList *)ldata->data, node);
    mkvar_set(interp, "TOPIC", ((NodeName *)node)->name, NULL, NULL, 1, 0, 1);
    mkinterp_exec_stmts(stmts, interp);
  }

  mkloop_set(interp, "item", loop_item, NULL);
  mkvar_set(interp, "FIRST", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "LAST",  "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "TOPIC", "", NULL, NULL, 1, 0, 1);
}


static void
loop_htmlitem(MKLoop *ldata, MKInterp *interp, NodeList *stmts)
{
  Node *node, *inode;

  for(node = ((NodeList *)ldata->data)->head; node; node = node->next) {
    mkvar_set(interp, "TOPIC", ((NodeName *)node)->name, NULL, NULL, 1, 0, 1);

    for(inode = ((HtmlTopic *)node)->items.head; inode; inode = inode->next) {
      if((((NodeList *)ldata->data)->head == node) &&
	 (((HtmlTopic *)node)->items.head == inode)) {
	mkvar_set(interp, "FIRST", "1", NULL, NULL, 1, 0, 1);
      } else {
	mkvar_set(interp, "FIRST", "", NULL, NULL, 1, 0, 1);
      }

      if((((NodeList *)ldata->data)->tail == node) &&
	 (((HtmlTopic *)node)->items.tail == inode)) {
	mkvar_set(interp, "LAST", "1", NULL, NULL, 1, 0, 1);
      } else {
	mkvar_set(interp, "LAST", "", NULL, NULL, 1, 0, 1);
      }

      mkvar_set(interp, "ITEM", ((NodeName *)inode)->name, NULL, NULL, 1, 0,1);
      mkinterp_exec_stmts(stmts, interp);
    }
  }

  mkvar_set(interp, "FIRST", "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "LAST",  "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "ITEM",  "", NULL, NULL, 1, 0, 1);
  mkvar_set(interp, "TOPIC", "", NULL, NULL, 1, 0, 1);
}


static void
loop_item(MKLoop *ldata, MKInterp *interp, NodeList *stmts)
{
  Node *node;

  if(!ldata->data) {
    return;
  }

  for(node = ((NodeList *)ldata->data)->head; node; node = node->next) {
    mkvar_fl_set(interp, (NodeList *)ldata->data, node);
    mkvar_set(interp, "ITEM", ((NodeName *)node)->name, NULL, NULL, 1, 0, 1);
    mkinterp_exec_stmts(stmts, interp);
  }

  mkvar_set(interp, "ITEM", "", NULL, NULL, 1, 0, 1);
}


static void
html_load(const char *filename, NodeList *list)
{
  FILE *file;
  char *s1, *s2;
  HtmlItem *item;
  HtmlTopic *topic;

  if(!(file = fopen(filename, "r"))) {
    return;
  }

  while((s1 = _getCString(file))) { /* Topic */
    if(!(s2 = _getCString(file))) { /* Item */
      FreeMem(s1);
      return;
    }

    if(!(topic = (HtmlTopic *)hsh_node_name_find(s1, list))) {
      topic = MallocMem(HtmlTopic);
      topic->nname.name = s1;
      hsh_nodelist_init(&topic->items);

      hsh_node_name_insert(&topic->nname, list, 0);
    } else {
      FreeMem(s1);
    }

    item = MallocMem(HtmlItem);
    item->nname.name = s2;
    hsh_node_name_insert(&item->nname, &topic->items, 0);
  }
}


static void
html_free(NodeList *list)
{
  Node *topic, *item;

  for(topic = list->head; topic; topic = list->head) {
    list->head = topic->next;

    for(item = ((HtmlTopic *)topic)->items.head; item; 
	item = ((HtmlTopic *)topic)->items.head) {
      ((HtmlTopic *)topic)->items.head = item->next;

      FreeMem(((NodeName *)item)->name);
      FreeMem((HtmlItem *)item);
    }

    FreeMem(((NodeName *)topic)->name);
    FreeMem((HtmlTopic *)topic);
  }
}
