/* HICKMK Lib 1.2.1 - HICK Make Utility
** Copyright (C) 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdarg.h>

#include "getfile.h"
#include "getpacket.h"

#include "mklang.h"

#define TOKEN_LEN 16

#define DEFAULT_MODE_DIR   493       /* 755 drwxr-xr-x */
#define DEFAULT_MODE_FILE  420       /* 622 -rw-r--r-- */

static int mklang_load_stmt(FILE *file, NodeList *list, unsigned int in_type);

static MKExpr *mkexpr_get(FILE *file, char *value1, int is_stmt);
static MKExpr *mkexpr_new(unsigned int op, char *value);
static void mkexpr_free(MKExpr *expr);

static MKStmt *mkstmt_new(unsigned int token, char *value, NodeList *list);
static void mkstmt_free(MKStmt *stmt);

static void mklang_ungets(FILE *file, char *value);
static int mklang_get_name(FILE *file, char *value, size_t maxlen);
static int mklang_get_string(FILE *file, char **value);
static int mklang_get_mode(FILE *file, char **value, mode_t *mode, 
			   unsigned int type);

static int is_octal(const char *str);
static unsigned int get_octal(const char *str);


static FileToken tokens[] = {
  { "dir",     MKT_DIR     },
  { "file",    MKT_FILE    },
  { "foreach", MKT_FOREACH },
  { "read",    MKT_READ    },
  { "write",   MKT_WRITE   },
  { "if",      MKT_IF      },
  { "unset",   MKT_UNSET   },
  { NULL, 0 }
};

static FileToken op_token[] = {
  { "==", MKO_EQUAL     },
  { "!=", MKO_NOT_EQUAL },
  { "=",  MKO_SET       },
  { NULL, 0 }
};

static FileToken path_tokens[] = {
  { "dir",   MKT_DIR  },
  { "file" , MKT_FILE },
  { NULL, 0 }
};

static FileToken class_tokens[] = {
  { "menu",   MKC_MENU   },
  { "script", MKC_SCRIPT },
  { "html",   MKC_HTML   },
  { NULL, 0 }
};


/* mklang_load_functions()
**   0  successfull, -1  can't open file, -2  failed parsing file,
**  -3  out of memory
*/

int
mklang_functions_load(const char *filename, NodeList *list)
{
  FILE *file;
  mode_t mode;
  int c, cnum;
  unsigned int tnum;
  MKFunction *mkfunc;
  char fclass[TOKEN_LEN+1], *name, token[TOKEN_LEN+1], *value;

  if(!(file = fopen(filename, "r"))) {
    return -1;
  }

  while((c = _getSpaces(file)) != EOF) {
    if(c == '#') {
      _getComment(file);
      continue;
    }

    ungetc(c, file);

    if(!mklang_get_name(file, fclass, TOKEN_LEN) || 
       !mklang_get_string(file, &name)) {
      break;
    }

    if(!mklang_get_name(file, token, TOKEN_LEN) || 
       !mklang_get_string(file, &value)) {
      FreeMem(name);
      break;
    }

    if((cnum = _token_lookup(fclass, class_tokens)) == -1) { 
      mklang_error(MKTLang, MKError, "invalid class '%s'.\n", fclass);
      FreeMem(name);
      FreeMem(value);
      continue;
    }

    if((tnum = _token_lookup(token, path_tokens)) == -1) {
      mklang_error(MKTLang, MKError, 
		   "invalid path token '%s', use dir or file.\n", token);
      FreeMem(name);
      FreeMem(value);
      continue;
    }

    if(!mklang_get_mode(file, &value, &mode, tnum)) {
      FreeMem(name);
      break;
    }

    if(!(mkfunc = MallocMem(MKFunction))) {
      mklang_error(MKTLang, MKFatal, "out of memory!\n");
      return -3;
    }

    mkfunc->fclass          = (enum MKClass)cnum;
    mkfunc->name            = name;
    mkfunc->stmt.token      = tnum;
    mkfunc->stmt.value      = value;
    mkfunc->stmt.u.mode     = mode;
    hsh_nodelist_init(&mkfunc->stmt.stmt1);

    if((tnum = mklang_load_stmt(file, &mkfunc->stmt.stmt1, tnum))) {
      mklang_error(MKTLang, MKError, "failed reading function '%s'.\n", name);
      return tnum;
    }

    hsh_node_append(&mkfunc->stmt.node, list);
  }

  fclose(file);
  return 0;
}


void
mklang_functions_free(NodeList *list)
{
  Node *node, *snode;

  for(node = list->head; node; node = list->head) {
    list->head = node->next;

    for(snode = ((MKFunction *)node)->stmt.stmt1.head; snode; 
	snode = ((MKFunction *)node)->stmt.stmt1.head) {
      ((MKFunction *)node)->stmt.stmt1.head = snode->next;

      mkstmt_free((MKStmt *)snode);
    }

    if(((MKFunction *)node)->name) {
      FreeMem(((MKFunction *)node)->name);
    }

    FreeMem(((MKFunction *)node)->stmt.value);
    FreeMem((MKFunction *)node);
  }
}


static int
mklang_load_stmt(FILE *file, NodeList *list, unsigned int in_type)
{
  mode_t mode;
  MKExpr *expr;
  int c, n, ret = 0;
  MKStmt *stmt, *tstmt;
  unsigned int tnum;
  char *token, *value;

  if((c = _getSpaces(file)) != '{') {
    mklang_error(MKTLang, MKError, "MKLang: missing a '{'.\n");
    return -2;
  }

  while(((c = _getSpaces(file)) != EOF) && (c != '}')) {
    if(c == '#') {
      _getComment(file);
      continue;
    }

    ungetc(c, file);

    if(!mklang_get_string(file, &token)) {
      return -2;
    }

    if((tnum = _token_lookup(token, tokens)) == -1) {
      if(!(expr = mkexpr_get(file, token, 1))) {
	mklang_error(MKTLang, MKWarning, "unknown token '%s'.\n", token);
	FreeMem(token);
      }

      if((stmt = mkstmt_new(MKT_EXPR, NULL, list))) {
	stmt->u.expr = expr;
      } else {
	mkexpr_free(expr);
      }
      continue;
    }

    FreeMem(token);

    switch(tnum) {
      case MKT_DIR:
      case MKT_FILE: {
	if(!mklang_get_string(file, &value) ||
	   !mklang_get_mode(file, &value, &mode, tnum)) {
	  return -2;
	}

	if(!(stmt = mkstmt_new(tnum, value, list))) {
	  FreeMem(value);
	  return -3;
	}

	stmt->u.mode = mode;

	if((n = mklang_load_stmt(file, &stmt->stmt1, tnum))) {
	  hsh_node_remove(&stmt->node, list);
	  mkstmt_free(stmt);
	  return n;
	}

	if(in_type == MKT_FILE) {
	  ret = -2;
	  mklang_error(MKTLang, MKError, "can't use '%s' in a 'file'.\n",
		       (tnum == MKT_DIR ? "dir" : "file"));
	  hsh_node_remove(&stmt->node, list);
	}
	break;
      }
      case MKT_FOREACH: {
	if(!mklang_get_string(file, &value)) {
	  return -2;
	}

	if(!(stmt = mkstmt_new(tnum, value, list))) {
	  FreeMem(value);
	  return -3;
	}

	if((n = mklang_load_stmt(file, &stmt->stmt1, in_type))) {
	  hsh_node_remove(&stmt->node, list);
	  mkstmt_free(stmt);
	  return n;
	}

	break;
      }
      case MKT_READ: {
	if(!(tstmt = mkstmt_new(tnum, NULL, NULL))) {
	  return -3;
	}

	do {
	  if(!mklang_get_string(file, &value)) {
	    mkstmt_free(tstmt);
	    return -2;
	  }

	  if(!(stmt = mkstmt_new(MKT_READ, value, &tstmt->stmt1))) {
	    mkstmt_free(tstmt);
	    FreeMem(value);
	    return -3;
	  }
	} while((c = _getSpaces(file)) == '|');

	ungetc(c, file);

	if(in_type == MKT_DIR) {
	  ret = -2;
	  mklang_error(MKTLang, MKError, "can't use 'read' in a 'dir'.\n");
	  mkstmt_free(tstmt);
	} else {
	  hsh_node_append(&tstmt->node, list);
	}
	break;
      }
      case MKT_WRITE: {
	if(!mklang_get_string(file, &value)) {
	  return -2;
	}

	if(in_type == MKT_DIR) {
	  FreeMem(value);
	  ret = -2;
	  mklang_error(MKTLang, MKError, "can't use 'write' in a 'dir'.\n");
	} else {
	  if(!(stmt = mkstmt_new(tnum, value, list))) {
	    FreeMem(value);
	    return -3;
	  }
	}
	break;
      }
      case MKT_IF: {
	if((c = _getSpaces(file)) != '(') {
	  ret = -2;
	  mklang_error(MKTLang, MKError, 
		       "missing '(' at begining of if statment.\n");
	  break;
	}

	if(!mklang_get_string(file, &value)) {
	  return -2;
	}

	if(!(expr = mkexpr_get(file, value, 0))) {
	  mklang_error(MKTLang, MKError, "failed reading expression.\n");
	  return -2;
	}

	if(!(stmt = mkstmt_new(tnum, NULL, list))) {
	  mkexpr_free(expr);
	  return -3;
	}

	stmt->u.expr = expr;

	if((c = _getSpaces(file)) != ')') {
	  ungetc(c, file);
	  ret = -2;
	  mklang_error(MKTLang, MKError,
		       "missing ')' at end of if statment.\n");
	  break;
	}

	if((n = mklang_load_stmt(file, &stmt->stmt1, in_type))) {
	  mklang_error(MKTLang, MKError, "failed reading if statement.\n");
	  return n;
	}

	if(!mklang_get_string(file, &value)) {
	  return -2;
	}

	if(!_token_case_cmp(value, "else")) { /* else statement? */
	  if((n = mklang_load_stmt(file, &stmt->stmt2, in_type))) {
	    FreeMem(value);
	    mklang_error(MKTLang, MKError, 
			 "failed reading if else statement.\n");
	    return n;
	  }
	} else {
	  mklang_ungets(file, value);
	}

	FreeMem(value);
	break;
      }
      case MKT_UNSET: {
	if(!mklang_get_string(file, &value)) {
	  return -2;
	}

	if(!(stmt = mkstmt_new(tnum, value, list))) {
	  FreeMem(value);
	  return -3;
	}
	break;
      }
    }
  }

  return ret;
}


static MKExpr *
mkexpr_get(FILE *file, char *value1, int is_stmt)
{
  MKExpr *expr;
  unsigned int onum;
  char *op, *value2;

  if(!mklang_get_string(file, &op)) {
    return NULL;
  }

  if((onum = _token_lookup(op, op_token)) == -1) { /* not a valid operator? */
    mklang_ungets(file, op);
    FreeMem(op);

    if(!is_stmt) {		/* in if statement? it's ok then. */
      return mkexpr_new(MKO_ID, value1);
    }

    return NULL;
  }
   
  FreeMem(op);

  if(!mklang_get_string(file, &value2)) {
    return NULL;
  }

  if(is_stmt && ((onum == MKO_EQUAL) || (onum == MKO_NOT_EQUAL))) {
    mklang_error(MKTLang, MKWarning, "statement with no effect.\n");
    return NULL;
  }

  if((expr = mkexpr_new(onum, NULL))) {
    expr->node.next = (Node *)mkexpr_new(MKO_ID, value1);
    expr->node.prev = (Node *)mkexpr_new(MKO_ID, value2);
  }

  return expr;
}


static MKExpr *
mkexpr_new(unsigned int op, char *value)
{
  MKExpr *expr;

  if(!(expr = MallocMem(MKExpr))) {
    mklang_error(MKTLang, MKFatal, "out of memory!\n");
    return NULL;
  }

  expr->op        = op;
  expr->value     = value;
  expr->node.next = NULL;
  expr->node.prev = NULL;
  return expr;
}


static void
mkexpr_free(MKExpr *expr)
{
  if(expr->node.next) {
    mkexpr_free((MKExpr *)expr->node.next);
  }

  if(expr->node.prev) {
    mkexpr_free((MKExpr *)expr->node.prev);
  }

  if(expr->value) {
    FreeMem(expr->value);
  }

  FreeMem(expr);
}


static MKStmt *
mkstmt_new(unsigned int token, char *value, NodeList *list)
{
  MKStmt *stmt;

  if(!(stmt = MallocMem(MKStmt))) {
    mklang_error(MKTLang, MKFatal, "out of memory!\n");
    return NULL;
  }

  stmt->token = token;
  stmt->value = value;
  stmt->node.next = stmt->node.prev = NULL;
  hsh_nodelist_init(&stmt->stmt1);
  hsh_nodelist_init(&stmt->stmt2);

  if(list) {
    hsh_node_append(&stmt->node, list);
  }

  return stmt;
}


static void
mkstmt_free(MKStmt *stmt)
{
  Node *node;

  for(node = stmt->stmt1.head; node; node = stmt->stmt1.head) {
    stmt->stmt1.head = node->next;

    mkstmt_free((MKStmt *)node);
  }

  for(node = stmt->stmt2.head; node; node = stmt->stmt2.head) {
    stmt->stmt2.head = node->next;

    mkstmt_free((MKStmt *)node);
  }

  if(((stmt->token == MKT_EXPR) || (stmt->token == MKT_IF)) &&
     (stmt->u.expr)) {
    mkexpr_free(stmt->u.expr);
  }

  if(stmt->value) {
    FreeMem(stmt->value);
  }

  FreeMem(stmt);
}


static void
mklang_ungets(FILE *file, char *value)
{
  size_t n, len;

  if((len = strlen(value))) {
    ungetc(' ', file);

    for(n = len; n > 0; n--) {
      ungetc(value[n - 1], file);
    }
  }
}


static int
mklang_get_name(FILE *file, char *value, size_t maxlen)
{
  if(!_getName(file, value, maxlen)) {
    mklang_error(MKTLang, MKError, "unexpected end of file.\n");
    return 0;
  }

  return 1;
}


static int
mklang_get_string(FILE *file, char **value)
{
  if(!(*value = _getString(file))) {
    mklang_error(MKTLang, MKError, "unexpected end of file.\n");
    return 0;
  }

  return 1;
}


static int
mklang_get_mode(FILE *file, char **value, mode_t *mode, unsigned int type)
{
  *mode = (type == MKT_DIR ? DEFAULT_MODE_DIR : DEFAULT_MODE_FILE);

  if(is_octal(*value)) {	/* got a octal number? */
    *mode = get_octal(*value);

    FreeMem(*value);

    if(!mklang_get_string(file, value)) {
      return 0;
    }
  }

  return 1;
}


static int 
is_octal(const char *str)
{
  const char *p;

  for(p = str; *p; p++) {
    if(!((*p >= '0') && (*p <= '7'))) {
      return 0;
    }
  }

  return 1;
}


static unsigned int
get_octal(const char *str)
{
  const char *p;
  unsigned int mode;

  for(mode = 0, p = str; ((*p >= '0') && (*p <= '7')); p++) {
    mode = mode * 8 + *p - '0';
  }

  return mode;
}


void
mklang_error(enum MKLang_Type type, enum MKLang_Err err, const char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);

  fprintf(stderr, "%s: %s: ", (type == MKTInterp ? "MKInterp" : 
			       (type == MKTParse ? "MKParse" : "MKLang")),
	  (err == MKError ? "Error" :
	   (err == MKFatal ? "FATAL ERROR" : "Warning")));

  vfprintf(stderr, fmt, ap);
}
