/* HickSH Lib 1.1.1 - HICK Make Utility
** Copyright (C) 98, 99, 2000 Daniel Bergstroem <daniel@sdf.se>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>

#include "nodes.h"


Node *
hsh_node_prepend(Node *node, NodeList *to)
{
  NODE_IS_NOT_NULL("_node_prepend", node, to);

  node->prev = NULL;
  node->next = to->head;
  to->head = (to->head ? (to->head->prev = node) : (to->tail = node));
  return node;
}


Node *
hsh_node_append(Node *node, NodeList *to)
{
  NODE_IS_NOT_NULL("_node_append", node, to);

  node->next = NULL;
  node->prev = to->tail;
  to->tail = (to->tail ? (to->tail->next = node) : (to->head = node));
  return node;
}


Node *
hsh_node_remove(Node *node, NodeList *from)
{
#ifdef NODE_PROTECTED
  NODE_IS_NOT_NULL("_node_remove", node, from);

  if(node->prev) {		/* can we safely remove the node? */
    NODE_PTR_NOT_EQUAL("_node_remove", node->prev->next, node);
  } else {
    NODE_PTR_NOT_EQUAL("_node_remove", from->head, node);
  }

  if(node->next) {
    NODE_PTR_NOT_EQUAL("_node_remove", node->next->prev, node);
  } else {
    NODE_PTR_NOT_EQUAL("_node_remove", from->tail, node);
  }
#endif /* _NODE_PROTECTED */

  if(node->prev) {
    node->prev->next = node->next;
  } else {
    from->head = node->next;
  }

  if(node->next) {
    node->next->prev = node->prev;
  } else {
    from->tail = node->prev;
  }

  node->next = node->prev = NULL;
  return node;
}


Node *
hsh_node_insert_after(Node *node, Node *after, NodeList *to)
{
  node->prev = after->prev;
  node->next = after;
  after->prev = node;

  if(node->prev) {
    node->prev->next = node;
  } else {
    to->head = node;
  }

  return node;
}


NodeName *
hsh_node_name_find(const char *name, NodeList *list)
{
  Node *node;

  NODE_IS_NOT_NULL("_node_name_find", name, list);

  for(node = list->head; node; node = node->next) {
    if(!strcasecmp(name, ((NodeName *)node)->name)) {
      return (NodeName *)node;
    }
  }

  return NULL;
}


Node *
hsh_node_name_insert(NodeName *nname, NodeList *to, int rev)
{
  Node *node;

  NODE_IS_NOT_NULL("_node_name_insert", nname, to);

  if(rev) {			/* highest value last */
    if(!to->head || 
       (strcasecmp(nname->name, ((NodeName *)to->head)->name) >= 0)) {
      return hsh_node_prepend(&nname->node, to);
    }

    for(node = to->head; node; node = node->next) {
      if(strcasecmp(nname->name, ((NodeName *)node)->name) >= 0) {
	return hsh_node_insert_after(&nname->node, node, to);
      }
    }
  } else {			/* highest value first */
    if(!to->head || 
       (strcasecmp(nname->name, ((NodeName *)to->head)->name) <= 0)) {
      return hsh_node_prepend(&nname->node, to);
    }

    for(node = to->head; node; node = node->next) {
      if(strcasecmp(nname->name, ((NodeName *)node)->name) <= 0) {
	return hsh_node_insert_after(&nname->node, node, to);
      }
    }
  }

  return hsh_node_append(&nname->node, to);
}


Node *
hsh_node_num_insert(NodeNum *nnum, NodeList *to, int rev)
{
  Node *node;

  NODE_IS_NOT_NULL("_node_num_insert", nnum, to);

  if(rev) {			/* highest value last */
    if(!to->head || (nnum->num <= ((NodeNum *)to->head)->num)) {
      return hsh_node_prepend(&nnum->node, to);
    }

    for(node = to->head; node; node = node->next) {
      if(nnum->num <= ((NodeNum *)node)->num) {
	return hsh_node_insert_after(&nnum->node, node, to);
      }
    }
  } else {			/* highest value first */
    if(!to->head || (nnum->num >= ((NodeNum *)to->head)->num)) {
      return hsh_node_prepend(&nnum->node, to);
    }

    for(node = to->head->next; node; node = node->next) {
      if(nnum->num >= ((NodeNum *)node)->num) {
	return hsh_node_insert_after(&nnum->node, node, to);
      }
    }
  }

  return hsh_node_append(&nnum->node, to);
}
